#include "control.h"
#include <string.h>
#include <iostream>
using namespace std;

int Base::MAX_LENGTH = Control::PRINT_LINE_SIZE - 4;

//
// Constructor
//
Base::Base(Control *control_) : control(control_),
                                option(control_ -> option),
                                grammar(control_ -> grammar),
                                node_pool(control_ -> node_pool)
{}


//
// MKFIRST constructs the FIRST and FOLLOW maps, the CLOSURE map,
// ADEQUATE_ITEM and ITEM_TABLE maps and all other basic maps.
//
void Base::Process(void)
{
    //
    // Initialize global arrays.
    //
    next_rule.Resize(grammar -> num_rules + 1);
    first_item_of.Resize(grammar -> num_rules + 1);
    next_item.Resize(grammar -> num_items + 1);
    lhs_rule.Resize(grammar -> num_terminals + 1, grammar -> num_symbols);
    lhs_rule.Initialize(NIL);
    nt_items.Resize(grammar -> num_terminals + 1, grammar -> num_symbols);
    nt_items.Initialize(NIL);

    item_table.Resize(grammar -> num_items + 1);
    null_nt.Resize(grammar -> num_terminals + 1, grammar -> num_symbols);
    null_nt.MemReset();

    closure.Resize(grammar -> num_terminals + 1, grammar -> num_symbols);
    clitems.Resize(grammar -> num_terminals + 1, grammar -> num_symbols);

    //
    // NT_FIRST is used to construct a mapping from non-terminals to the
    // set of terminals taht may appear first in a string derived from
    // the non-terminal.
    //
    nt_first.Resize(grammar -> num_terminals + 1, grammar -> num_symbols);
    for (int i = grammar -> FirstNonTerminal(); i <= grammar -> LastNonTerminal(); i++)
        nt_first[i].Initialize(grammar -> num_terminals + 1);

    //
    // In this loop, we construct the LHS_RULE map which maps
    // each non-terminal symbol into the set of rules it produces
    //
    for (int rule_no = grammar -> FirstRule(); rule_no <= grammar -> LastRule(); rule_no++)
    {
        int symbol = grammar -> rules[rule_no].lhs;
        if (lhs_rule[symbol] == NIL)
            next_rule[rule_no] = rule_no;
        else
        {
            next_rule[rule_no] = next_rule[lhs_rule[symbol]];
            next_rule[lhs_rule[symbol]]  = rule_no;
        }
        lhs_rule[symbol] = rule_no;
    }

    //
    // Check if there are any non-terminals that do not produce
    // any rules.
    //
    NoRulesProduced();

    //
    // Construct the CLOSURE map of non-terminals.
    //
    index_of.Resize(grammar -> num_terminals + 1, grammar -> num_symbols);
    index_of.Initialize(OMEGA);
    for (int nt = grammar -> FirstNonTerminal(); nt <= grammar -> LastNonTerminal(); nt++)
    {
        if (index_of[nt] == OMEGA)
            ComputeClosure(nt);
    }

    //
    // Construct the NULL_NT map for non-terminals.
    // A non-terminal B is said to be nullable if either:
    //
    //    B -> %empty  or  B -> B1 B2 B3 ... Bk  where Bi is
    //                         nullable for 1 <= i <= k
    //
    NullablesComputation();

    //
    // Construct the FIRST map for non-terminals and also a list
    // of non-terminals whose first set is empty.
    //
    index_of.Initialize(OMEGA);
    for (int nt = grammar -> FirstNonTerminal(); nt <= grammar -> LastNonTerminal(); nt++)
    {
        if (index_of[nt] == OMEGA)
            ComputeFirst(nt);
    }

    //
    //  Since every input source will be followed by the EOFT
    //  symbol, FIRST[accept_image] cannot contain empty but
    //  instead must contain the EOFT symbol.
    //
    if (null_nt[grammar -> accept_image])
    {
        null_nt[grammar -> accept_image] = false;
        nt_first[grammar -> accept_image].RemoveElement(grammar -> empty);
        nt_first[grammar -> accept_image].AddElement(grammar -> eof_image);
    }

    //
    // Check whether there are any non-terminals that do not
    // generate any terminal strings. If so, signal error and stop.
    //
    CheckNonTerminals();

    //
    // Construct the ITEM_TABLE, FIRST_ITEM_OF, and NT_ITEMS maps.
    //
    first_table.Resize(grammar -> num_symbols + 1);
    first_table.Initialize(NIL);

    assert(first.Length() == 0);
    first.Next(); // skip zeroth element

    int first_of_empty = first.NextIndex(); // use 1st slot for empty
    first[first_of_empty].suffix_root = 1;
    first[first_of_empty].suffix_tail = 0;

    int item_no = 0;
    item_table[item_no].rule_number = 0;
    item_table[item_no].symbol = grammar -> empty;
    item_table[item_no].dot = 0;
    item_table[item_no].suffix_index = NIL;

    for (int rule_no = grammar -> FirstRule(); rule_no <= grammar -> LastRule(); rule_no++)
    {
        first_item_of[rule_no] = item_no + 1;
        int j = 0,
            k = grammar -> rules[rule_no + 1].rhs_index - 1;
        for (int i = grammar -> FirstRhsIndex(rule_no); i < grammar -> EndRhsIndex(rule_no); i++)
        {
            item_no++;
            int symbol = grammar -> rhs_sym[i];
            item_table[item_no].rule_number = rule_no;
            item_table[item_no].symbol = symbol;
            item_table[item_no].dot = j;

            if (option -> lalr_level > 1 ||
                grammar -> IsNonTerminal(symbol) ||
                symbol == grammar -> error_image)
            {
                if (i == k)
                    item_table[item_no].suffix_index = first_of_empty;
                else
                    item_table[item_no].suffix_index = FirstMap(i + 1, k);
            }
            else
                item_table[item_no].suffix_index = NIL;

            if (grammar -> IsNonTerminal(symbol))
            {
                next_item[item_no] = nt_items[symbol];
                nt_items[symbol] = item_no;
            }
            j++;
        }

        item_table[++item_no].rule_number = rule_no;
        item_table[item_no].symbol = grammar -> empty;
        item_table[item_no].dot = j;
        item_table[item_no].suffix_index = NIL;
    }


    //
    // We now compute the first set for all suffixes that were
    // inserted in the FIRST map.
    // Extra space is also allocated to compute the first set for
    // suffixes whose left-hand side is the ACCEPT non-terminal.
    // The first set for these suffixes are the sets needed to
    // construct the FOLLOW map and compute look-ahead sets.  They
    // are placed in the FIRST table in the range 1..NUM_FIRST_SETS
    // The first element in the FIRST table contains the first sets
    // for the empty sequence.
    //
    for (int i = 1; i < first.Length(); i++)
    {
        first[i].set.Initialize(grammar -> num_terminals + 1);
        SuffixFirst(i);
    }

    //
    // There must be only one rule whose left-hand side is accep_image.
    //
    int index = first.NextIndex();
    item_no = first_item_of[lhs_rule[grammar -> accept_image]];
    item_table[item_no].suffix_index = index;
    first[index].set.Initialize(grammar -> num_terminals + 1);
    first[index].set.AddElement(grammar -> eof_image);

    //
    // If the READ/REDUCE option is on, we precalculate the kernel
    // of the final states which simply consists of the last item
    // in  the corresponding rule.  Rules with the ACCEPT
    // non-terminal as their left-hand side are not considered so
    // as to let the Accpet action remain as a Reduce action
    // instead of a Goto/Reduce action.
    //
    if (option -> read_reduce)
    {
        adequate_item.Resize(grammar -> num_rules + 1);
        adequate_item.MemReset();
        for (int rule_no = grammar -> FirstRule(); rule_no <= grammar -> LastRule(); rule_no++)
        {
            int rhs_size = grammar -> RhsSize(rule_no);
            if (grammar -> rules[rule_no].lhs != grammar -> accept_image)
            {
                int item_no = first_item_of[rule_no] + rhs_size;
                Node *p = node_pool -> AllocateNode();
                p -> value = item_no;
                p -> next = NULL;
                adequate_item[rule_no] = p;
            }
        }
    }


    //
    // Construct the CLITEMS map. Each element of CLITEMS points
    // to a circular linked list of items.
    //
    for (int nt = grammar -> FirstNonTerminal(); nt <= grammar -> LastNonTerminal(); nt++)
    {
        assert(clitems[nt].Length() == 0);
        int rule_no;
        for (bool end_node = ((rule_no = lhs_rule[nt]) == NIL);
             ! end_node;
             end_node = (rule_no == lhs_rule[nt]))
        {
            rule_no = next_rule[rule_no];

            clitems[nt].Next() = first_item_of[rule_no];
        }
    }

    //
    // If LALR_LEVEL > 1, we need to calculate RMPSELF, a set that
    // identifies the nonterminals that can right-most produce
    // themselves. In order to compute RMPSELF, the map PRODUCES
    // must be constructed which identifies for each nonterminal
    // the set of nonterminals that it can right-most produce.
    //
    if (option -> lalr_level > 1)
    {
        direct_produces.Resize(grammar -> num_terminals + 1, grammar -> num_symbols);
        produces.Resize(grammar -> num_terminals + 1, grammar -> num_symbols);
        for (int nt = grammar -> FirstNonTerminal(); nt <= grammar -> LastNonTerminal(); nt++)
            produces[nt].Initialize(grammar -> num_nonterminals + 1, grammar -> num_terminals);

        for (int nt = grammar -> FirstNonTerminal(); nt <= grammar -> LastNonTerminal(); nt++)
        {
            for (int k = 0; k < clitems[nt].Length(); k++)
            {
                int item_no = clitems[nt][k],
                    symbol = item_table[item_no].symbol;
                if (grammar -> IsNonTerminal(symbol))
                {
                    int i = item_table[item_no].suffix_index;
                    if (first[i].set[grammar -> empty] && (! produces[nt][symbol]))
                    {
                        produces[nt].AddElement(symbol);
                        direct_produces[nt].Next() = symbol;
                    }
                }
            }
        }

        //
        // Complete the construction of the RIGHT_MOST_PRODUCES map
        // for non-terminals using the digraph algorithm.
        //
        index_of.Initialize(OMEGA);
        for (int nt = grammar -> FirstNonTerminal(); nt <= grammar -> LastNonTerminal(); nt++)
        {
            if (index_of[nt] == OMEGA)
                ComputeProduces(nt);
        }

        InitRmpself(produces);
    }

    //
    // Construct the FOLLOW map if
    //   1) an SLR table is requested
    //   2) if we have to print the FOLLOW map
    //   3) Error-maps are requested
    //   4) There are more than one starting symbol.
    //
    if (option -> slr || option -> follow || option -> error_maps ||
        next_rule[lhs_rule[grammar -> accept_image]] != lhs_rule[grammar -> accept_image])
    {
        follow.Resize(grammar -> num_terminals + 1, grammar -> num_symbols);
        for (int i = grammar -> FirstNonTerminal(); i <= grammar -> LastNonTerminal(); i++)
            follow[i].Initialize(grammar -> num_terminals + 1);
        follow[grammar -> accept_image].AddElement(grammar -> eof_image);


        index_of.Initialize(OMEGA);
        index_of[grammar -> accept_image] = INFINITY;  // mark computed
        for (int nt = grammar -> FirstNonTerminal(); nt <= grammar -> LastNonTerminal(); nt++)
        {
            if (index_of[nt] == OMEGA) // not yet computed ?
                ComputeFollow(nt);
        }

        //
        // Initialize FIRST for suffixes that can follow each starting
        // non-terminal ( except the main symbol) with the FOLLOW set
        // of the non-terminal in question.
        //
        int rule_no = lhs_rule[grammar -> accept_image];
        if (next_rule[rule_no] != rule_no)
        {
            rule_no = next_rule[rule_no];   // first rule
            int top = item_table[first_item_of[rule_no]].suffix_index;
            for (int i = top + 1; i < first.Length(); i++)
            {
                rule_no = next_rule[rule_no];

                int item_no = first_item_of[rule_no],
                    symbol = item_table[item_no].symbol;
                if (grammar -> IsNonTerminal(symbol))
                    first[i].set = follow[symbol];
            }
        }
    }

    //
    // If we are constructing a backtracking parser, we need to order
    // the rules based on the relation:
    //
    //    (A, B) is in the relation iff A ::=* B xxx
    //
    // Note that this implies that (A, B) if A ::= yyy B xxx if
    // yyy =>* empty.
    //
    if (option -> backtrack)
    {
        rank.Resize(grammar -> num_rules + 1);

        lhs_ranked_rules.Resize(grammar -> num_terminals + 1, grammar -> num_symbols);

        //
        // In this loop, we construct the LHS_RANKED_RULES map which maps
        // each non-terminal symbol into the set of rules it produces.
        // When we encounter a rule where the priority is explicitly set,
        // we process it right away.
        //
        for (int rule_no = grammar -> FirstRule(); rule_no <= grammar -> LastRule(); rule_no++)
        {
            if (option -> priority && grammar -> rules[rule_no].IsPriorityProduction())
                ordered_rules.Next() = rule_no;
            else
            {
                int symbol = grammar -> rules[rule_no].lhs;
                lhs_ranked_rules[symbol].Next() = rule_no;
            }
        }

        //
        // Now assign priorities to rules based on partial order of the
        // nonterminals (left-hand sides). See ComputeRank for the
        // relation on which the partial order is constructed.
        //
        index_of.Initialize(OMEGA);
        for (int nt = grammar -> FirstNonTerminal(); nt <= grammar -> LastNonTerminal(); nt++)
        {
            if (index_of[nt] == OMEGA)
                ComputeRank(nt);
        }

        for (int i = 0; i < ordered_rules.Length(); i++)
            rank[ordered_rules[i]] = i + 1;
        lhs_ranked_rules.Resize(0, -1); // free space

        //
        // Report any problem found in computing the rank of the rules.
        //
        if (option -> priority)
        {
            //
            // First, flush any data left in the report buffer.
            //
            option -> report.Print(Control::syslis);
            option -> report.Flush(stdout);

            if (SCC_in_ranks.Length() > 0)
            {
                fprintf(Control::syslis, "\n\nStrongly-connected components in ranks:\n\n");
                for (int i = 0; i < SCC_in_ranks.Length(); i++)
                {
                    for (int k = 0; k < SCC_in_ranks[i].Length(); k++)
                        fprintf(Control::syslis, "    %s\n", grammar -> RetrieveString(SCC_in_ranks[i][k]));
                    fprintf(Control::syslis, "\n");
                }
            }

            fprintf(Control::syslis, "\n\n" "RANK RULE\n" "==== ====\n");
            for (int i = 0; i < ordered_rules.Length(); i++)
            {
                int rule_no = ordered_rules[i];
                fprintf(Control::syslis, "%-4d %-4d  ", rank[rule_no], rule_no);
                fprintf(Control::syslis, "%s ::=", grammar -> RetrieveString(grammar -> rules[rule_no].lhs));
                for (int k = grammar -> FirstRhsIndex(rule_no); k < grammar -> EndRhsIndex(rule_no); k++)
                {
                    int symbol = grammar -> rhs_sym[k];
                    fprintf(Control::syslis, " %s", grammar -> RetrieveString(symbol));
                }
                fprintf(Control::syslis, "\n");
            }
        }

        SCC_in_ranks.Resize(0); // free space
    }

    //
    // If WARNINGS option is turned on, the unreachable symbols in
    // the grammar are printed.
    //
    if (option -> warnings)
        PrintUnreachables();

    //
    // If a Cross_Reference listing is requested, it is generated
    // here.
    //
    if (option -> xref)
        PrintXref();

    //
    // If a listing of the FIRST map is requested, it is generated
    // here.
    //
    if (option -> first)
        PrintNonTerminalFirst();

    //
    // If a listing of the FOLLOW map is requested, it is generated
    // here.
    //
    if (option -> follow)
        PrintFollowMap();

    option -> report.Print(Control::syslis);
    option -> report.Flush(stdout);

    return;
}


//
// DEPRECATED: TODO: Note that since the new input format does
// not require terminal symbols to be declared, all non-declared
// grammar symbols are assumed to be terminals. Therefore, it
// will never be the case that we detect a nonterminal that does not
// produce any rule.
//
void Base::NoRulesProduced(void)
{
    char line[Control::PRINT_LINE_SIZE + 1],
         tok[Control::SYMBOL_SIZE + 1];

    //
    // Build a list of all non-terminals that do not produce any
    // rules.
    //
    int nt_root = NIL,
        nt_last = NIL;
    BoundedArray<int> nt_list(grammar -> num_terminals + 1, grammar -> num_symbols);
    for (int symbol = grammar -> FirstNonTerminal(); symbol <= grammar -> LastNonTerminal(); symbol++)
    {
        if (lhs_rule[symbol] == NIL)
        {
            if (nt_root == NIL)
                 nt_root = symbol;
            else nt_list[nt_last] = symbol;
            nt_last = symbol;
        }
    }

    //
    // If the list of non-terminals that do not produce any rules
    // is not empty, signal error and stop.
    //
    if (nt_root != NIL)
    {
        nt_list[nt_last] = NIL;

        option -> report.Put("***ERROR: ");
        option -> report.Put(nt_list[nt_root] == NIL
                                ? "The following Non-terminal does not produce any rules: "
                                : "The following Non-terminals do not produce any rules: ");
        option -> report.PutChar('\n');
        strcpy(line, "        ");

        for (int symbol = nt_root; symbol != NIL; symbol = nt_list[symbol])
        {
            grammar -> RestoreSymbol(tok, grammar -> RetrieveString(symbol));
            if (strlen(line) + strlen(tok) > Control::PRINT_LINE_SIZE)
            {
                option -> report.Put(line);
                option -> report.PutChar('\n');
                grammar -> ReportLargeToken(line, tok, "    ", MAX_LENGTH);
            }
            else strcat(line, tok);
            strcat(line, " ");
        }

        option -> report.Put(line);
        option -> report.PutChar('\n');

        option -> report.Print(Control::syslis);
        option -> report.Flush(stdout);

        exit(12);
    }

    return;
}


//
//  This function computes the closure of a non-terminal LHS_SYMBOL passed
// to it as an argument using the digraph algorithm.
// The closure of a non-terminal A is the set of all non-terminals Bi that
// can directly or indirectly start a string generated by A.
// I.e., A *::= Bi X where X is an arbitrary string.
//
void Base::ComputeClosure(int lhs_symbol)
{
    BoundedArray<int> nont_list(grammar -> num_terminals + 1, grammar -> num_symbols, OMEGA);
    nont_list[lhs_symbol] = NIL;
    int nt_root = lhs_symbol;

    stack.Push(lhs_symbol);
    int indx = stack.Length();
    index_of[lhs_symbol] = indx;

    assert(closure[lhs_symbol].Length() == 0);

    int rule_no;
    for (bool end_node = ((rule_no = lhs_rule[lhs_symbol]) == NIL);
         ! end_node;          // Iterate over all rules of LHS_SYMBOL
         end_node = (rule_no == lhs_rule[lhs_symbol]))
    {
        rule_no = next_rule[rule_no];

        int symbol = (grammar -> RhsSize(rule_no) == 0
                               ? grammar -> empty
                               : grammar -> rhs_sym[grammar -> rules[rule_no].rhs_index]);

        if (grammar -> IsNonTerminal(symbol))
        {
            if (nont_list[symbol] == OMEGA)
            {
                if (index_of[symbol] == OMEGA) // if first time seen
                    ComputeClosure(symbol);

                index_of[lhs_symbol] = Util::Min(index_of[lhs_symbol],
                                                    index_of[symbol]);
                nont_list[symbol] = nt_root;
                nt_root = symbol;

                //
                // add closure[symbol] to closure of LHS_SYMBOL.
                //
                for (int k = 0; k < closure[symbol].Length(); k++)
                {
                    int element = closure[symbol][k];
                    if (nont_list[element] == OMEGA)
                    {
                        nont_list[element] = nt_root;
                        nt_root = element;
                    }
                }
            }
        }
    }

    assert(closure[lhs_symbol].Length() == 0);

    for (; nt_root != lhs_symbol; nt_root = nont_list[nt_root])
        closure[lhs_symbol].Next() = nt_root;

    if (index_of[lhs_symbol] == indx)
    {
        for (int symbol = stack.Top(); symbol != lhs_symbol; symbol = stack.Top())
        {
            closure[symbol].Next() = lhs_symbol;

            for (int k = 0; k < closure[lhs_symbol].Length(); k++)
            {
                int element = closure[lhs_symbol][k];
                if (element != symbol)
                    closure[symbol].Next() = element;
            }

            index_of[symbol] = INFINITY;
            stack.Pop();
        }

        index_of[lhs_symbol] = INFINITY;
        stack.Pop();
    }

    return;
}


//
//   This procedure computes the set of non-terminal symbols that can
// generate the empty string.  Such non-terminals are said to be nullable.
//
// A non-terminal "A" can generate empty if the grammar in question contains
// a rule:
//
//          A ::= B1 B2 ... Bn     n >= 0,  1 <= i <= n
// and Bi, for all i, is a nullable non-terminal.
//
void Base::NullablesComputation(void)
{
    //
    // First, initialize
    // RHS_START. RHS_START is a mapping from each rule in the grammar
    // into the next symbol in its right-hand side that has not yet
    // proven to be nullable.
    //
    Array<int> rhs_start(grammar -> num_rules + 1);
    for (int rule_no = grammar -> FirstRule(); rule_no <= grammar -> LastRule(); rule_no++)
        rhs_start[rule_no] = grammar -> rules[rule_no].rhs_index;

    //
    // We now iterate over the rules and try to advance the RHS_START
    // pointer thru each right-hand side as far as we can.  If one or
    // more non-terminals are found to be nullable, they are marked
    // as such and the process is repeated.
    //
    // If we go through all the rules and no new non-terminal is found
    // to be nullable then we stop and return.
    //
    // Note that for each iteration, only rules associated with
    // non-terminals that are non-nullable are considered.  Further,
    // as soon as a non-terminal is found to be nullable, the
    // remaining rules associated with it are not considered.  I.e.,
    // we quit the inner loop.
    //
    bool changed = true;
    while(changed)
    {
        changed = false;

        for (int nt = grammar -> FirstNonTerminal(); nt <= grammar -> LastNonTerminal(); nt++)
        {
            int rule_no;
            for (bool end_node = ((rule_no = lhs_rule[nt]) == NIL);
                 ! null_nt[nt] && ! end_node;
                 end_node = (rule_no == lhs_rule[nt]))
            {
                rule_no = next_rule[rule_no];

                if (IsNullableRhs(rhs_start, rule_no))
                {
                    changed = true;
                    null_nt[nt] = true;
                }
            }
        }
    }

    return;
}


//
// This procedure tries to advance the RHS_START pointer.  If the current
// symbol identified by the RHS_START element is a terminal it returns false
// to indicate that it cannot go any further.  If it encounters a  non-null-
// lable non-terminal, it also returns false. Otherwise, the whole right-hand
// side is consumed, and it returns the value true.
//
bool Base::IsNullableRhs(Array<int> &rhs_start, int rule_no)
{
    for(rhs_start[rule_no] = rhs_start[rule_no];
        rhs_start[rule_no] <= grammar -> rules[rule_no + 1].rhs_index - 1;
        rhs_start[rule_no]++)
    {
        int symbol = grammar -> rhs_sym[rhs_start[rule_no]];
        if (grammar -> IsTerminal(symbol))
             return(false);
        else if (! null_nt[symbol]) // symbol is a non-terminal
             return(false);
    }

    return(true);
}


//
// This subroutine computes FIRST(NT) for some non-terminal NT using the
// digraph algorithm.
// FIRST(NT) is the set of all terminals Ti that may start a string generated
// by NT. That is, NT *::= Ti X where X is an arbitrary string.
//
void Base::ComputeFirst(int nt)
{
    stack.Push(nt);
    int indx = stack.Length();
    index_of[nt] = indx;

    //
    // Iterate over all rules generated by non-terminal NT...
    // In this application of the transitive closure algorithm,
    //
    //  G(A) := { t | A ::= t X for a terminal t and a string X }
    //
    // The relation R is defined as follows:
    //
    //    R(A, B) iff A ::= B1 B2 ... Bk B X
    //
    // where Bi is nullable for 1 <= i <= k
    //
    int rule_no;
    for (bool end_node = ((rule_no = lhs_rule[nt]) == NIL);
         ! end_node;    // Iterate over all rules produced by NT
         end_node = (rule_no == lhs_rule[nt]))
    {
        rule_no = next_rule[rule_no];

        bool blocked = false;
        for (int i = grammar -> FirstRhsIndex(rule_no); i < grammar -> EndRhsIndex(rule_no); i++)
        {
            int symbol = grammar -> rhs_sym[i];
            if (grammar -> IsNonTerminal(symbol))
            {
                if (index_of[symbol] == OMEGA)
                    ComputeFirst(symbol);
                index_of[nt] = Util::Min( index_of[nt], index_of[symbol]);

                BitSet temp_set(nt_first[symbol]);
                temp_set.RemoveElement(grammar -> empty);
                nt_first[nt] += temp_set;
                blocked = ! (null_nt[symbol]);
            }
            else
            {
                nt_first[nt].AddElement(symbol);
                blocked = true;
            }

            if (blocked)
                break;
        }

        if (! blocked)
        {
            nt_first[nt].AddElement(grammar -> empty);
        }
    }

    if (index_of[nt] == indx)
    {
        for (int symbol = stack.Top(); symbol != nt; symbol = stack.Top())
        {
            nt_first[symbol] = nt_first[nt];
            index_of[symbol] = INFINITY;
            stack.Pop();
        }
        index_of[nt] = INFINITY;
        stack.Pop();
    }

    return;
}


//
// Compute rank for each rule if needed.
// TODO: This is a temporary algorithm. A much deeper
// analysis is required...
//
void Base::ComputeRank(int nt)
{
    stack.Push(nt);
    int indx = stack.Length();
    index_of[nt] = indx;

    //
    // Iterate over all rules generated by non-terminal NT...
    // In this application of the transitive closure algorithm,
    //
    //  G(A) := { t | A ::= t X for a terminal t and a string X }
    //
    // The relation R is defined as follows:
    //
    //    R(A, Bi) and R(A, B) iff A ::= B1 B2 ... Bk B X
    //
    // where Bi is nullable for 0 <= i <= k and B is not nullable
    //
    for (int k = 0; k < lhs_ranked_rules[nt].Length(); k++)
    {
        int rule_no = lhs_ranked_rules[nt][k];
        for (int i = grammar -> FirstRhsIndex(rule_no); i < grammar -> EndRhsIndex(rule_no); i++)
        {
            int symbol = grammar -> rhs_sym[i];
            if (grammar -> IsNonTerminal(symbol))
            {
                if (index_of[symbol] == OMEGA)
                    ComputeRank(symbol);
                index_of[nt] = Util::Min( index_of[nt], index_of[symbol]);

                if (! null_nt[symbol])
                    break;
            }
            else break;
        }
    }

    if (index_of[nt] == indx)
    {
        int symbol;
        if (nt == stack.Top())
        {
            symbol = stack.Top();
            ProcessRank(symbol);
            index_of[symbol] = INFINITY;
            stack.Pop();
        }
        else // a non-trivial SCC
        {
            Tuple<int> &scc = SCC_in_ranks.Next();
            do
            {
                symbol = stack.Top();
                scc.Next() = symbol;
                ProcessRank(symbol);
                index_of[symbol] = INFINITY;
                stack.Pop();
            } while (symbol != nt);
        }
    }

    return;
}

//
// Assign a rank to all rules whose left-hand side is "symbol" who
// were not explicitly prioritized.
//
void Base::ProcessRank(int symbol)
{
    //
    // List of empty rules associated with symbol.
    // Note that we only consider explicitly specified NULL
    // productions; nullable productions, i.e., a production
    // whose right-hand side consists of a sequence of nullable
    // nonterminals are not considered.
    //
    Tuple<int> empty_rules,
               nullable_rules;

    //
    // In this loop, as we iterate over the non-prioritized rules associated
    // with a nonterminal, we assign ranks to them in the order in which they
    // appear unless they happen to be empty or nullable rules in which case
    // we add them to separate respective lists.
    //
    for (int i = 0; i < lhs_ranked_rules[symbol].Length(); i++)
    {
        int rule_no = lhs_ranked_rules[symbol][i],
            first_item = first_item_of[rule_no],
            first_rhs_symbol = item_table[first_item].symbol;
        if (grammar -> RhsSize(rule_no) == 0)
             empty_rules.Next() = rule_no;
        else if (grammar -> IsNonTerminal(first_rhs_symbol) &&
                 null_nt[first_rhs_symbol] &&
                 First(item_table[first_item].suffix_index)[grammar -> empty])
             nullable_rules.Next() = rule_no;
        else ordered_rules.Next() = rule_no;
    }

    //
    // We now assign ranks to the nullable rules.
    //
    for (int i = 0; i < nullable_rules.Length(); i++)
        ordered_rules.Next() = nullable_rules[i];

    //
    // We now assign ranks to the empty rules.
    //
    for (int i = 0; i < empty_rules.Length(); i++)
        ordered_rules.Next() = empty_rules[i];

    return;
}


//
// This procedure checks whether or not any non-terminal symbols can fail to
// generate a string of terminals.
//
// A non-terminal "A" can generate a terminal string if the grammar in
// question contains a rule of the form:
//
//         A ::= X1 X2 ... Xn           n >= 0,  1 <= i <= n
//
// and Xi, for all i, is a terminal or a non-terminal that can generate a
// string of terminals.
// This routine is structurally identical to COMPUTE_NULLABLES.
//
void Base::CheckNonTerminals(void)
{
    BoundedArray<bool> produces_terminals(grammar -> num_terminals + 1, grammar -> num_symbols, false);
    produces_terminals[grammar -> accept_image] = true;

    //
    // First, assume that only accept produces terminals. Then,
    // initialize RHS_START. RHS_START is a mapping from each rule in
    // the grammar into the next symbol in its right-hand side that
    // has not yet proven to be a symbol that generates terminals.
    //
    Array<int> rhs_start(grammar -> num_rules + 1);
    for (int rule_no = grammar -> FirstRule(); rule_no <= grammar -> LastRule(); rule_no++)
        rhs_start[rule_no] = grammar -> rules[rule_no].rhs_index;

    //
    // We now iterate over the rules and try to advance the RHS_START
    // pointer to each right-hand side as far as we can.  If one or
    // more non-terminals are found to be "all right", they are
    // marked as such and the process is repeated.
    //
    // If we go through all the rules and no new non-terminal is
    // found to be "all right" then we stop and return.
    //
    // Note that on each iteration, only rules associated with
    // non-terminals that are not "all right" are considered. Further,
    // as soon as a non-terminal is found to be "all right", the
    // remaining rules associated with it are not considered. I.e.,
    // we quit the inner loop.
    //
    bool changed = true;
    while(changed)
    {
        changed = false;

        for (int nt = grammar -> FirstNonTerminal(); nt <= grammar -> LastNonTerminal(); nt++)
        {
            int rule_no;
            for (bool end_node = ((rule_no = lhs_rule[nt]) == NIL);
                 (! produces_terminals[nt]) && (! end_node);
                 end_node = (rule_no == lhs_rule[nt]))
            {
                rule_no = next_rule[rule_no];

                if (IsTerminalRhs(rhs_start, produces_terminals, rule_no))
                {
                    changed = true;
                    produces_terminals[nt] = true;
                }
            }
        }
    }

    //
    // Construct a list of all non-terminals that do not generate
    // terminal strings.
    //
    int nt_root = NIL,
        nt_last = NIL;
    BoundedArray<int> nt_list(grammar -> num_terminals + 1, grammar -> num_symbols);
    for (int nt = grammar -> FirstNonTerminal(); nt <= grammar -> LastNonTerminal(); nt++)
    {
        if (! produces_terminals[nt])
        {
            if (nt_root == NIL)
                nt_root = nt;
            else
                nt_list[nt_last] = nt;
            nt_last = nt;
        }
    }

    //
    // If there are non-terminal symbols that do not generate
    // terminal strings, print them out and stop the program.
    //
    if (nt_root != NIL)
    {
        nt_list[nt_last] = NIL;  // mark end of list

        char line[Control::PRINT_LINE_SIZE + 1];
        strcpy(line, "*** ERROR: The following Non-terminal");
        if (nt_list[nt_root] == NIL)
            strcat(line, " does not generate any terminal strings: ");
        else
        {
            strcat(line, "s do not generate any terminal strings: ");
            option -> report.Put(line);
            option -> report.PutChar('\n');
            strcpy(line, "        "); // 8 spaces
        }

        for (int symbol = nt_root; symbol != NIL; symbol = nt_list[symbol])
        {
            char tok[Control::SYMBOL_SIZE + 1];
            grammar -> RestoreSymbol(tok, grammar -> RetrieveString(symbol));
            if (strlen(line) + strlen(tok) > Control::PRINT_LINE_SIZE-1)
            {
                option -> report.Put(line);
                option -> report.PutChar('\n');
                grammar -> ReportLargeToken(line, tok, "    ", MAX_LENGTH);
            }
            else strcat(line, tok);
            strcat(line, " ");
        }
        option -> report.Put(line);
        option -> report.PutChar('\n');

        option -> report.Print(Control::syslis);
        option -> report.Flush(stdout);

        exit(12);
    }

    return;
}


//
// This procedure tries to advance the RHS_START pointer.  If the current
// symbol identified by the RHS_START element is a bad non-terminal it
// returns false.  Otherwise, the whole right-hand side is traversed, and it
// returns the value true.
//
bool Base::IsTerminalRhs(Array<int> &rhs_start, BoundedArray<bool> &produces_terminals, int rule_no)
{
    for(rhs_start[rule_no] = rhs_start[rule_no];
        rhs_start[rule_no] <= grammar -> rules[rule_no + 1].rhs_index - 1;
        rhs_start[rule_no]++)
    {
        int symbol = grammar -> rhs_sym[rhs_start[rule_no]];
        if (grammar -> IsNonTerminal(symbol))
        {
            if (! produces_terminals[symbol])
                return(false);
        }
    }

    return(true);
}


//
//  FIRST_MAP takes as arguments two pointers, ROOT and TAIL, to a sequence
// of symbols in RHS which it inserts in FIRST_TABLE.  The vector FIRST_TABLE
// is used as the base for a hashed table where collisions are resolved by
// links.  Elements added to this hash table are allocated from the vector
// FIRST, with the variable TOP always indicating the position of the
// last element in FIRST that was allocated.
// NOTE: The suffix indentified by ROOT and TAIL is presumed not to be empty.
//       That is, ROOT <= TAIL !!!
//
int Base::FirstMap(int root, int tail)
{
    for (int i = first_table[grammar -> rhs_sym[root]]; i != NIL; i = first[i].link)
    {
        int j,
            k;
        for (j = root + 1,
             k = first[i].suffix_root + 1;
             (j <= tail && k <= first[i].suffix_tail);
             j++, k++)
        {
            if (grammar -> rhs_sym[j] != grammar -> rhs_sym[k])
                break;
        }
        if (j > tail && k > first[i].suffix_tail)
            return i;
    }

    int index = first.NextIndex();
    first[index].suffix_root = root;
    first[index].suffix_tail = tail;
    first[index].link = first_table[grammar -> rhs_sym[root]];
    first_table[grammar -> rhs_sym[root]] = index;

    return(index);
}


//
// SuffixFirst computes the set of all terminals that can appear as the first symbol
// in a rule suffix  and places the result in the FIRST set indexable by INDEX.
//
void Base::SuffixFirst(int index)
{
    int root = first[index].suffix_root,
        tail = first[index].suffix_tail,
        symbol = (root > tail ? grammar -> empty : grammar -> rhs_sym[root]);

    if (grammar -> IsTerminal(symbol))
    {
        first[index].set.SetEmpty();
        first[index].set.AddElement(symbol); // add it to set
    }
    else first[index].set = nt_first[symbol];

    for (int i = root + 1; i <= tail && first[index].set[grammar -> empty]; i++)
    {
        symbol = grammar -> rhs_sym[i];
        first[index].set.RemoveElement(grammar -> empty); // remove EMPTY
        if (grammar -> IsTerminal(symbol))
             first[index].set.AddElement(symbol); // add it to set
        else first[index].set += nt_first[symbol];
    }

    return;
}


//
// For a given symbol, complete the computation of
// PRODUCES[symbol].
//
void Base::ComputeProduces(int symbol)
{
    stack.Push(symbol);
    int indx = stack.Length();
    index_of[symbol] = indx;

    for (int i = 0; i < direct_produces[symbol].Length(); i++)
    {
        int new_symbol = direct_produces[symbol][i];
        if (index_of[new_symbol] == OMEGA)  // first time seen?
            ComputeProduces(new_symbol);
        index_of[symbol] = Util::Min(index_of[symbol], index_of[new_symbol]);
        produces[symbol] += produces[new_symbol];
    }

    if (index_of[symbol] == indx)  //symbol is SCC root
    {
        for (int new_symbol = stack.Top();
             new_symbol != symbol; new_symbol = stack.Top())
        {
            produces[new_symbol] = produces[symbol];
            index_of[new_symbol] = INFINITY;
            stack.Pop();
        }

        index_of[symbol] = INFINITY;
        stack.Pop();
    }

    return;
}


//
// COMPUTE_FOLLOW computes FOLLOW[nt] for some non-terminal NT using the
// digraph algorithm.  FOLLOW[NT] is the set of all terminals Ti that
// may immediately follow a string X generated by NT. I.e., if NT *::= X
// then X Ti is a valid substring of a class of strings that may be
// recognized by the language.
//
void Base::ComputeFollow(int nt)
{
    //
    // FOLLOW[NT] was initialized to 0 for all non-terminals.
    //
    stack.Push(nt);
    int indx = stack.Length();
    index_of[nt] = indx;

    //
    // iterate over all items of NT
    //
    for (int item_no = nt_items[nt]; item_no != NIL; item_no = next_item[item_no])
    {
        BitSet temp_set(first[item_table[item_no].suffix_index].set);
        if (temp_set[grammar -> empty])
        {
            temp_set.RemoveElement(grammar -> empty);
            int rule_no = item_table[item_no].rule_number,
                lhs_symbol = grammar -> rules[rule_no].lhs;
            if (index_of[lhs_symbol] == OMEGA)
                ComputeFollow(lhs_symbol);
            follow[nt] += follow[lhs_symbol];
            index_of[nt] = Util::Min( index_of[nt], index_of[lhs_symbol]);
        }
        follow[nt] += temp_set;
    }

    if (index_of[nt] == indx)
    {
        for (int lhs_symbol = stack.Top(); lhs_symbol != nt; lhs_symbol = stack.Top())
        {
            follow[lhs_symbol] = follow[nt];
            index_of[lhs_symbol] = INFINITY;
            stack.Pop();
        }
        index_of[nt] = INFINITY;
        stack.Pop();
    }

    return;
}


void Base::PrintUnreachables(void)
{
    //
    // SYMBOL_LIST is used for two purposes:
    //  1) to mark symbols that are reachable from the Accepting
    //        non-terminal.
    //  2) to construct lists of symbols that are not reachable.
    //
    Array<int> symbol_list(grammar -> num_symbols + 1, Util::OMEGA);
    symbol_list[grammar -> eof_image] = NIL;
    symbol_list[grammar -> empty] = NIL;
    if (option -> error_maps)
        symbol_list[grammar -> error_image] = NIL;

    //
    // Initialize a list consisting only of the Accept non-terminal.
    // This list is a work pile of non-terminals to process as follows:
    // Each non-terminal in the front of the list is removed in turn and
    // 1) All terminal symbols in one of its right-hand sides are
    //     marked reachable.
    // 2) All non-terminals in one of its right-hand sides are placed
    //     in the the work pile of it had not been processed previously
    //
    int nt_root = grammar -> accept_image;
    symbol_list[nt_root] = NIL;
    for (int nt = nt_root; nt != NIL; nt = nt_root)
    {
        nt_root = symbol_list[nt];

        int rule_no;
        for (bool end_node = ((rule_no = lhs_rule[nt]) == NIL);
             ! end_node;
             end_node = (rule_no == lhs_rule[nt]))
        {
            rule_no = next_rule[rule_no];

            for (int i = grammar -> FirstRhsIndex(rule_no); i < grammar -> EndRhsIndex(rule_no); i++)
            {
                int symbol = grammar -> rhs_sym[i];
                if (grammar -> IsTerminal(symbol))
                    symbol_list[symbol] = NIL;
                else if (symbol_list[symbol] == OMEGA)
                {
                    symbol_list[symbol] = nt_root;
                    nt_root = symbol;
                }
            }
        }
    }

    //
    // We now iterate (backwards to keep things in order) over the
    // terminal symbols, and place each unreachable terminal in a
    // list. If the list is not empty, we signal that these symbols
    // are unused.
    //
    int t_root = NIL;
    for (int symbol = grammar -> LastTerminal(); symbol >= grammar -> FirstTerminal(); symbol--)
    {
        if (symbol_list[symbol] == OMEGA)
        {
            symbol_list[symbol] = t_root;
            t_root = symbol;
        }
    }

    char line[Control::PRINT_LINE_SIZE + 1],
         tok[Control::SYMBOL_SIZE + 1];
    if (t_root != NIL)
    {
        if (symbol_list[t_root] != NIL)
        {
            option -> report.Put("*** The following Terminals are useless: ");
            option -> report.PutChar('\n');
            strcpy(line, "        "); // 8 spaces
        }
        else
            strcpy(line, "*** The following Terminal is useless: ");

        for (int symbol = t_root; symbol != NIL; symbol = symbol_list[symbol])
        {
            grammar -> RestoreSymbol(tok, grammar -> RetrieveString(symbol));
            if (strlen(line) + strlen(tok) > Control::PRINT_LINE_SIZE)
            {
                option -> report.Put(line);
                option -> report.PutChar('\n');
                grammar -> ReportLargeToken(line, tok, "    ", MAX_LENGTH);
            }
            else
            {
                strcat(line, tok);
                strcat(line, " ");
            }
            strcat(line, " ");
        }
        option -> report.Put(line);
        option -> report.PutChar('\n');
    }


    //
    // We now iterate (backward to keep things in order) over the
    // non-terminals, and place each unreachable non-terminal in a
    // list.  If the list is not empty, we signal that these
    // symbols are unused.
    //
    nt_root = NIL;
    for (int symbol = grammar -> LastNonTerminal(); symbol >= grammar -> FirstNonTerminal(); symbol--)
    {
        if (symbol_list[symbol] == OMEGA)
        {
            symbol_list[symbol] = nt_root;
            nt_root = symbol;
        }
    }

    if (nt_root != NIL)
    {
        if (symbol_list[nt_root] != NIL)
        {
            option -> report.Put("*** The following Non-Terminals are useless: ");
            option -> report.PutChar('\n');
            strcpy(line, "        "); // 8 spaces
        }
        else strcpy(line, "*** The following Non-Terminal is useless: ");

        for (int symbol = nt_root; symbol != NIL; symbol = symbol_list[symbol])
        {
            grammar -> RestoreSymbol(tok, grammar -> RetrieveString(symbol));
            if (strlen(line) + strlen(tok) > Control::PRINT_LINE_SIZE)
            {
                option -> report.Put(line);
                option -> report.PutChar('\n');
                grammar -> PrintLargeToken(line, tok, "    ", MAX_LENGTH);
            }
            else
                strcat(line, tok);
            strcat(line, " ");
        }
        option -> report.Put(line);
        option -> report.PutChar('\n');
    }

    return;
}


//
// PRINT_XREF prints out the Cross-reference map. We build a map from each
// terminal into the set of items whose Dot-symbol (symbol immediately
// following the dot ) is the terminal in question.  Note that we iterate
// backwards over the rules to keep the rules associated with the items
// sorted, since they are inserted in STACK fashion in the lists:  Last-in,
// First out.
//
void Base::PrintXref(void)
{
    //
    // SORT_SYM is used to sort the symbols for cross_reference listing.
    //
    Array<int> sort_sym(grammar -> num_symbols + 1),
               t_items(grammar -> num_terminals + 1);
    t_items.Initialize(NIL);

    for (int rule_no = grammar -> LastRule(); rule_no >= grammar -> FirstRule(); rule_no--)
    {
        int item_no = first_item_of[rule_no];
        for (int i = grammar -> FirstRhsIndex(rule_no); i < grammar -> EndRhsIndex(rule_no); i++)
        {
            int symbol = grammar -> rhs_sym[i];
            if (grammar -> IsTerminal(symbol))
            {
                next_item[item_no] = t_items[symbol];
                t_items[symbol] = item_no;
            }
            item_no++;
        }
    }
    for (int i = 1; i <= grammar -> num_symbols; i++)
        sort_sym[i] = i;
    QuickSortSymbols(sort_sym, 1, grammar -> num_symbols);

    char line[Control::PRINT_LINE_SIZE + 1],
         tok[Control::SYMBOL_SIZE + 1];

    //
    // First, flush any data left in the report buffer.
    //
    option -> report.Print(Control::syslis);
    option -> report.Flush(stdout);

    control -> PrintHeading();
    fprintf(Control::syslis, "\n\nCross-reference table:\n");
    for (int i = 1; i <= grammar -> num_symbols; i++)
    {
        int symbol = sort_sym[i];
        if (symbol != grammar -> accept_image && symbol != grammar -> eof_image
                                   && symbol != grammar -> empty)
        {
            fprintf(Control::syslis, "\n");
            grammar -> RestoreSymbol(tok, grammar -> RetrieveString(symbol));
            grammar -> PrintLargeToken(line, tok, "", Control::PRINT_LINE_SIZE-7);
            strcat(line, "  ==>> ");
            int offset = strlen(line) - 1,
                item_no;
            if (grammar -> IsNonTerminal(symbol))
            {
                int rule_no;
                for (bool end_node = ((rule_no = lhs_rule[symbol]) == NIL);
                     ! end_node;
                     end_node = (rule_no == lhs_rule[symbol]))
                {
                    rule_no = next_rule[rule_no];

                    IntToString num(rule_no);
                    if (num.Length() + strlen(line) > Control::PRINT_LINE_SIZE)
                    {
                        fprintf(Control::syslis, "\n%s", line);
                        strcpy(line, " ");
                        for (int j = 1; j <= offset; j++)
                             strcat(line, " ");
                    }
                    strcat(line, num.String());
                    strcat(line, " ");
                }
                fprintf(Control::syslis, "\n%s", line);
                item_no = nt_items[symbol];
            }
            else
            {
                for (item_no = t_items[symbol]; item_no != NIL; item_no = next_item[item_no])
                {
                    int rule_no = item_table[item_no].rule_number;

                    IntToString num(rule_no);
                    if (num.Length() + strlen(line) > Control::PRINT_LINE_SIZE)
                    {
                        fprintf(Control::syslis, "\n%s", line);
                        strcpy(line, " ");
                        for (int j = 1; j <= offset; j++)
                             strcat(line, " ");
                    }
                    strcat(line, num.String());
                    strcat(line, " ");
                }
                fprintf(Control::syslis, "\n%s",line);
            }
        }
    }
    fprintf(Control::syslis, "\n\n");

    return;
}


//
// QUICK_SYM takes as arguments an array of pointers whose elements point to
// nodes and two integer arguments: L, H. L and H indicate respectively the
// lower and upper bound of a section in the array.
//
void Base::QuickSortSymbols(Array<int> &array, int low, int high)
{
    Stack<int> lostack,
               histack;

    lostack.Push(low);
    histack.Push(high);

    while(! lostack.IsEmpty())
    {
        int lower = lostack.Pop(),
            upper = histack.Pop();

        while(upper > lower)
        {
            //
            // Split the array section indicated by LOWER and UPPER
            // using ARRAY[LOWER] as the pivot.
            //
            int i = lower,
                pivot = array[lower];
            for (int j = lower + 1; j <= upper; j++)
            {
                if (strcmp(grammar -> RetrieveString(array[j]),
                           grammar -> RetrieveString(pivot)) < 0)
                {
                    int temp = array[++i];
                    array[i] = array[j];
                    array[j] = temp;
                }
            }

            array[lower] = array[i];
            array[i] = pivot;

            if ((i - lower) < (upper - i))
            {
                lostack.Push(i + 1);
                histack.Push(upper);
                upper = i - 1;
            }
            else
            {
                histack.Push(i - 1);
                lostack.Push(lower);
                lower = i + 1;
            }
        }
    }

    return;
}


//
// PRINT_NON_TERMINAL_FIRST prints the first set for each non-terminal.
//
void Base::PrintNonTerminalFirst(void)
{
    char line[Control::PRINT_LINE_SIZE + 1],
         tok[Control::SYMBOL_SIZE + 1];

    //
    // First, flush any data left in the report buffer.
    //
    option -> report.Print(Control::syslis);
    option -> report.Flush(stdout);

    control -> PrintHeading();
    fprintf(Control::syslis, "\nFirst map for non-terminals:\n\n");

    for (int nt = grammar -> FirstNonTerminal(); nt <= grammar -> LastNonTerminal(); nt++)
    {
        grammar -> RestoreSymbol(tok, grammar -> RetrieveString(nt));
        grammar -> PrintLargeToken(line, tok, "", Control::PRINT_LINE_SIZE - 7);
        strcat(line, "  ==>> ");
        for (int t = grammar -> FirstTerminal(); t <= grammar -> LastTerminal(); t++)
        {
            if (nt_first[nt][t])
            {
                grammar -> RestoreSymbol(tok, grammar -> RetrieveString(t));
                if (strlen(line) + strlen(tok) > Control::PRINT_LINE_SIZE - 1)
                {
                    fprintf(Control::syslis, "\n%s", line);
                    grammar -> PrintLargeToken(line, tok, "    ", MAX_LENGTH);
                }
                else
                    strcat(line, tok);
                strcat(line, " ");
            }
        }
        fprintf(Control::syslis, "\n%s\n", line);
    }

    return;
}


//
// PRINT_FOLLOW_MAP prints the follow map.
//
void Base::PrintFollowMap(void)
{
    //
    // First, flush any data left in the report buffer.
    //
    option -> report.Print(Control::syslis);
    option -> report.Flush(stdout);

    control -> PrintHeading();
    fprintf(Control::syslis, "\nFollow Map:\n\n");
    for (int nt = grammar -> FirstNonTerminal(); nt <= grammar -> LastNonTerminal(); nt++)
    {
        char line[Control::PRINT_LINE_SIZE + 1],
             tok[Control::SYMBOL_SIZE + 1];

        grammar -> RestoreSymbol(tok, grammar -> RetrieveString(nt));
        grammar -> PrintLargeToken(line, tok, "", Control::PRINT_LINE_SIZE-7);
        strcat(line, "  ==>> ");
        for (int t = grammar -> FirstTerminal(); t <= grammar -> LastTerminal(); t++)
        {
            if (follow[nt][t])
            {
                grammar -> RestoreSymbol(tok, grammar -> RetrieveString(t));
                if (strlen(line) + strlen(tok) > Control::PRINT_LINE_SIZE-2)
                {
                    fprintf(Control::syslis, "\n%s", line);
                    grammar -> PrintLargeToken(line, tok, "    ", MAX_LENGTH);
                }
                else
                    strcat(line, tok);
                strcat(line, " ");
            }
        }
        fprintf(Control::syslis, "\n%s\n", line);
    }
    return;
}


//
// This procedure is invoked when LALR_LEVEL > 1 to construct the
// RMPSELF set which identifies the nonterminals that can right-most
// produce themselves. It takes as argumen the map PRODUCES which
// identifies for each nonterminal the set of nonterminals that it can
// right-most produce.
//
void Base::InitRmpself(BoundedArray<BitSetWithOffset> &produces)
{
    rmpself.Resize(grammar -> num_terminals + 1, grammar -> num_symbols);

    //
    // Note that each element of the map produces is a boolean vector
    // that is indexable in the range 1..num_nonterminals. Since each
    // nonterminal is offset by the value num_terminals (to distinguish
    // it from the terminals),it must therefore be adjusted accordingly
    // when dereferencing an element in the range of the produces map.
    //
    for (int nt = grammar -> FirstNonTerminal(); nt <= grammar -> LastNonTerminal(); nt++)
        rmpself[nt] = produces[nt][nt];

    return;
}
