#include <sys/stat.h>
#include "control.h"
#include "scanner.h"
#include "parser.h"
#include "jikespg_sym.h"

int (*Scanner::scan_keyword[13]) (char *p1) =
{
    ScanKeyword0,
    ScanKeyword0,
    ScanKeyword0,
    ScanKeyword0,
    ScanKeyword4,
    ScanKeyword0,
    ScanKeyword6,
    ScanKeyword7,
    ScanKeyword8,
    ScanKeyword9,
    ScanKeyword10,
    ScanKeyword11,
    ScanKeyword12
};

void Scanner::Setup()
{
    //
    // If this assertion fails, the Token structure in stream.h must be redesigned !!!
    //
    assert(NUM_TERMINALS < 64);

    for (int c = 0; c < 256; c++)
        classify_token[c] = (IsSpace(c) ? &Scanner::ClassifyBadToken : &Scanner::ClassifySymbol);

    classify_token[(int) NULL_CHAR] = &Scanner::ClassifyEof;
    classify_token[(int) CTL_Z] = &Scanner::ClassifyEof;

    classify_token[(int) option -> escape] = &Scanner::ClassifyEscapedSymbol;
    classify_token[(int) option -> or_marker] = &Scanner::ClassifyOr;

    classify_token[(int) ':']  = &Scanner::ClassifyEquivalence;
    classify_token[(int) '-']  = &Scanner::ClassifyArrow;
    classify_token[(int) '\''] = &Scanner::ClassifySingleQuotedSymbol;
    classify_token[(int) '\"'] = &Scanner::ClassifyDoubleQuotedSymbol;
    classify_token[(int) '<']  = &Scanner::ClassifyLess;
}


//
// Scan the main file
//
void Scanner::Scan()
{
    lex_stream -> ResetBadTokens();
    lex_stream -> ResetTokenStream();

    InputFileSymbol *main_input_file = lex_stream -> FindOrInsertFile(option -> grm_file);
    if (! main_input_file)
    {
        option -> report.Put("***Error: Input file \"");
        option -> report.Put(option -> grm_file);
        option -> report.Put("\" could not be read");
        option -> report.Flush(stdout);
    }
    else
    {
        main_input_file -> Lock();
        main_input_file -> ReadInput();
        input_buffer = main_input_file -> Buffer();
        if (input_buffer == NULL)
        {
            option -> report.Put("***Error: Input file \"");
            option -> report.Put(option -> grm_file);
            option -> report.Put("\" could not be read");
            option -> report.Flush(stdout);
            return;
        }
        line_location = main_input_file -> LineLocationReference();
        line_location -> Next() = 0; // mark starting location of line # 0

        current_token_index = lex_stream -> GetNextToken(main_input_file, 0); // Get 0th token and set its location to 0!
        current_token = lex_stream -> GetTokenReference(current_token_index);
        current_token -> SetKind(0);

        cursor = input_buffer;

        //
        // Assign the null string symbol to the 0th token.
        //
        current_token -> SetSymbol(variable_table -> FindOrInsertName(cursor, 0));
        current_token -> SetEndLocation(0);

        //
        // Process all options if any that was specified by the user or set
        // defaults otherwise. First, we process the options that were hard-wired
        // in the input. Then, we process any option that was issued from the
        // command line.
        //
        ScanOptions();

        option -> ProcessCommandOptions(); // process options from command line

        char *main_cursor = cursor; // Remember where we are in the main file

        //
        // If we need to include a template, load its environment and
        // process its options.
        //
        InputFileSymbol *template_file = NULL;
        if (option -> template_name)
        {
            //
            // First, look in the template search directory path for the
            // template file name.
            //
            template_file = lex_stream -> FindOrInsertFile(option -> template_search_directory, option -> template_name);

            //
            // If we were not successful in the template search directory path,
            // look in the include search directory path.
            //
            if (template_file == NULL)
                template_file = lex_stream -> FindOrInsertFile(option -> include_search_directory, option -> template_name);

            //
            // If we could not open the template file, issue an error.
            //
            if (template_file == NULL)
                lex_stream -> AddBadToken(LexStream::NO_TEMPLATE, current_token_index);
            else
            {
                template_file -> Lock();
                template_file -> ReadInput();
                input_buffer = template_file -> Buffer();
                if (input_buffer == NULL)
                {
                    option -> report.Put("***Error: Input file \"");
                    option -> report.Put(option -> template_name);
                    option -> report.Put("\" could not be read");
                    option -> report.Flush(stdout);
                    return;
                }
                line_location = template_file -> LineLocationReference();
                line_location -> Next() = 0; // mark starting location of line # 0
                cursor = input_buffer;
                ScanOptions(); // scan the template options
            }
        }

        option -> CompleteOptionProcessing();
        if (option -> return_code != 0) // if any bad option was found, stop
        {
            option -> report.Flush(stdout);
            exit(option -> return_code);
        }
        option -> PrintOptionsInEffect();

        Setup();

        //
        // If a template is present, continue scanning it...
        //
        if (template_file)
        {
            input_file = template_file;
            Scan(cursor, &(input_buffer[template_file -> BufferLength()]));
            while(lex_stream -> Kind(current_token_index) == TK_EOF)
                current_token_index--;
            lex_stream -> ResetTokenStream(current_token_index + 1);
            template_file -> Unlock(); // unlock the template file.
        }

        //
        // Restore the main file environment and continue scanning 
        //
        input_file = main_input_file;
        line_location = main_input_file -> LineLocationReference();
        input_buffer = main_input_file -> Buffer();
        Scan(main_cursor, &(input_buffer[input_file -> BufferLength()]));

        for (int k = 0; k < this -> option -> import_file.Length(); k++)
            ImportTerminals(option -> import_file[k]);

        for (int k = 0; k < this -> option -> filter_file.Length(); k++)
            ProcessFilters(option -> filter_file[k]);
    }

    //
    // If any bad tokens were scanned, flag them.
    //
    if (lex_stream -> NumBadTokens() > 0)
        lex_stream -> ReportErrors();

    return;
}


//
// Scan the file specified
//
void Scanner::Scan(int file_index)
{
    lex_stream -> ResetBadTokens();
    Setup();

    int error_code = IncludeFile(lex_stream -> NameString(file_index));
    if (error_code != LexStream::NO_ERROR)
        lex_stream -> AddBadToken(error_code, file_index);

    //
    // If any bad tokens were scanned, flag them.
    //
    if (lex_stream -> NumBadTokens() > 0)
         lex_stream -> ReportErrors();
    else
    {
        current_token_index = lex_stream -> GetNextToken(input_file, cursor - input_buffer);
        current_token = lex_stream -> GetTokenReference(current_token_index);
        current_token -> SetKind(TK_EOF);
    }

    return;
}

//
// This is one of the main entry point for the Java lexical analyser.
// Its input is the name of a regular text file. Its output is a stream
// of tokens.
//
void Scanner::Scan(char *buffer_start, char *buffer_tail)
{
    cursor = buffer_start;

    //
    // CURSOR is assumed to point to the next character to be scanned.
    // Using CURSOR,we jump to the proper classification function
    // which scans and classifies the token and returns the location of
    // the character immediately following it.
    //
    do
    {
        SkipSpaces();

        //
        // Allocate space for next token and set its location.
        //
        current_token_index = lex_stream -> GetNextToken(input_file, cursor - input_buffer);
        current_token = lex_stream -> GetTokenReference(current_token_index);
        Tuple<BlockSymbol *> &blocks = action_blocks -> Blocks((unsigned char) *cursor);
        if (blocks.Length())
             ClassifyBlock(blocks);
        else (this ->* classify_token[(unsigned char) *cursor])();
    } while (cursor < buffer_tail);

    //
    // Add a gate after the last line.
    //
    line_location -> Next() = buffer_tail - input_buffer;

    return;
}


//
//
//
void Scanner::ScanOptions()
{
    for (SkipSpaces(); strxeq(cursor, "%options"); SkipSpaces())
    {
        char *start = cursor + strlen("%options"),
             *ptr;
        if (! (IsSpace(*start) || (start[0] == '-' && start[1] == '-')))
            break;
        //
        // Note that we start scanning at "start" in case start points to a
        // new line character or a comment.
        //
        for (ptr = start; ! (IsNewline(*ptr) || (ptr[0] == '-' && ptr[1] == '-')); ptr++)  // skip all until \n
            ;
        option -> ProcessUserOptions(start, ptr - start);

        cursor = ptr;
    }

    return;
}


//
//
//
void Scanner::SkipOptions()
{
    for (SkipSpaces(); strxeq(cursor, "%options"); SkipSpaces())
    {
        char *start = cursor + strlen("%options"),
             *ptr;
        if (! (IsSpace(*start) || (start[0] == '-' && start[1] == '-')))
            break;
        //
        // Note that we start scanning at "start" in case start points to a
        // new line character or a comment.
        //
        for (ptr = start; ! (IsNewline(*ptr) || (ptr[0] == '-' && ptr[1] == '-')); ptr++)  // skip all until \n
            ;
        cursor = ptr;
    }

    return;
}


//
//
//
void Scanner::ImportTerminals(const char *filename)
{
    InputFileSymbol *file_symbol = this -> lex_stream -> FindOrInsertFile(option -> include_search_directory, filename);
    const char *name = (file_symbol == NULL ? filename : file_symbol -> Name());
    ImportArguments arguments(this -> option, name);
    Option option(arguments.argc, arguments.argv);
    LexStream lex_stream(&option);
    VariableLookupTable variable_table;
    MacroLookupTable macro_table;
    Scanner scanner(&option, &lex_stream, &variable_table, &macro_table);
    scanner.Scan();

    if (lex_stream.NumTokens() == 0 || lex_stream.NumBadTokens() > 0)
        exit(12);
    else
    {
        //
        // This is the code one would write in order to retrieve only the
        // terminals exported from the imported file without processing the
        // whole grammar.
        //
        //     Parser parser(&option, &lex_stream, &variable_table, &macro_table);
        //     parser.Parse();
        //
        //     //
        //     // Transfer each imported terminal symbol into the local variable
        //     // table and add it to the list of imported symbols.
        //     //
        //     for (int i = 0; i < parser.export.Length(); i++)
        //     {
        //     Tuple<int> imports;
        //     for (int i = 0; i < lex_stream.NumImportedFilters(); i++)
        //         imports.Next() = lex_stream.ImportedFilter(i);
        //     for (int i = 0; i < parser.exports.Length(); i++)
        //         imports.Next() = parser.exports[i];
        //
        //     for (int i = 0; i < imports.Length(); i++)
        //     {
        //         int import = imports[i];
        //
        //         VariableSymbol *symbol = lex_stream.GetVariableSymbol(import);
        //         InputFileSymbol *file_symbol = this -> lex_stream -> FindOrInsertFile(lex_stream.FileName(import));
        //
        //         current_token_index = this -> lex_stream -> GetNextToken(file_symbol, lex_stream.StartLocation(import));
        //         current_token = this -> lex_stream -> GetTokenReference(current_token_index);
        //         current_token -> SetKind(lex_stream.Kind(import));
        //         current_token -> SetSymbol(this -> variable_table -> FindOrInsertName(symbol -> Name(), symbol -> NameLength()));
        //
        //         this -> lex_stream -> AddImportedTerminal(current_token_index);
        //     }
        //  
    
        //
        // Process the imported grammar.
        //
        Control control(&option, &lex_stream, &variable_table, &macro_table);
        control.grammar -> Process();
        control.ConstructParser();

        //
        // Transfer each exported terminal symbol into the local variable
        // table and add it to the list of imported terminals.
        //
        Tuple<int> imports;
        for (int i = 0; i < lex_stream.NumImportedFilters(); i++)
            imports.Next() = lex_stream.ImportedFilter(i);
        for (int i = 0; i < control.grammar -> parser.exports.Length(); i++)
            imports.Next() = control.grammar -> parser.exports[i];
        for (int i = 0; i < imports.Length(); i++)
        {
            int import = imports[i];

            VariableSymbol *symbol = lex_stream.GetVariableSymbol(import);
            InputFileSymbol *old_file_symbol = lex_stream.GetFileSymbol(import),
                            *file_symbol = this -> lex_stream -> FindOrInsertFile(old_file_symbol -> Name());
            if (file_symbol -> LineLocation().Length() != old_file_symbol -> LineLocation().Length())
            {
                assert(file_symbol -> LineLocation().Length() == 0);
                for (int i = 0; i < old_file_symbol -> LineLocation().Length(); i++)
                    file_symbol -> LineLocation().Next() = old_file_symbol -> LineLocation()[i];
            }
            current_token_index = this -> lex_stream -> GetNextToken(file_symbol, lex_stream.StartLocation(import));
            current_token = this -> lex_stream -> GetTokenReference(current_token_index);
            current_token -> SetEndLocation(lex_stream.EndLocation(import));
            current_token -> SetKind(lex_stream.Kind(import));
            current_token -> SetSymbol(this -> variable_table -> FindOrInsertName(symbol -> Name(), symbol -> NameLength()));

            this -> lex_stream -> AddImportedTerminal(current_token_index);
        }
    }
    
    return;
}


//
//
//
void Scanner::ProcessFilters(const char *filename)
{
    InputFileSymbol *file_symbol = this -> lex_stream -> FindOrInsertFile(option -> include_search_directory, filename);
    const char *name = (file_symbol == NULL ? filename : file_symbol -> Name());
    FilterArguments arguments(this -> option, name);
    Option option(arguments.argc, arguments.argv);

    LexStream lex_stream(&option);
    VariableLookupTable variable_table;
    MacroLookupTable macro_table;
    Scanner scanner(&option, &lex_stream, &variable_table, &macro_table);
    scanner.Scan();

    if (lex_stream.NumTokens() == 0 || lex_stream.NumBadTokens() > 0)
        exit(12);
    else
    {
        this -> lex_stream -> AddFilterMacro(filename, (option.DefaultActionFile()
                                                              ? option.DefaultActionFile() -> Name()
                                                              : option.file_prefix));

        //
        // This is the code one would write in order to retrieve only the
        // terminals exported from the imported file without processing the
        // whole grammar.
        //
        //     Parser parser(&option, &lex_stream, &variable_table, &macro_table);
        //     parser.Parse();
        //
        //     //
        //     // Transfer each imported terminal symbol into the local variable
        //     // table and add it to the list of imported symbols.
        //     //
        //     for (int i = 0; i < parser.export.Length(); i++)
        //     {
        //         int import = parser.exports[i];
        // 
        //         VariableSymbol *symbol = lex_stream.GetVariableSymbol(import);
        //         InputFileSymbol *file_symbol = this -> lex_stream -> FindOrInsertFile(lex_stream.FileName(import));
        // 
        //         current_token_index = this -> lex_stream -> GetNextToken(file_symbol, lex_stream.StartLocation(import));
        //         current_token = this -> lex_stream -> GetTokenReference(current_token_index);
        //         current_token -> SetKind(lex_stream.Kind(import));
        //         current_token -> SetSymbol(this -> variable_table -> FindOrInsertName(symbol -> Name(), symbol -> NameLength()));
        // 
        //         this -> lex_stream -> AddImportedFilter(current_token_index);
        //     }
        //  

        //
        // Process the imported grammar.
        //
        Control control(&option, &lex_stream, &variable_table, &macro_table);
        control.grammar -> Process();
        control.ConstructParser();
    
        //
        // Transfer each exported terminal symbol into the local variable
        // table and add it to the list of imported filters.
        //
        Tuple<int> imports;
        for (int i = 0; i < lex_stream.NumImportedFilters(); i++)
            imports.Next() = lex_stream.ImportedFilter(i);
        for (int i = 0; i < control.grammar -> parser.exports.Length(); i++)
            imports.Next() = control.grammar -> parser.exports[i];
        for (int i = 0; i < imports.Length(); i++)
        {
            int import = imports[i];

            VariableSymbol *symbol = lex_stream.GetVariableSymbol(import);
            InputFileSymbol *old_file_symbol = lex_stream.GetFileSymbol(import),
                            *file_symbol = this -> lex_stream -> FindOrInsertFile(old_file_symbol -> Name());
            if (file_symbol -> LineLocation().Length() != old_file_symbol -> LineLocation().Length())
            {
                assert(file_symbol -> LineLocation().Length() == 0);
                for (int i = 0; i < old_file_symbol -> LineLocation().Length(); i++)
                    file_symbol -> LineLocation().Next() = old_file_symbol -> LineLocation()[i];
            }
            current_token_index = this -> lex_stream -> GetNextToken(file_symbol, lex_stream.StartLocation(import));
            current_token = this -> lex_stream -> GetTokenReference(current_token_index);
            current_token -> SetEndLocation(lex_stream.EndLocation(import));
            current_token -> SetKind(lex_stream.Kind(import));
            current_token -> SetSymbol(this -> variable_table -> FindOrInsertName(symbol -> Name(), symbol -> NameLength()));

            this -> lex_stream -> AddImportedFilter(current_token_index);
        }
    }

    return;
}


//
//
//
void Scanner::ScanComment()
{
    for (cursor += 2; ! IsNewline(*cursor); cursor++)  // skip all until \n
        ;
    return;
}


//
// This procedure is invoked to skip useless spaces in the input.
// It assumes upon entry that CURSOR points to the next character to
// be scanned.  Before returning it sets CURSOR to the location of the
// first non-space character following its initial position.
//
void Scanner::SkipSpaces()
{
    do
    {
        while (IsSpaceButNotNewline(*cursor))
            cursor++;
        while (IsNewline(*cursor)) // starting a new line?
        {
            if (*cursor == '\n') // For the sequence \r\n only count \n.
                line_location -> Next() = cursor + 1 - input_buffer;
            cursor++;
            while (IsSpaceButNotNewline(*cursor))
                cursor++;
        }

        while (cursor[0] == '-' && cursor[1] == '-')
              ScanComment();
    } while (IsSpace(*cursor));

    return;
}


//
// Scan a symbol of length I and determine if it is a keyword.
//
int Scanner::ScanKeyword0(char *p1)
{
    return TK_MACRO_NAME;
}

int Scanner::ScanKeyword4(char *p1)
{
    if (p1[1] == 'e' || p1[1] == 'E')
    {
        if (p1[2] == 'o' || p1[2] == 'O')
        {
             if (p1[3] == 'l' || p1[3] == 'L')
                  return TK_EOL_KEY;
             else if (p1[3] == 'f' || p1[3] == 'F')
                  return TK_EOF_KEY;
        }
        else if ((p1[2] == 'n' || p1[2] == 'N') &&
                 (p1[3] == 'd' || p1[3] == 'D'))
             return TK_END_KEY;
    }
    else if ((p1[1] == 'a' || p1[1] == 'A') &&
             (p1[2] == 's' || p1[2] == 'S') &&
             (p1[3] == 't' || p1[3] == 'T'))
         return TK_AST_KEY;

    return TK_MACRO_NAME;
}

int Scanner::ScanKeyword6(char *p1)
{
    switch (p1[1])
    {
        case 'a': case 'A':
            if ((p1[2] == 'l' || p1[2] == 'L') &&
                (p1[3] == 'i' || p1[3] == 'I') &&
                (p1[4] == 'a' || p1[4] == 'A') &&
                (p1[5] == 's' || p1[5] == 'S'))
                return TK_ALIAS_KEY;
            break;
        case 'e': case 'E':
            if ((p1[2] == 'm' || p1[2] == 'M') &&
                (p1[3] == 'p' || p1[3] == 'P') &&
                (p1[4] == 't' || p1[4] == 'T') &&
                (p1[5] == 'y' || p1[5] == 'Y'))
                return TK_EMPTY_KEY;
            else if ((p1[2] == 'r' || p1[2] == 'R') &&
                     (p1[3] == 'r' || p1[3] == 'R') &&
                     (p1[4] == 'o' || p1[4] == 'O') &&
                     (p1[5] == 'r' || p1[5] == 'R'))
                return TK_ERROR_KEY;
            break;
        case 'n': case 'N':
            if ((p1[2] == 'a' || p1[2] == 'A') &&
                (p1[3] == 'm' || p1[3] == 'M') &&
                (p1[4] == 'e' || p1[4] == 'E') &&
                (p1[5] == 's' || p1[5] == 'S'))
                return TK_NAMES_KEY;
            break;
        case 'r': case 'R':
            if ((p1[2] == 'u' || p1[2] == 'U') &&
                (p1[3] == 'l' || p1[3] == 'L') &&
                (p1[4] == 'e' || p1[4] == 'E') &&
                (p1[5] == 's' || p1[5] == 'S'))
                return TK_RULES_KEY;
            break;
        case 's': case 'S':
            if ((p1[2] == 't' || p1[2] == 'T') &&
                (p1[3] == 'a' || p1[3] == 'A') &&
                (p1[4] == 'r' || p1[4] == 'R') &&
                (p1[5] == 't' || p1[5] == 'T'))
                return TK_START_KEY;
            break;
        case 't': case 'T':
            if ((p1[2] == 'y' || p1[2] == 'Y') &&
                (p1[3] == 'p' || p1[3] == 'P') &&
                (p1[4] == 'e' || p1[4] == 'E') &&
                (p1[5] == 's' || p1[5] == 'S'))
                return TK_TYPES_KEY;
            break;
    }

    return TK_MACRO_NAME;
}

int Scanner::ScanKeyword7(char *p1)
{
    if ((p1[1] == 'd' || p1[1] == 'D') &&
        (p1[2] == 'e' || p1[2] == 'E') &&
        (p1[3] == 'f' || p1[3] == 'F') &&
        (p1[4] == 'i' || p1[4] == 'I') &&
        (p1[5] == 'n' || p1[5] == 'N') &&
        (p1[6] == 'e' || p1[6] == 'E'))
        return TK_DEFINE_KEY;
    else if ((p1[1] == 'e' || p1[1] == 'E') &&
             (p1[2] == 'x' || p1[2] == 'X') &&
             (p1[3] == 'p' || p1[3] == 'P') &&
             (p1[4] == 'o' || p1[4] == 'O') &&
             (p1[5] == 'r' || p1[5] == 'R') &&
             (p1[6] == 't' || p1[6] == 'T'))
        return TK_EXPORT_KEY;
    else if ((p1[1] == 'i' || p1[1] == 'I') &&
             (p1[2] == 'm' || p1[2] == 'M') &&
             (p1[3] == 'p' || p1[3] == 'P') &&
             (p1[4] == 'o' || p1[4] == 'O') &&
             (p1[5] == 'r' || p1[5] == 'R') &&
             (p1[6] == 't' || p1[6] == 'T'))
        return TK_IMPORT_KEY;
    else if ((p1[1] == 'n' || p1[1] == 'N') &&
             (p1[2] == 'o' || p1[2] == 'O') &&
             (p1[3] == 't' || p1[3] == 'T') &&
             (p1[4] == 'i' || p1[4] == 'I') &&
             (p1[5] == 'c' || p1[5] == 'C') &&
             (p1[6] == 'e' || p1[6] == 'E'))
        return TK_NOTICE_KEY;

    return TK_MACRO_NAME;
}

int Scanner::ScanKeyword8(char *p1)
{
    if((p1[1] == 'g' || p1[1] == 'G') &&
       (p1[2] == 'l' || p1[2] == 'L') &&
       (p1[3] == 'o' || p1[3] == 'O') &&
       (p1[4] == 'b' || p1[4] == 'B') &&
       (p1[5] == 'a' || p1[5] == 'A') &&
       (p1[6] == 'l' || p1[6] == 'L') &&
       (p1[7] == 's' || p1[7] == 'S'))
        return TK_GLOBALS_KEY;
    else if((p1[1] == 'h' || p1[1] == 'H') &&
            (p1[2] == 'e' || p1[2] == 'E') &&
            (p1[3] == 'a' || p1[3] == 'A') &&
            (p1[4] == 'd' || p1[4] == 'D') &&
            (p1[5] == 'e' || p1[5] == 'E') &&
            (p1[6] == 'r' || p1[6] == 'R') &&
            (p1[7] == 's' || p1[7] == 'S'))
        return TK_HEADERS_KEY;
    else if((p1[1] == 'i' || p1[1] == 'I') &&
            (p1[2] == 'n' || p1[2] == 'N') &&
            (p1[3] == 'c' || p1[3] == 'C') &&
            (p1[4] == 'l' || p1[4] == 'L') &&
            (p1[5] == 'u' || p1[5] == 'U') &&
            (p1[6] == 'd' || p1[6] == 'D') &&
            (p1[7] == 'e' || p1[7] == 'E'))
        return TK_INCLUDE_KEY;
    else if((p1[1] == 'r' || p1[1] == 'R') &&
            (p1[2] == 'e' || p1[2] == 'E') &&
            (p1[3] == 'c' || p1[3] == 'C') &&
            (p1[4] == 'o' || p1[4] == 'O') &&
            (p1[5] == 'v' || p1[5] == 'V') &&
            (p1[6] == 'e' || p1[6] == 'E') &&
            (p1[7] == 'r' || p1[7] == 'R'))
        return TK_RECOVER_KEY;

    return TK_MACRO_NAME;
}

int Scanner::ScanKeyword9(char *p1)
{
    if((p1[1] == 'k' || p1[1] == 'K') &&
       (p1[2] == 'e' || p1[2] == 'E') &&
       (p1[3] == 'y' || p1[3] == 'Y') &&
       (p1[4] == 'w' || p1[4] == 'W') &&
       (p1[5] == 'o' || p1[5] == 'O') &&
       (p1[6] == 'r' || p1[6] == 'R') &&
       (p1[7] == 'd' || p1[7] == 'D') &&
       (p1[8] == 's' || p1[8] == 'S'))
        return TK_KEYWORDS_KEY;
    else if((p1[1] == 't' || p1[1] == 'T') &&
            (p1[2] == 'r' || p1[2] == 'R') &&
            (p1[3] == 'a' || p1[3] == 'A') &&
            (p1[4] == 'i' || p1[4] == 'I') &&
            (p1[5] == 'l' || p1[5] == 'L') &&
            (p1[6] == 'e' || p1[6] == 'E') &&
            (p1[7] == 'r' || p1[7] == 'R') &&
            (p1[8] == 's' || p1[8] == 'S'))
        return TK_TRAILERS_KEY;

    return TK_MACRO_NAME;
}

int Scanner::ScanKeyword10(char *p1)
{
    if((p1[1] == 'd' || p1[1] == 'D') &&
       (p1[2] == 'r' || p1[2] == 'R') &&
       (p1[3] == 'o' || p1[3] == 'O') &&
       (p1[4] == 'p' || p1[4] == 'P') &&
       (p1[5] == 'r' || p1[5] == 'R') &&
       (p1[6] == 'u' || p1[6] == 'U') &&
       (p1[7] == 'l' || p1[7] == 'L') &&
       (p1[8] == 'e' || p1[8] == 'E') &&
       (p1[9] == 's' || p1[9] == 'S'))
        return TK_DROPRULES_KEY;
    else if((p1[1] == 't' || p1[1] == 'T') &&
       (p1[2] == 'e' || p1[2] == 'E') &&
       (p1[3] == 'r' || p1[3] == 'R') &&
       (p1[4] == 'm' || p1[4] == 'M') &&
       (p1[5] == 'i' || p1[5] == 'I') &&
       (p1[6] == 'n' || p1[6] == 'N') &&
       (p1[7] == 'a' || p1[7] == 'A') &&
       (p1[8] == 'l' || p1[8] == 'L') &&
       (p1[9] == 's' || p1[9] == 'S'))
        return TK_TERMINALS_KEY;

    return TK_MACRO_NAME;
}

int Scanner::ScanKeyword11(char *p1)
{
    if((p1[1] == 'i' || p1[1] == 'I') &&
       (p1[2] == 'd' || p1[2] == 'D') &&
       (p1[3] == 'e' || p1[3] == 'E') &&
       (p1[4] == 'n' || p1[4] == 'N') &&
       (p1[5] == 't' || p1[5] == 'T') &&
       (p1[6] == 'i' || p1[6] == 'I') &&
       (p1[7] == 'f' || p1[7] == 'F') &&
       (p1[8] == 'i' || p1[8] == 'I') &&
       (p1[9] == 'e' || p1[9] == 'E') &&
       (p1[10] == 'r' || p1[10] == 'R'))
        return TK_IDENTIFIER_KEY;

    return TK_MACRO_NAME;
}


int Scanner::ScanKeyword12(char *p1)
{
    if((p1[1] == 'd' || p1[1] == 'D') &&
       (p1[2] == 'r' || p1[2] == 'R') &&
       (p1[3] == 'o' || p1[3] == 'O') &&
       (p1[4] == 'p' || p1[4] == 'P') &&
       (p1[5] == 's' || p1[5] == 'S') &&
       (p1[6] == 'y' || p1[6] == 'Y') &&
       (p1[7] == 'm' || p1[7] == 'M') &&
       (p1[8] == 'b' || p1[8] == 'B') &&
       (p1[9] == 'o' || p1[9] == 'O') &&
       (p1[10] == 'l' || p1[10] == 'L') &&
       (p1[11] == 's' || p1[11] == 'S'))
        return TK_DROPSYMBOLS_KEY;
    else if((p1[1] == 'd' || p1[1] == 'D') &&
       (p1[2] == 'r' || p1[2] == 'R') &&
       (p1[3] == 'o' || p1[3] == 'O') &&
       (p1[4] == 'p' || p1[4] == 'P') &&
       (p1[5] == 'a' || p1[5] == 'A') &&
       (p1[6] == 'c' || p1[6] == 'C') &&
       (p1[7] == 't' || p1[7] == 'T') &&
       (p1[8] == 'i' || p1[8] == 'I') &&
       (p1[9] == 'o' || p1[9] == 'O') &&
       (p1[10] == 'n' || p1[10] == 'N') &&
       (p1[11] == 's' || p1[11] == 'S'))
        return TK_DROPACTIONS_KEY;

    return TK_MACRO_NAME;
}


//
//
//
void Scanner::ClassifyBlock(Tuple<BlockSymbol *> &blocks)
{
    for (int i = 0; i < blocks.Length(); i++)
    {
        BlockSymbol *block = blocks[i];
        if (strncmp(cursor, block -> BlockBegin(), block -> BlockBeginLength()) == 0)
        {
            char *block_end = block -> BlockEnd();
            int length = block -> BlockEndLength();

            for (cursor = cursor + length; *cursor != NULL_CHAR && (! strxeq(cursor, block_end)); cursor++)
            {
                // if (IsNewline(*cursor))
                if (*cursor == '\n') // when we encounter the sequence \r\n, we only count \n
                    line_location -> Next() = cursor - input_buffer + 1; // character after newline.
            }

            if (*cursor == NULL_CHAR)
                 lex_stream -> AddBadToken(LexStream::UNTERMINATED_BLOCK, current_token_index);
            else cursor = cursor + length;

            current_token -> SetKind(TK_BLOCK);
            current_token -> SetSymbol(block);
            current_token -> SetEndLocation((cursor - 1) - input_buffer);

            return;
        }
    }

    ClassifySymbol();

    return;
}


//
//
//
void Scanner::ClassifyEscapedSymbol()
{
    char *ptr = cursor + 1;

    while (IsAlnum(*ptr) && *ptr != option -> escape)
        ptr++;
    int len = ptr - cursor;

    current_token -> SetKind((scan_keyword[len < 13 ? len : 0])(cursor));
    current_token -> SetEndLocation((ptr - 1) - input_buffer);
    if (current_token -> Kind() == TK_MACRO_NAME)
        current_token -> SetSymbol(macro_table -> FindOrInsertName(cursor, len));
    else if (current_token -> Kind() == TK_INCLUDE_KEY)
    {
        //
        // . Save position of INCLUDE keyword token 
        // . Scan the include section
        // . Include the files specified in the include section
        //
        int include_key_index = current_token_index;
        cursor = ptr;
        SkipSpaces();
        if (*cursor != option -> escape)
        {
            current_token_index = lex_stream -> GetNextToken(input_file, cursor - input_buffer);
            current_token = lex_stream -> GetTokenReference(current_token_index);
            Tuple<BlockSymbol *> &blocks = action_blocks -> Blocks((unsigned char) *cursor);
            if (blocks.Length())
                 ClassifyBlock(blocks);
            else (this ->* classify_token[(unsigned char) *cursor])();
            SkipSpaces();
            ptr = cursor;
            if (current_token -> Kind() == TK_SYMBOL)
            {
                char *filename = new char[lex_stream -> NameStringLength(current_token_index) + 1];
                strcpy(filename, lex_stream -> NameString(current_token_index));

                Token save_include_key = *(lex_stream -> GetTokenReference(include_key_index)),
                      save_filename = *current_token;
                char *save_cursor = cursor;

                lex_stream -> ResetTokenStream(include_key_index);
                if (*cursor == option -> escape)
                {
                    char *p = cursor + 1;
                    while (IsAlnum(*p) && *p != option -> escape)
                        p++;
                    int len = p - cursor;
                    if ((scan_keyword[len < 13 ? len : 0])(cursor) == TK_END_KEY)
                        ptr = p; // move the pointer
                }

                InputFileSymbol *save_current_input_file = input_file;

                int error_code = IncludeFile(filename);
                if (error_code != LexStream::NO_ERROR)
                {
                    current_token_index = lex_stream -> GetNextToken(input_file, save_include_key.StartLocation());
                    current_token = lex_stream -> GetTokenReference(current_token_index);
                    *current_token = save_include_key;

                    current_token_index = lex_stream -> GetNextToken(input_file, save_filename.StartLocation());
                    current_token = lex_stream -> GetTokenReference(current_token_index);
                    *current_token = save_filename;

                    lex_stream -> AddBadToken(error_code, current_token_index);
                    ptr = save_cursor;
                }

                //
                // Restore the enclosing (saved) input file environment.
                //
                input_file = save_current_input_file;
                input_buffer = input_file -> Buffer();
                line_location = input_file -> LineLocationReference();

                delete [] filename;
            }
        }
    }

    cursor = ptr;

    return;
}


//
//
//
void Scanner::ClassifySymbol()
{
    char *ptr = cursor + 1;

    while ((! IsSpace(*ptr)) && (*ptr != option -> escape))
        ptr++;
    int len = ptr - cursor;

    current_token -> SetKind(TK_SYMBOL);
    current_token -> SetSymbol(variable_table -> FindOrInsertName(cursor, len));
    current_token -> SetEndLocation((ptr - 1) - input_buffer);

    cursor = ptr;

    return;
}


//
//
//
void Scanner::ClassifyEquivalence()
{
    if (cursor[1] == ':' && cursor[2] == '=' && IsSpace(cursor[3]))
    {
        current_token -> SetKind(TK_EQUIVALENCE);
        current_token -> SetEndLocation((cursor + 2) - input_buffer);
        cursor += 3;
    }
    else if (cursor[1] == ':' && cursor[2] == '=' &&
             cursor[3] == '?' && IsSpace(cursor[4]))
    {
        current_token -> SetKind(TK_PRIORITY_EQUIVALENCE);
        current_token -> SetEndLocation((cursor + 3) - input_buffer);
        cursor += 4;
    }
    else ClassifySymbol();

    return;
}


//
//
//
void Scanner::ClassifyArrow()
{
    if (cursor[1] == '>' && IsSpace(cursor[2]))
    {
        current_token -> SetKind(TK_ARROW);
        current_token -> SetEndLocation((cursor + 1) - input_buffer);
        cursor += 2;
    }
    else if (cursor[1] == '>' && cursor[2] == '?' && IsSpace(cursor[3]))
    {
        current_token -> SetKind(TK_PRIORITY_ARROW);
        current_token -> SetEndLocation((cursor + 2) - input_buffer);
        cursor += 3;
    }
    else ClassifySymbol();

    return;
}


//
//
//
void Scanner::ClassifyOr()
{
    if (IsSpace(cursor[1]))
    {
        current_token -> SetKind(TK_OR_MARKER);
        current_token -> SetEndLocation(cursor - input_buffer);
        cursor++;
    }
    else ClassifySymbol();

    return;
}


void Scanner::ClassifySingleQuotedSymbol()
{
    char delimiter = *cursor,
         *ptr;

    for (ptr = cursor + 1; ! IsNewline(*ptr); ptr++)
    {
        if (ptr[0] == delimiter)
            if (ptr[1] == delimiter)
                 ptr++;
            else break;
    }

    int length = ptr - cursor - 1;
    if (length == 0)
    {
        current_token -> SetKind(TK_EMPTY_KEY);
        current_token -> SetEndLocation(ptr - input_buffer);
    }
    else
    {
        char *name = new char[length + 1];
        char *p1 = name,
             *p2 = cursor + 1;

        current_token -> ResetInfoAndSetLocation(input_file, p2 - input_buffer);

        do
        {
            *(p1++) = *p2;
            p2 += (p2[0] == delimiter && p2[1] == delimiter ? 2 : 1); // turn 2 consecutive quotes into one.
        } while(p2 < ptr);

        current_token -> SetKind(TK_SYMBOL);
        current_token -> SetSymbol(variable_table -> FindOrInsertName(name, p1 - name));
        current_token -> SetEndLocation((ptr - 1) - input_buffer);

        delete [] name;
    }

    if (*ptr == delimiter)
    {
        cursor = ptr + 1;
        //
        // TODO: This test appeared to have been here for legacy reasons. 
        // Thus, it has been disabled until someone screams and then it will be
        // reevaluated.
        //
        // if (! IsSpace(*cursor))
        //    lex_stream -> AddBadToken(LexStream::UNDELIMITED_STRING_SYMBOL, current_token_index);
    }
    else
    {
        cursor = ptr;
        lex_stream -> AddBadToken(LexStream::UNTERMINATED_STRING_SYMBOL, current_token_index);
    }

    return;
}


void Scanner::ClassifyDoubleQuotedSymbol()
{
    char delimiter = *cursor,
         *ptr;

    for (ptr = cursor + 1; ! IsNewline(*ptr); ptr++)
    {
        if (ptr[0] == delimiter)
            if (ptr[1] == delimiter)
                 ptr++;
            else break;
    }

    int length = ptr - cursor - 1;
    if (length == 0)
    {
        current_token -> SetKind(TK_EMPTY_KEY);
        current_token -> SetEndLocation(ptr - input_buffer);
    }
    else
    {
        char *name = new char[length + 1];
        char *p1 = name,
             *p2 = cursor + 1;

        current_token -> ResetInfoAndSetLocation(input_file, p2 - input_buffer);

        do
        {
            if (p2[0] == '\\')
            {
                switch(p2[1])
                {
                    case 'b':
                        *(p1++) = '\b';
                        p2 += 2;
                        break;
                    case 't':
                        *(p1++) = '\t';
                        p2 += 2;
                        break;
                    case 'n':
                        *(p1++) = '\n';
                        p2 += 2;
                        break;
                    case 'f':
                        *(p1++) = '\f';
                        p2 += 2;
                        break;
                    case 'r':
                        *(p1++) = '\r';
                        p2 += 2;
                        break;
                    case '\"':
                        *(p1++) = '\"';
                        p2 += 2;
                        break;
                    case '\'':
                        *(p1++) = '\'';
                        p2 += 2;
                        break;
                    case '\\':
                        *(p1++) = '\\';
                        p2 += 2;
                        break;
                    case '0': case '1': case '2': case '3': case '4':
                    case '5': case '6': case '7': case '8': case '9':
                        {
                            int num = 0,
                                i;
                            for (i = 1; i < 4 ; i++)
                            {
                                char c = p2[i];
                                if (IsDigit(c) && c < '8')
                                     num = num * 8 + (c - '0');
                                else break;
                            }

                            //
                            // If the character is a ASCII code save it.
                            //
                            if (i > 1 && num <= 128)
                            {
                                *(p1++) = num;
                                p2 += i;
                            }
                            else
                            {
                                *(p1++) = *(p2++);
                                lex_stream -> AddBadToken(LexStream::BAD_OCTAL_ASCII_CODE, current_token_index);
                            }
                        }
                        break;
                    case 'u':
                        {
                            int num = 0,
                                i;
                            for (i = 2; i < 6; i++)
                            {
                                char c = p2[i];
                                if (IsDigit(c))
                                     num = num * 16 + (c - '0');
                                else if (IsAlpha(c)) 
                                {
                                    c = ToUpper(c);
                                    if (ToUpper(c) <= 'F')
                                         num = num * 16 + 10 + (c - 'A');
                                    else break;
                                }
                                else break;
                            }

                            //
                            // If the character is a valid unicode character, convert
                            // it to its UTF8 format. 
                            //
                            if (i == 6)
                            {
                                if (num == 0)
                                {
                                     *(p1++) = (char) 0xC0;
                                     *(p1++) = (char) 0x80;
                                }
                                else if (num <= 0x007F)
                                     *(p1++) = (char) num;
                                else if (num <= 0x07FF)
                                {
                                     *(p1++) = (char) ((char) 0xC0 | (char) ((num >> 6) & 0x0000001F)); // bits 6-10
                                     *(p1++) = (char) ((char) 0x80 | (char) (num & 0x0000003F));        // bits 0-5
                                }
                                else
                                {
                                     *(p1++) = (char) ((char) 0xE0 | (char) ((num >> 12) & 0x0000000F));
                                     *(p1++) = (char) ((char) 0x80 | (char) ((num >> 6) & 0x0000003F));
                                     *(p1++) = (char) ((char) 0x80 | (char) (num & 0x0000003F));
                                }
                                p2 += i;
                            }
                            else
                            {
                                *(p1++) = *(p2++);
                                lex_stream -> AddBadToken(LexStream::BAD_UNICODE, current_token_index);
                            }
                        }
                        break;
                    default:
                        *(p1++) = *(p2++);
                        lex_stream -> AddBadToken(LexStream::ISOLATED_BACKSLASH, current_token_index);
                        break;
                }
            }
            else
            {
                *(p1++) = *p2;
                p2 += (p2[0] == delimiter && p2[1] == delimiter ? 2 : 1); // turn 2 consecutive quotes into one.
            }
        } while(p2 < ptr);

        current_token -> SetKind(TK_SYMBOL);
        current_token -> SetSymbol(variable_table -> FindOrInsertName(name, p1 - name));
        current_token -> SetEndLocation((ptr - 1) - input_buffer);

        delete [] name;
    }

    if (*ptr == delimiter)
    {
        cursor = ptr + 1;
        //
        // TODO: This test appeared to have been here for legacy reasons. 
        // Thus, it has been disabled until someone screams and then it will be
        // reevaluated.
        //
        //if (! IsSpace(*cursor))
        //    lex_stream -> AddBadToken(LexStream::UNDELIMITED_STRING_SYMBOL, current_token_index);
    }
    else
    {
        cursor = ptr;
        lex_stream -> AddBadToken(LexStream::UNTERMINATED_STRING_SYMBOL, current_token_index);
    }

    return;
}


void Scanner::ClassifyLess()
{
    char *ptr = cursor + 1;

    if (IsAlpha(*ptr))
    {
        for (ptr++; *ptr != '>' && (! IsNewline(*ptr)); ptr++)
            ;
        if (*ptr == '>')
             ptr++;
        else lex_stream -> AddBadToken(LexStream::UNTERMINATED_BRACKET_SYMBOL, current_token_index);

        current_token -> SetKind(TK_SYMBOL);
        current_token -> SetSymbol(variable_table -> FindOrInsertName(cursor, ptr - cursor));
        current_token -> SetEndLocation((ptr - 1) - input_buffer);
        cursor = ptr;
    }
    else ClassifySymbol();

    return;
}


void Scanner::ClassifyEof()
{
    current_token -> SetKind(TK_EOF);

    cursor++;

    return;
}


void Scanner::ClassifyBadToken()
{
    assert(! "this can happen?");

    return;
}


int Scanner::IncludeFile(const char *filename)
{
    input_file = lex_stream -> FindOrInsertFile(option -> include_search_directory, filename);

    //
    // If we could not open the include file, issue an error.
    //
    if (input_file == NULL)
        return LexStream::NO_INCLUDE;
    else
    {
        //
        // If the file is already in the process of being included, issue an error and stop.
        //
        if (input_file -> IsLocked())
            return LexStream::RECURSIVE_INCLUDE;
        else
        {
            //
            // Lock the include_file to avoid looping...
            //
            input_file -> Lock();
            input_file -> ReadInput();
            input_buffer = input_file -> Buffer();
            if (input_buffer == NULL)
                return LexStream::NO_INCLUDE;
            line_location = input_file -> LineLocationReference();
            line_location -> Next() = 0; // mark starting location of line # 0
            cursor = input_buffer;
            SkipOptions();
            Scan(cursor, &(input_buffer[input_file -> BufferLength()]));
            while(lex_stream -> Kind(current_token_index) == TK_EOF)
                current_token_index--;
            lex_stream -> ResetTokenStream(current_token_index + 1);
            input_file -> Unlock();
        }
    }

    return LexStream::NO_ERROR;
}
