/*******************************************************************************
 * Copyright (c) 2009 STMicroelectronics.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *   Xavier Raynaud <xavier.raynaud@st.com> - initial API and implementation
 *******************************************************************************/
package org.eclipse.linuxtools.gprof.action;

import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.OutputStream;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.eclipse.cdt.core.model.CModelException;
import org.eclipse.cdt.core.model.CoreModel;
import org.eclipse.cdt.core.model.IBinary;
import org.eclipse.cdt.core.model.ICProject;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.linuxtools.gprof.Activator;
import org.eclipse.linuxtools.gprof.dialog.OpenGmonDialog;
import org.eclipse.linuxtools.gprof.view.GmonView;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.IEditorLauncher;
import org.eclipse.ui.PlatformUI;

import com.xmos.cdt.core.CdtCorePluginDbg;
import com.xmos.cdt.core.utils.CommandRunner;
import com.xmos.libs.core.utils.Directory;
import com.xmos.libs.core.utils.Logger;
import com.xmos.libs.core.utils.Pair;
import com.xmos.libs.core.xnloader.XnLoader;

/**
 * Action performed when user clicks on a gmon file
 *
 * @author Xavier Raynaud <xavier.raynaud@st.com>
 */
public class OpenGmonAction implements IEditorLauncher {

	private static int instanceNum = 0;

	public void open(IPath file) {
		String s = getDefaultBinary(file);
		Shell shell = PlatformUI.getWorkbench().getDisplay().getActiveShell();
		OpenGmonDialog d = new OpenGmonDialog(shell, s, file);
		if (d.open() != OpenGmonDialog.OK) {
			return;
		}
		String binaryPath = d.getBinaryFile();
		IProject project = getProject(file);

		// XMOS
		try {
		    final File tempDir = Directory.createTempDirectory("xde");
            String elfImagePath = getElfImagePath(tempDir, binaryPath, file.toFile(), project);
            GmonView.displayGprofView(elfImagePath, file.toOSString(), Integer.toString(instanceNum++));            

            Job job = new Job("Deleting temporary files") {
                protected IStatus run(IProgressMonitor monitor) {
                    try {
                        Directory.deleteDirectory(tempDir);
                    } catch (Exception e) {}       
                    return Status.OK_STATUS;
                }
            };
            job.setUser(false);
            job.schedule();            

        } catch (Exception e) {
            e.printStackTrace();
        }		 
	}

	private String getDefaultBinary(IPath file) {
		File gmonFile = file.toFile();
		File parent = gmonFile.getParentFile();
		File info = new File(parent,"AnalysisInfo.txt");
		IProject project = null;
		IFile ifile = null;
		try {
			String line;
			String  tab[];
			if (info.exists()) {
				BufferedReader br = new BufferedReader(new FileReader(info.toString()));

				while ((line = br.readLine())!= null){
					tab = line.split("=");
					String name="", value="";
					if (tab.length > 1){
						name=tab[0];
						value=tab[1].trim();	
						if (name.equals("Project Name ")){
							project = ResourcesPlugin.getWorkspace().getRoot().getProject(value);
						}
						else if (name.equals("Program Name ")){
							if(project != null){
								ifile = project.getFile(value);
								br.close();
								if (ifile.exists()) {
									return ifile.getLocation().toString();
								}
							}
						}	
					}
				}
			}else{
				IFile c = ResourcesPlugin.getWorkspace().getRoot()
				.getFileForLocation(file);
				if (c != null) {
					project = c.getProject();
					if (project != null && project.exists()) {
						ICProject cproject = CoreModel.getDefault().create(project);
						if (cproject != null) {
							try {
								IBinary[] b = cproject.getBinaryContainer()
								.getBinaries();
								if (b != null && b.length > 0 && b[0] != null) {
									IResource r = b[0].getResource();
									return r.getLocation().toOSString();
								}
							} catch (CModelException _) {
							}
						}
					}
				}
			}
		} catch (FileNotFoundException e) {
			Status status = new Status(
					Status.WARNING,
					Activator.PLUGIN_ID,
					IStatus.WARNING,
					e.getMessage(),
					e
			);
			Activator.getDefault().getLog().log(status);
		} catch (IOException e) {
			Status status = new Status(
					Status.ERROR,
					Activator.PLUGIN_ID,
					IStatus.ERROR,
					e.getMessage(),
					e
			);
			Activator.getDefault().getLog().log(status);
		}
		return "";
	}	
	
	private IProject getProject(IPath file) {
        IFile c = ResourcesPlugin.getWorkspace().getRoot().getFileForLocation(file);
        if (c != null) {
            return c.getProject();
        }
        return null;
	}
	
	private String getElfImagePath(File tempDir, String xeFile, File gprofFile, IProject project) throws Exception {        
        OutputStream consoleOut = new ByteArrayOutputStream();
        OutputStream consoleErr = new ByteArrayOutputStream();          
        
        String command = "xobjdump";
        String[] args = {"--split", "--split-dir", tempDir.getAbsolutePath(), xeFile};
        
        if (CdtCorePluginDbg.DEBUG) {
            Logger.printf("OpenGmonAction: Running command: %s", command);
            for (String arg : args) {
                Logger.printf(" %s", arg);
            }
            Logger.printf("\n");
        }
        
        CommandRunner commandRunner = new CommandRunner(project);
        commandRunner.run(consoleOut, consoleErr, command, args, new String[] {});
                    
        if (CdtCorePluginDbg.DEBUG) {
            Logger.printf("STDOUT: %s\n", consoleOut.toString());
            Logger.printf("STDERR: %s\n", consoleErr.toString());            
        }

        consoleOut.close();
        consoleErr.close();        
        
        return getElfFileName(tempDir, gprofFile, xeFile); 
	}
	    
    private String getElfFileName(File tempDir, File gprofFile, String xeFile) throws Exception {
        Pair<String, String> nodeCore = getNodeCore(gprofFile, xeFile);
        
        // Note: Due to the multistage boot, there could be multiple elf files for a given node/core.
        // In this case we want to find the one with the highest number and use this as the elf image..
        String baseName = tempDir + File.separator + "image_n" + nodeCore.first + "c" + nodeCore.second;        
        String imageName = baseName + ".elf";
        int extension = 2;        
        while(true) {
            String image = baseName + "_" + extension + ".elf";
            if (new File(image).exists()) {
                imageName = image;
                
            } else {
                break;
            }
            ++extension;
        }        
        return imageName;      
    }
	
    private Pair<String, String> getNodeCore(File gprofFile, String xeFile) throws Exception {
        // Note: The grof files are named in one of 2 ways...
        String[] parts = gprofFile.getName().split("_");
        if (parts.length == 2) {
            // 1) tile[0]_core0.gprof 
            //    In this case, we get the node and core indices
            //    for this tile from the xn file in the binary.
            String coreReference = parts[0];
            XnLoader xnLoader = new XnLoader();
            xnLoader.load(new File(xeFile));
            Pair<Integer, Integer> nodeCore = xnLoader.getNodeChainIndexCoreNumber(coreReference);            
            return new Pair<String, String>(
                    Integer.toString(nodeCore.first),
                    Integer.toString(nodeCore.second));
            
        } else {
            // 2) n0t0c0.gprof
            //    In this case, we get the node and core indices
            //    directly from the filename.
            return new Pair<String, String>(
                    getField(gprofFile, "Node", 1),
                    getField(gprofFile, "Tile", 2));                
        }
    }
    
    private String getField(File gprofFile, String fieldName, int fieldNum) throws Exception {
        Pattern pattern = Pattern.compile("^n(\\d+)t(\\d+).*\\.gprof");
        Matcher matcher = pattern.matcher(gprofFile.getName());
        if (matcher.find()) {
            return matcher.group(fieldNum).trim();
        }
        throw new Exception(fieldName + " number not found in filename: " + gprofFile.getName());
    }    
}
