#include <sys/stat.h>
#include "control.h"
#include "grammar.h"

#include <iostream>
using namespace std;

Grammar::Grammar(Control *control_,
                 Action *action_blocks_,
                 LexStream *lex_stream_,
                 VariableLookupTable *variable_table_,
                 MacroLookupTable *macro_table_)
                                  : control(control_),
                                    action_blocks(action_blocks_),
                                    lex_stream(lex_stream_),
                                    variable_table(variable_table_),
                                    macro_table(macro_table_),
                                    parser(control_, lex_stream_, variable_table_, macro_table_),

                                    symbol_index(4, variable_table_ -> Size()),
                                    rules(2, 2048),
                                    rhs_sym(4, 8192),
                                    name(2, variable_table_ -> Size())
{
    option = control_ -> option;
    identifier_symbol = NULL;
    eol_symbol = NULL;
    eof_symbol = NULL;
    error_symbol = NULL;

    num_items = 0;
    num_symbols = 0;
    num_single_productions = 0;
    num_rules = 0;

    empty = 0;
    identifier_image = 0;
    eol_image = 0;
    eof_image = 0;
    error_image = 0;
    accept_image = 0;
    start_image = 0;

    variable_symbol_pool.Next() = null_symbol = new VariableSymbol("", 0, 0, 0);

    empty_symbol = allocate_variable_symbol("empty");
    accept_symbol = allocate_variable_symbol("accept");

    return;
}

void Grammar::Process()
{
    parser.Parse();
    if (parser.rules.Length() == 0)
    {
        Tuple<const char *> msg;
        msg.Next() = option -> grm_file;
        msg.Next() = " is an empty grammar... Processing stopped";
        lex_stream -> EmitError(msg, lex_stream -> NumTokens() - 2); // point to the last token
    }

    //
    // Add all the exported symbols to the export_macro_table.
    //
    for (int i = 0; i < parser.exports.Length(); i++)
    {
        int export_token = parser.exports[i];
        (void) InsertExportMacro(lex_stream -> NameString(export_token), lex_stream -> NameStringLength(export_token));
    }
    for (int i = 0; i < lex_stream -> NumImportedFilters(); i++)
    {
        int export_token = lex_stream -> ImportedFilter(i);
        (void) InsertExportMacro(lex_stream -> NameString(export_token), lex_stream -> NameStringLength(export_token));
    }

    //
    // Make sure that none of these macros also appear in the local macro table
    //
    for (int i = 0; i < export_macro_table.Size(); i++)
    {
        SimpleMacroSymbol *macro = export_macro_table[i];
        if (FindLocalMacro(macro -> Name(), macro -> NameLength()))
        {
            option -> report.Put("*** The name of the exported symbol ");
            option -> report.Put(macro -> Name());
            option -> report.Put(" conflicts with the predefined macro of the same name\n");
        }
    }

    //
    // We now make sure that none of the user-defined macros appear
    // in either the local macro table or the export_macro_table.
    //
    for (int i = 0; i < macro_table -> Size(); i++)
    {
        MacroSymbol *macro = macro_table -> Element(i);
        if (FindLocalMacro(macro -> Name(), macro -> NameLength()))
        {
            option -> report.Put("*** The user-defined macro ");
            option -> report.Put(macro -> Name());
            option -> report.Put(" conflicts with the predefined macro of the same name\n");
        }
        else if (export_macro_table.FindName(macro -> Name(), macro -> NameLength()))
        {
            option -> report.Put("*** The user-defined macro ");
            option -> report.Put(macro -> Name());
            option -> report.Put(" conflicts with an exported symbol of the same name\n");
        }
    }

    symbol_index.Next(); // skip the 0th element

    ProcessTerminals();
    ProcessRules();
    ProcessExportedTerminals();
    ProcessNames();
    if (option -> list)
        DisplayInput();

    option -> report.Print(Control::syslis);
    option -> report.Flush(stdout);

    return;
}


VariableSymbol *Grammar::GetSymbol(int sym_index)
{
    VariableSymbol *symbol;

    switch(lex_stream -> Kind(sym_index))
    {
        case TK_EMPTY_KEY:
             symbol = empty_symbol;
             break;
        case TK_IDENTIFIER_KEY:
             symbol = identifier_symbol;
             break;
        case TK_ERROR_KEY:
             symbol = error_symbol;
             break;
        case TK_EOF_KEY:
             symbol = eof_symbol;
             break;
        case TK_EOL_KEY:
             symbol = eol_symbol;
             break;
        default:
            assert(lex_stream -> Kind(sym_index) == TK_SYMBOL ||
                   lex_stream -> Kind(sym_index) == TK_MACRO_NAME);
            symbol = lex_stream -> GetVariableSymbol(sym_index);
            break;
    }

    return symbol;
}


int Grammar::GetSymbolIndex(int index)
{
    VariableSymbol *symbol = GetSymbol(index);
    return (symbol ? symbol -> SymbolIndex() : 0);
}


int Grammar::AssignSymbolIndex(VariableSymbol *symbol)
{
    int index = symbol -> SymbolIndex();
    if (index == 0)
    {
        index = ++num_symbols;
        symbol -> SetSymbolIndex(index);

        int symno_index = symbol_index.NextIndex();
        assert(symno_index == index);
        symbol_index[symno_index].symbol = symbol;
        symbol_index[symno_index].external_name_index = symbol -> NameIndex();
    }

    return index;
}


//
// Construct set of exported symbols. Make sure that there are no duplicates.
//
void Grammar::ProcessExportedTerminals(void)
{
    Array<int> import_index(variable_table -> Size(), -1);

    //
    // Merge the exported symbols inherited from filter files
    // with the symbols that are to be locally exported
    // and add them to the exported list.
    //
    Tuple<int> imports;
    for (int i = 0; i < lex_stream -> NumImportedFilters(); i++)
        imports.Next() = lex_stream -> ImportedFilter(i);
    for (int i = 0; i < parser.exports.Length(); i++)
        imports.Next() = parser.exports[i];
    for (int i = 0; i < imports.Length(); i++)
    {
        int import = imports[i];
        VariableSymbol *symbol = lex_stream -> GetVariableSymbol(import);
        int decl = import_index[symbol -> Index()];
        if (decl == -1)
        {
            import_index[symbol -> Index()] = import;
            exported_symbols.Next() = symbol;
        }
        else
        {
            Tuple<const char *> msg;
            msg.Next() = "Symbol \"";
            msg.Next() = symbol -> Name();
            msg.Next() = "\" was specified more than once in an Export section. Previous specification was located at ";
            msg.Next() = lex_stream -> FileName(decl);
            msg.Next() = ":";
            IntToString start_line(lex_stream -> Line(decl));
            msg.Next() = start_line.String();
            msg.Next() = ":";
            IntToString start_column(lex_stream -> Column(decl));
            msg.Next() = start_column.String();
            msg.Next() = ":";
            IntToString end_line(lex_stream -> EndLine(decl));
            msg.Next() = end_line.String();
            msg.Next() = ":";
            IntToString end_column(lex_stream -> EndColumn(decl));
            msg.Next() =  end_column.String();
            msg.Next() = ":";
            IntToString start_location(lex_stream -> StartLocation(decl));
            msg.Next() = start_location.String();
            msg.Next() = ":";
            IntToString end_location(lex_stream -> EndLocation(decl));
            msg.Next() = end_location.String();
            lex_stream -> EmitWarning(msg, import);
        }
    }

    return;
}


void Grammar::ProcessTerminals(void)
{
    empty = AssignSymbolIndex(empty_symbol);

    Tuple<int> declared_terminals;
    declared_terminals.Next() = empty;

    //
    // Process the declared set of terminal symbols.
    // This is useful for upward compatibility with old versions where
    // the declaration of terminals was required. It will also pick up
    // terminal symbols that are declared but never used.
    //
    for (int i = 0; i < parser.terminals.Length(); i++)
    {
        VariableSymbol *terminal = lex_stream -> GetVariableSymbol(parser.terminals[i]);
        assert(terminal);
        declared_terminals.Next() = AssignSymbolIndex(terminal);
    }
    for (int i = 0; i < parser.recovers.Length(); i++)
    {
        VariableSymbol *terminal = lex_stream -> GetVariableSymbol(parser.recovers[i].symbol_index);
        assert(terminal);
        declared_terminals.Next() = AssignSymbolIndex(terminal);
    }
    for (int i = 0; i < lex_stream -> NumImportedTerminals(); i++) // imported from another grammar
    {
        int import = lex_stream -> ImportedTerminal(i);
        VariableSymbol *terminal = lex_stream -> GetVariableSymbol(import);
        declared_terminals.Next() = AssignSymbolIndex(terminal);
    }
    for (int j = 0; j < parser.keywords.Length(); j++)
    {
        VariableSymbol *keyword = lex_stream -> GetVariableSymbol(parser.keywords[j]);
        assert(keyword);
        declared_terminals.Next() = AssignSymbolIndex(keyword);
    }

    //
    // Compute the set of terminals directly from the grammar, even
    // though the user may have declared them using a terminals and keywords section;
    // Traverse the right-hand side and initially, add all symbols found to the
    // term_set.
    //
    BitSet term_set(variable_table -> Size(), BitSet::EMPTY);
    for (int k = 0; k < parser.rules.Length(); k++)
    {
        for (int i = lex_stream -> Next(parser.rules[k].separator_index);
                 i < parser.rules[k].end_rhs_index;
                 i = lex_stream -> Next(i))
        {
            VariableSymbol *symbol = lex_stream -> GetVariableSymbol(i);
            if (symbol)
                term_set.AddElement(symbol -> Index());
        }
    }

    //
    // Add all symbols appearing on the right-hand side of an alias rule to the term_set.
    //
    for (int l = 0; l < parser.aliases.Length(); l++)
    {
        VariableSymbol *symbol = lex_stream -> GetVariableSymbol(parser.aliases[l].rhs_index); // the right-hand-side symbol
        if (symbol) // no symbol is defined for EMPTY
            term_set.AddElement(symbol -> Index());
    }

    //
    // Remove all symbols appearing on the left-hand side of an alias rule from the term_set.
    //
    for (int m = 0; m < parser.aliases.Length(); m++)
    {
        VariableSymbol *symbol = lex_stream -> GetVariableSymbol(parser.aliases[m].lhs_index); // the left-hand-side symbol
        assert(symbol);
        term_set.RemoveElement(symbol -> Index());
    }

    //
    // Remove all symbols appearing on the left-hand side of a grammar rule from the term_set.
    //
    BitSet nterm_set(variable_table -> Size(), BitSet::EMPTY);
    for (int n = 0; n < parser.rules.Length(); n++)
    {
        VariableSymbol *symbol = lex_stream -> GetVariableSymbol(parser.rules[n].lhs_index);
        assert(symbol);
        term_set.RemoveElement(symbol -> Index());
        nterm_set.AddElement(symbol -> Index());
    }

    //
    // Remove all symbols to whom a known non-terminal is aliased.
    // This iterative algorithm is in general INEFFICIENT. However, since
    // the Alias section is likely to be small, it does not matter.
    //
    // TODO: compute this right... first compute partial order of
    //       nonterminals based on Alias relation, etc...
    //
    bool changed = true;
    while (changed)
    {
        changed = false;

        //
        // Remove all symbols to whom a known non-terminal is aliased.
        // I.e., any symbol appearing on the right-hand side of an alias rule
        // where the left-hand side symbol was also used as the left-hand
        // side of a grammar rule.
        //
        for (int i = 0; i < parser.aliases.Length(); i++)
        {
            VariableSymbol *symbol = lex_stream -> GetVariableSymbol(parser.aliases[i].lhs_index);
            assert(symbol);
            if (nterm_set[symbol -> Index()]) // The lhs is a nonterminal?
            {
                symbol = lex_stream -> GetVariableSymbol(parser.aliases[i].rhs_index); // the left-hand-side symbol
                if (! nterm_set[symbol -> Index()]) // The rhs is NOT a nonterminal?
                {
                    assert(symbol);
                    changed = true;
                    term_set.RemoveElement(symbol -> Index());
                    nterm_set.AddElement(symbol -> Index());
                }
            }
        }
    }

    //
    // Now process the declared set of terminal symbols.
    //
    for (int o = 0; o < variable_table -> Size(); o++)
    {
        VariableSymbol *symbol = variable_table -> Element(o);
        if (term_set[symbol -> Index()])
            (void) AssignSymbolIndex(symbol);
    }

    //
    // Now take care of the terminal aliasses that are not aliassed to special symbols.
    //
    Tuple<int> remaining_aliases;
    ProcessInitialAliases(remaining_aliases);

    //
    // Process the special symbols.
    //
    if (parser.identifier_index != 0)
    {
         identifier_symbol = lex_stream -> GetVariableSymbol(parser.identifier_index);
         assert(identifier_symbol);
         identifier_image = AssignSymbolIndex(identifier_symbol);
         declared_terminals.Next() = identifier_image;
    }

    if (parser.eol_index != 0)
    {
         eol_symbol = lex_stream -> GetVariableSymbol(parser.eol_index);
         assert(eol_symbol);
         eol_image = AssignSymbolIndex(eol_symbol);
         declared_terminals.Next() = eol_image;
    }

    if (parser.eof_index != 0)
    {
         eof_symbol = lex_stream -> GetVariableSymbol(parser.eof_index);
         assert(eof_symbol);
         eof_image = AssignSymbolIndex(eof_symbol);
    }
    else
    {
        eof_symbol = allocate_variable_symbol("eof");
        eof_image = AssignSymbolIndex(eof_symbol);
    }
    declared_terminals.Next() = eof_image;

    if (parser.error_index != 0)
    {
         error_symbol = lex_stream -> GetVariableSymbol(parser.error_index);
         assert(error_symbol);
         error_image = AssignSymbolIndex(error_symbol);
         declared_terminals.Next() = error_image;
    }
    else if (option -> error_maps)
    {
         error_symbol = allocate_variable_symbol("error");
         error_image = AssignSymbolIndex(error_symbol);
         declared_terminals.Next() = error_image;
    }
    else error_image = DEFAULT_SYMBOL;

    //
    // Compute the set of "recover" symbols identified by the user.
    // Two representations are used: a bit set for ease of
    // queries and a tuple for ease of iteration.
    //
    // TODO: save the block (containing the allocation expression) for
    // each recover symbol, if it was specified. If the block was not
    // specified, save the default expression (getRhsIToken(...)).
    //
    recover_set.Initialize(num_symbols + 1);
    for (int p = 0; p < parser.recovers.Length(); p++)
    {
        VariableSymbol *recover = lex_stream -> GetVariableSymbol(parser.recovers[p].symbol_index);
        if (! recover_set[recover -> SymbolIndex()])
        {
            recovers.Next() = recover -> SymbolIndex();
            recover_set.AddElement(recover -> SymbolIndex());
        }
    }

    //
    // Compute the set of keywords identified by the user.
    // Two representations are used: a bit set for ease of
    // queries and a tuple for ease of iteration.
    //
    keyword_set.Initialize(num_symbols + 1);
    for (int p = 0; p < parser.keywords.Length(); p++)
    {
        VariableSymbol *keyword = lex_stream -> GetVariableSymbol(parser.keywords[p]);
        if (! keyword_set[keyword -> SymbolIndex()])
        {
            keywords.Next() = keyword -> SymbolIndex();
            keyword_set.AddElement(keyword -> SymbolIndex());
        }
    }

    //
    // Compute the number of terminals in the language.
    // It is important that this operation be performed here!
    // Note that the next symbol following last terminal is the
    // Accept nonterminal.
    //
    num_terminals = num_symbols;
    accept_image = AssignSymbolIndex(accept_symbol);

    ProcessRemainingAliases(remaining_aliases);

    //
    // If the warnings option is on then check to see if the grammar
    // contained undeclared terminals and if so, print them.
    //
    if (option -> warnings)
    {
        BitSet declared_set(num_terminals + 1, BitSet::EMPTY);
        int num_undeclared_terminals = num_terminals; // Assume that all terminals are not declared
        for (int i = 0; i < declared_terminals.Length(); i++)
        {
            int symbol = declared_terminals[i];
            if (! declared_set[symbol])
            {
                declared_set.AddElement(symbol);
                num_undeclared_terminals--; // Each time we find a terminal declaration, decrease undeclared count
            }
        }
        if (num_undeclared_terminals > 0) // Some terminals were not declared?
        {
            enum { MAX_LENGTH = Control::PRINT_LINE_SIZE - 4 };

            option -> report.Put("*** The following undeclared symbol");
            option -> report.Put(num_undeclared_terminals > 1 ? "s are" : " is");
            option -> report.Put(" assumed to be ");
            option -> report.Put(num_undeclared_terminals > 1 ? "Terminals:\n\n" : "a terminal: ");

            char line[Control::PRINT_LINE_SIZE + 1],
                 tok[Control::SYMBOL_SIZE + 1];
            if (num_undeclared_terminals > 1)
                 strcpy(line, "        "); // 8 spaces
            else line[0] = '\0';
            for (int symbol = 1; symbol <= num_terminals; symbol++)
            {
                if (! declared_set[symbol])
                {
                    RestoreSymbol(tok, RetrieveString(symbol));
                    if (strlen(line) + strlen(tok) > Control::PRINT_LINE_SIZE)
                    {
                        option -> report.Put(line);
                        option -> report.Put("\n");
                        ReportLargeToken(line, tok, "    ", MAX_LENGTH);
                    }
                    else
                    {
                        strcat(line, tok);
                        strcat(line, " ");
                    }
                    strcat(line, " ");
                }
            }
            option -> report.Put(line);
            option -> report.Put("\n");
        }
    }

    return;
}


void Grammar::ProcessInitialAliases(Tuple<int> &remaining_aliases)
{
    for (int i = 0; i < parser.aliases.Length(); i++)
    {
        int rhs_kind = lex_stream -> Kind(parser.aliases[i].rhs_index);

        if (rhs_kind == TK_EMPTY_KEY ||
            rhs_kind == TK_IDENTIFIER_KEY ||
            rhs_kind == TK_ERROR_KEY ||
            rhs_kind == TK_EOF_KEY ||
            rhs_kind == TK_EOL_KEY)
            remaining_aliases.Next() = i;
        else
        {
            int image = GetSymbolIndex(parser.aliases[i].rhs_index); // the right-hand-side symbol
            if (image == 0)
                remaining_aliases.Next() = i;
            else // If the right-hand side symbol is a declared symbol
            {
                //
                // Note that when one of the special symbols such as %eof,
                // or %eol etc.. are used on the left side of an alias, they
                // are preprocessed in the grammar as if they were specified
                // in a separate section headed by the keyword in question.
                // The use of these keywords on the left-hand side in the
                // Alias section has been deprecated.
                //
                assert(lex_stream -> Kind(parser.aliases[i].lhs_index) == TK_SYMBOL ||
                       lex_stream -> Kind(parser.aliases[i].lhs_index) == TK_MACRO_NAME);

                VariableSymbol *symbol = lex_stream -> GetVariableSymbol(parser.aliases[i].lhs_index);
                assert(symbol);
                if (symbol -> SymbolIndex() != 0)
                    lex_stream -> EmitError("This symbol was previously defined", parser.aliases[i].lhs_index);
                symbol -> SetSymbolIndex(image);
            }
        }
    }

    return;
}


void Grammar::ProcessRemainingAliases(Tuple<int> &remaining_aliases)
{
    for (int k = 0; k < remaining_aliases.Length(); k++)
    {
        int i = remaining_aliases[k];

        if ((lex_stream -> Kind(parser.aliases[i].rhs_index) == TK_IDENTIFIER_KEY && identifier_symbol == NULL) ||
            (lex_stream -> Kind(parser.aliases[i].rhs_index) == TK_EOL_KEY && eol_symbol == NULL) ||
            (lex_stream -> Kind(parser.aliases[i].rhs_index) == TK_ERROR_KEY && error_symbol == NULL))
        {
            option -> report.Put("*** Error: ");
            option -> report.Put(lex_stream -> NameString(parser.aliases[i].rhs_index));
            option -> report.Put(" can\'t be used as an alias for \"");
            option -> report.Put(lex_stream -> NameString(parser.aliases[i].lhs_index));
            option -> report.Put("\" because it was not defined.\n");
        }
        else
        {
            int image = GetSymbolIndex(parser.aliases[i].rhs_index); // the right-hand-side symbol
            if (image == 0) // an as yet undefined symbol
            {
                VariableSymbol *symbol = lex_stream -> GetVariableSymbol(parser.aliases[i].rhs_index);
                assert(symbol);
                image = AssignSymbolIndex(symbol);
            }

            //
            // Note that when one of the special symbols such as %eof,
            // or %eol etc.. are used on the left side of an alias, they
            // are preprocessed in the grammar as if they were specified
            // in a separate section headed by the keyword in question.
            // The use of these keywords on the left-hand side in the
            // Alias section has been deprecated.
            //
            assert(lex_stream -> Kind(parser.aliases[i].lhs_index) == TK_SYMBOL ||
                   lex_stream -> Kind(parser.aliases[i].lhs_index) == TK_MACRO_NAME);

            VariableSymbol *symbol = lex_stream -> GetVariableSymbol(parser.aliases[i].lhs_index);
            assert(symbol);
            if (symbol -> SymbolIndex() != 0)
                lex_stream -> EmitError("This symbol was previously defined", parser.aliases[i].lhs_index);
            symbol -> SetSymbolIndex(image);
        }
    }

    if (eol_image == 0)
        eol_image = eof_image;

    return;
}


void Grammar::ProcessTitleOrGlobalBlock(int block_token, ActionBlockElement &action)
{
    BlockSymbol *block = lex_stream -> GetBlockSymbol(block_token);
    TextBuffer *buffer = block -> Buffer();

    if (block != option -> DefaultBlock())
        lex_stream -> EmitError("Only default blocks may appear in a Title or Global segment", block_token);
    action.rule_number = 0;
    action.location = ActionBlockElement::INITIALIZE; // does not matter - block must be default block...
    action.block_token = block_token;
    action.buffer = (buffer ? buffer : block -> ActionfileSymbol() -> InitialHeadersBuffer());

    return;
}


char *Grammar::InsertInterface(SymbolLookupTable &symbol_set, char *name)
{
    int length = strlen(name) + 1;
    char *type_name = new char[length + 1];
    strcpy(type_name, "I");
    strcat(type_name, name);
    Symbol *symbol = symbol_set.FindOrInsertName(type_name, length);
    delete [] type_name;

    return symbol -> Name();
}


void Grammar::ProcessRules(void)
{
    Tuple<ActionBlockElement> notice_actions,
                              header_actions,
                              initial_actions,
                              code_actions,
                              ast_actions,
                              trailer_actions;
    Tuple<AttributeElement> attribute_actions;

    //
    // Add starting rule.
    //
    int start_index = (parser.start_index != 0 ? parser.start_index : parser.rules[0].lhs_index);
    start_symbol = lex_stream -> GetVariableSymbol(start_index);
    AssignSymbolIndex(start_symbol);
    int rule_index = rules.NextIndex();
    rules[rule_index].lhs = accept_image;
    rules[rule_index].separator_token = TK_EQUIVALENCE;
    rules[rule_index].produces_token = TK_EQUIVALENCE;
    rules[rule_index].rhs_index = rhs_sym.Length();
    rules[rule_index].source_index = 0;
    if (start_symbol -> SymbolIndex() != empty)
    {
        start_image = start_symbol -> SymbolIndex();
        rhs_sym.Next() = start_image;
    }
    else
    {
        Tuple <const char *> msg;
        msg.Next() = "The start symbol, \"";
        msg.Next() = start_symbol -> Name();
        msg.Next() = "\", is aliased to the Empty symbol";

        lex_stream -> EmitError(msg, start_index);
    }

    //
    // Are there any notice action blocks?
    //
    for (int j = 0; j < parser.notice_blocks.Length(); j++)
    {
        LexStream::TokenIndex block_token = parser.notice_blocks[j];
        BlockSymbol *block = lex_stream -> GetBlockSymbol(block_token);
        if (block != option -> DefaultBlock())
        {
            lex_stream -> EmitError("Only default blocks may appear in a notice segment", block_token);
            exit(12);
        }

        ActionBlockElement &action = notice_actions.Next();
        action.rule_number = 0;
        action.location = ActionBlockElement::INITIALIZE; // does not matter - block must be default block...
        action.block_token = block_token;
        action.buffer = &notice_buffer;
    }

    //
    // Now, process the title block, if present.
    // Then, process all global blocks specified.
    //
    for (int j = 0; j < parser.global_blocks.Length(); j++)
        ProcessTitleOrGlobalBlock(parser.global_blocks[j], header_actions.Next());

    //
    // Are there any action blocks preceding the first rule?
    // These can be either specified in a Header segment or
    // in a Rule segment (legacy). The Header blocks come first.
    //
    for (int j = 0; j < parser.header_blocks.Length(); j++)
    {
        LexStream::TokenIndex block_token = parser.header_blocks[j];
        BlockSymbol *block = lex_stream -> GetBlockSymbol(block_token);
        TextBuffer *buffer = block -> Buffer();

        ActionBlockElement &action = header_actions.Next();
        action.rule_number = 0;
        action.location = ActionBlockElement::INITIALIZE;
        action.block_token = block_token;
        action.buffer = (buffer ? buffer : block -> ActionfileSymbol() -> InitialHeadersBuffer());
    }

    for (int n = 0; n < parser.initial_blocks.Length(); n++)
    {
        LexStream::TokenIndex block_token = parser.initial_blocks[n];
        BlockSymbol *block = lex_stream -> GetBlockSymbol(block_token);
        TextBuffer *buffer = block -> Buffer();

        ActionBlockElement &action = initial_actions.Next();
        action.rule_number = 0;
        action.location = ActionBlockElement::BODY;
        action.block_token = block_token;
        action.buffer = (buffer ? buffer : block -> ActionfileSymbol() -> BodyBuffer());
    }

    //
    // In this loop, the grammar is placed in the rule table structure and the
    // right-hand sides are placed in the RHS table.  A check is made to prevent
    // terminals from being used as left hand sides.
    // Also, if macro symbols are associated with symbols in the right-hand
    // side of certain rules, keep track of such macro names and their position
    // in processed_rule_map.
    //
    Tuple< Tuple<ProcessedRuleElement> > processed_rule_map;
    processed_rule_map.Next(); // rule 0 with no macros has already been allocated.
    SymbolLookupTable classname_set;
    Tuple<ClassnameElement> classname;

    //
    // Process the Token classname.
    //
    {
        int length = strlen(option -> ast_type) + strlen("Token");
        char *name = new char[length + 1];
        strcpy(name, option -> ast_type);
        strcat(name, "Token");
        this -> ast_token_classname = classname_set.FindOrInsertName(name, length) -> Name();
        delete [] name;

        ClassnameElement &element = classname.Next();
        element.specified_name = ast_token_classname;
        element.real_name = ast_token_classname;
        element.is_terminal_class = true;
        element.array_element_type_symbol = NULL;
    }

    for (int k = 0; k < parser.rules.Length(); k++)
    {
        int rule_index = rules.NextIndex();
        Tuple<ProcessedRuleElement> &processed_rule_elements = processed_rule_map.Next();
        VariableSymbol *lhs_symbol = lex_stream -> GetVariableSymbol(parser.rules[k].lhs_index);
        int lhs_image = AssignSymbolIndex(lhs_symbol),
            separator_token = lex_stream -> Kind(parser.rules[k].separator_index);

        if (IsTerminal(lhs_image))
            lex_stream -> EmitError("Terminal symbol used as left hand side", parser.rules[k].lhs_index);

        rules[rule_index].lhs = lhs_image;
        rules[rule_index].separator_token = separator_token;
        rules[rule_index].rhs_index = rhs_sym.Length();
        rules[rule_index].source_index = k;
        rules[rule_index].produces_token = (separator_token != TK_OR_MARKER
                                                             ? separator_token
                                                             : rules[rule_index - 1].produces_token);

        int last_symbol_index = parser.rules[k].separator_index;
        for (int j = parser.rules[k].end_rhs_index - 1;
                 j > parser.rules[k].separator_index;
                 j = lex_stream -> Previous(j))
        {
            if (lex_stream -> Kind(j) == TK_SYMBOL || lex_stream -> Kind(j) == TK_EMPTY_KEY)
            {
                last_symbol_index = j;
                break;
            }
        }

        for (int i = lex_stream -> Next(parser.rules[k].separator_index);
                 i < parser.rules[k].end_rhs_index;
                 i = lex_stream -> Next(i))
        {
            if (lex_stream -> Kind(i) == TK_BLOCK)
            {
                //
                // Check whether or not the rule is a single production
                // and if so, issue an error and stop.
                //
                if (rules[rule_index].IsArrowProduction())
                    lex_stream -> EmitError("This action is associated with a single production", i);

                if ((! option -> attributes) || i > last_symbol_index)
                {
                    BlockSymbol *block = lex_stream -> GetBlockSymbol(i);
                    TextBuffer *buffer = block -> Buffer();

                    ActionBlockElement &action = (option -> automatic_ast != Option::NONE ? ast_actions.Next() : code_actions.Next());
                    action.rule_number = rule_index;
                    action.location = ActionBlockElement::BODY;
                    action.block_token = i;
                    action.buffer = (buffer ? buffer : block -> ActionfileSymbol() -> BodyBuffer());
                }
                else
                {
                    IntToString number(attribute_actions.Length());
                    int length = number.Length() + 1; // +1 for escape character
                    char *name = new char[length + 1]; // +1 for \0
                    name[0] = option -> escape;
                    strcpy(&(name[1]), number.String());

                    VariableSymbol *lhs_symbol = variable_table -> FindName(name, length);
                    if (lhs_symbol)
                    {
                        // TODO: GENERATE ERROR MESSAGE !!!
                        assert(0);
                    }
                    else lhs_symbol = variable_table -> InsertName(name, length);

                    int index = attribute_actions.NextIndex();
                    attribute_actions[index].lhs_image = AssignSymbolIndex(lhs_symbol);
                    attribute_actions[index].block_token = i;

                    rhs_sym.Next() = attribute_actions[index].lhs_image;

                    delete [] name;
                }
            }
            else if (lex_stream -> Kind(i) == TK_MACRO_NAME)
            {
                int image = GetSymbolIndex(i - 1); // preceding symbol
                assert(image != 0);

                //
                // This may occur if the first symbol appearing in a rule was aliased to 
                // %empty.
                //
                if (image == empty)
                    lex_stream -> EmitError("Misplaced rule macro associated with Empty symbol", i);

                //
                // We have two cases to consider here. Either the empty macro, $,
                // was specified or a named macro was specified.
                // 
                // 1. A named macro was specified:
                //    a. If there was a definition already generated automatically for
                //       that symbol, we simply need to override it with the user's
                //       definition.
                // 
                //    b. otherwise, we create a new definition for the macro
                // 
                // 2. The empty macro ($) was specified:
                //    if there was a definition already generated for that symbol,
                //    remove it.
                //
                int position = rhs_sym.Length() - rules[rule_index].rhs_index;
                if (lex_stream -> NameStringLength(i) > 1)
                {
                    if ((option -> variables == Option::BOTH) || 
                        (option -> variables == Option::TERMINALS && IsTerminal(image)) || 
                        (option -> variables == Option::NON_TERMINALS && (! IsTerminal(image)))) // override previous definition
                    {
                        ProcessedRuleElement &processed_rule_element = processed_rule_elements[processed_rule_elements.Length() - 1];
                        processed_rule_element.image = image;
                        processed_rule_element.token_index = i;
                        assert(processed_rule_element.position == position);
                    }
                    else // add new definition
                    {
                        ProcessedRuleElement &processed_rule_element = processed_rule_elements.Next();
                        processed_rule_element.image = image;
                        processed_rule_element.token_index = i;
                        processed_rule_element.position = position;
                    }
                }
                else if ((option -> variables == Option::BOTH) || 
                         (option -> variables == Option::TERMINALS && IsTerminal(image)) || 
                         (option -> variables == Option::NON_TERMINALS && (! IsTerminal(image)))) // remove previous definition
                {
                    assert(processed_rule_elements[processed_rule_elements.Length() - 1].position == position);
                    processed_rule_elements.Reset(processed_rule_elements.Length() - 1);
                }
            }
            else
            {
                VariableSymbol *rhs_symbol = GetSymbol(i);
                int image = (rhs_symbol ? AssignSymbolIndex(rhs_symbol) : 0);
                if (image == 0)
                     lex_stream -> EmitError("This symbol was neither declared nor aliased to a terminal symbol", i);
                else if (image == eof_image)
                     lex_stream -> EmitError("End-of-file symbol cannot be used in a rule", i);
                else if (image != empty)
                {
                    rhs_sym.Next() = image;

                    if ((option -> variables == Option::BOTH) || 
                        (option -> variables == Option::TERMINALS && IsTerminal(image)) || 
                        (option -> variables == Option::NON_TERMINALS && (!IsTerminal(image))))
                    {
                        int position = rhs_sym.Length() - rules[rule_index].rhs_index;
                        ProcessedRuleElement &processed_rule_element = processed_rule_elements.Next();
                        processed_rule_element.image = image;
                        processed_rule_element.token_index = i;
                        processed_rule_element.position = position;
                    }
                }
            }
        }

        if (rules[rule_index].IsArrowProduction() && (rules[rule_index].rhs_index + 1 == rhs_sym.Length()))
            num_single_productions++;

        //
        // If we will be generating Ast classes, we update the classname info.
        //
        if (option -> automatic_ast != Option::NONE)
        {
            int classname_index = parser.rules[k].classname_index,
                array_element_type_index = parser.rules[k].array_element_type_index;
            Symbol *symbol;
            {
                const char *name = (classname_index > 0
                                       ? lex_stream -> NameString(classname_index)
                                       : RetrieveString(rules[rule_index].lhs));
                if (! Code::IsValidVariableName(name))
                    lex_stream -> EmitError("Illegal variable name",
                                            (classname_index > 0 ? classname_index
                                                                 : parser.rules[k].lhs_index));
                if (array_element_type_index != 0 &&
                    lex_stream -> GetVariableSymbol(array_element_type_index) -> NameLength() > 0 &&
                    lex_stream -> NameStringLength(classname_index) == 1) // The classname is the null macro?
                {
                    char *array_element_name = lex_stream -> GetVariableSymbol(array_element_type_index) -> Name();
                    int length = 1 + lex_stream -> GetVariableSymbol(array_element_type_index) -> NameLength() + strlen("List");
                    char *list_name = new char[length + 1];
                         *list_name = option -> escape;
                         strcpy(list_name + 1, array_element_name);
                         strcat(list_name, "List");
                         symbol = classname_set.FindOrInsertName(list_name, length);
                    delete [] list_name;
                }
                else symbol = classname_set.FindOrInsertName(name, strlen(name));
            }

            int index = symbol -> Index();
            if (index == classname.Length())
            {
                ClassnameElement &element = classname.Next();
                element.specified_name = symbol -> Name();
                element.real_name = (*(symbol -> Name()) == option -> escape
                                               ? symbol -> Name() + 1
                                               : symbol -> Name());
                element.array_element_type_symbol = (array_element_type_index == 0
                                                          ? NULL
                                                          : lex_stream -> GetVariableSymbol(array_element_type_index));
            }
            else if (array_element_type_index != 0 &&
                     lex_stream -> GetVariableSymbol(array_element_type_index) -> NameLength() > 0)
            {
                if (classname[index].array_element_type_symbol != lex_stream -> GetVariableSymbol(array_element_type_index))
                    lex_stream -> EmitError("Respecification of an array with a different element type", classname_index);
            }

            //
            // Process rules that are explicitly associated with a non-null classname
            // or are not empty rules or are not single productions with a nonterminal
            // on the right-hand side.
            //
            int rule_size = rhs_sym.Length() - FirstRhsIndex(rule_index);
            if (rule_size > 1 ||
                (classname[index].specified_name != classname[index].real_name) ||
                (rule_size == 1 && IsTerminal(rhs_sym[FirstRhsIndex(rule_index)])))
                 classname[index].rule.Next() = rule_index;
            else classname[index].ungenerated_rule.Next() = rule_index;
        }
    }

    //
    // If we have to generate extra rules because of the presence of
    // of "attributes", we do so here.
    //
    for (int l = 0; l < attribute_actions.Length(); l++)
    {
        int rule_index = rules.NextIndex();
        (void) processed_rule_map.Next(); // allocate an empty macro table for rule_index;
        rules[rule_index].lhs = attribute_actions[l].lhs_image;
        rules[rule_index].separator_token = TK_EQUIVALENCE;
        rules[rule_index].rhs_index = rhs_sym.Length();

        LexStream::TokenIndex block_token = attribute_actions[l].block_token;
        BlockSymbol *block = lex_stream -> GetBlockSymbol(block_token);
        TextBuffer *buffer = block -> Buffer();
        ActionBlockElement &action = code_actions.Next();
        action.rule_number = rule_index;
        action.location = ActionBlockElement::BODY;
        action.block_token = block_token;
        action.buffer = (buffer ? buffer : block -> ActionfileSymbol() -> BodyBuffer());
    }

    //
    // add a gate marker for the last rule.
    //
    rules.Next().rhs_index = rhs_sym.Length();

    //
    // Are there any trailing action blocks following the last rule?
    //
    for (int a = 0; a < parser.trailer_blocks.Length(); a++)
    {
        LexStream::TokenIndex block_token = parser.trailer_blocks[a];
        BlockSymbol *block = lex_stream -> GetBlockSymbol(block_token);
        TextBuffer *buffer = block -> Buffer();

        //
        // Note that trailer actions are associated with rule 0 instead of
        // the last rule (parser.rules.Length()). This is because we do not
        // want to have the variable name macros defined for the last rule
        // to be visible when processing trailer actions. (See ProcessCodeActions).
        //
        ActionBlockElement &action = trailer_actions.Next();
        action.rule_number = 0;
        action.location = ActionBlockElement::FINALIZE;
        action.block_token = block_token;
        action.buffer = (buffer ? buffer : block -> ActionfileSymbol() -> FinalTrailersBuffer());
    }

    num_nonterminals = num_symbols - num_terminals;
    num_rules = 1 /* start rule */ + (parser.rules.Length() - 1) + attribute_actions.Length();
    num_items = rhs_sym.Length() + (num_rules + 1);

    assert(num_rules == rules.Length() - 2);

    //
    // First, insert local macros. Then, process all actions
    //
    rule_number_macro = InsertLocalMacro("rule_number");
    rule_text_macro = InsertLocalMacro("rule_text");
    rule_size_macro = InsertLocalMacro("rule_size");
    input_file_macro = InsertLocalMacro("input_file");
    current_line_macro = InsertLocalMacro("current_line");
    next_line_macro = InsertLocalMacro("next_line");
    symbol_declarations = InsertLocalMacro("symbol_declarations", "");

    (void) InsertLocalMacro("num_rules", num_rules);
    (void) InsertLocalMacro("num_terminals", num_terminals);
    (void) InsertLocalMacro("num_nonterminals", num_nonterminals);
    (void) InsertLocalMacro("num_non_terminals", num_nonterminals); // for upward compatibility with old version
    (void) InsertLocalMacro("num_symbols", num_symbols);
    (void) InsertLocalMacro("template", option -> template_name);
    (void) InsertLocalMacro("file_prefix", option -> file_prefix);
    (void) InsertLocalMacro("package", option -> package);
    (void) InsertLocalMacro("ast_package", option -> ast_package);
    (void) InsertLocalMacro("ast_type", option -> ast_type);
    (void) InsertLocalMacro("exp_type", option -> exp_type);
    (void) InsertLocalMacro("prs_type", option -> prs_type);
    (void) InsertLocalMacro("sym_type", option -> sym_type);
    (void) InsertLocalMacro("action_type", option -> action_type);
    (void) InsertLocalMacro("vititor_type", option -> visitor_type);

    //
    // Process filters macros 
    //
    for (int i = 0; i < lex_stream -> NumFilterMacros(); i++)
        InsertFilterMacro(lex_stream -> FilterMacro(i).macro_name, lex_stream -> FilterMacro(i).macro_value);

    //
    // If automatic generation of the AST nodes is requested, we generate
    // the typestring map. Otherwise, we construct the typestring map based
    // on information provided by the user in the Type section of his grammar.
    // Note that this array has an upper bound of "num_symbols + 1" instead
    // of "num_symbols". This is because the index "num_symbols + 1" will be
    // used to represent the interface associated with Tokens.
    //
    SymbolLookupTable symbol_set;
    this -> ast_token_interface = num_symbols + 1;
    Array<const char *> typestring(ast_token_interface + 1, NULL);
    if (option -> automatic_ast != Option::NONE)
    {
        typestring[0] = option -> ast_type;

        typestring[ast_token_interface] = InsertInterface(symbol_set, ast_token_classname);
        for (int symbol = 1; symbol <= num_terminals; symbol++)
            typestring[symbol] = typestring[ast_token_interface];

        for (int symbol = num_terminals + 1; symbol <= num_symbols; symbol++)
            typestring[symbol] = InsertInterface(symbol_set, RetrieveString(symbol));
    }
    else if (parser.types.Length() > 0)
    {
        const char *terminal_default_type = NULL,
                   *nonterminal_default_type = NULL;
        if (option -> programming_language == Option::C || option -> programming_language == Option::CPP)
             terminal_default_type = nonterminal_default_type = "void *";
        else if (option -> programming_language == Option::JAVA)
        {
             terminal_default_type = "IToken";
             nonterminal_default_type = "Object";
        }
        else lex_stream -> EmitError("Type information can only be processed for C, C++ or Java tables", parser.types[0].type_index);

        typestring[0] = terminal_default_type;
        for (int i = 1; i <= num_terminals; i++)
            typestring[i] = terminal_default_type;

        typestring[accept_image] = nonterminal_default_type;
        for (int i = 0; i < parser.types.Length(); i++)
        {
            VariableSymbol *symbol = GetSymbol(parser.types[i].symbol_index);
            int image = (symbol ? symbol -> SymbolIndex() : 0);
            if (image == 0)
                lex_stream -> EmitError("Undefined nonterminal symbol", parser.types[i].symbol_index);

            if (option -> warnings && typestring[image] != NULL)
                lex_stream -> EmitWarning("Duplicate specification of this type", parser.types[i].symbol_index);

            typestring[image] = lex_stream -> NameString(parser.types[i].type_index);
        }

        Tuple<int> untyped;
        for (int symbol = num_terminals + 1; symbol <= num_symbols; symbol++)
        {
            if (typestring[symbol] == NULL)
            {
                untyped.Next() = symbol;
                typestring[symbol] = nonterminal_default_type;
            }
        }

        if (option -> warnings && untyped.Length() > 0)
        {
            enum { MAX_LENGTH = Control::PRINT_LINE_SIZE - 4 };

            char line[Control::PRINT_LINE_SIZE + 1],
                 tok[Control::SYMBOL_SIZE + 1];
            strcpy(line, "        "); // 8 spaces

            option -> report.Put(untyped.Length() == 1
                                     ? "*** The following nonterminal was not typed: \n\n"
                                     : "*** The following nonterminal(s) are not typed:\n\n");
            for (int i = 0; i < untyped.Length(); i++)
            {
                int symbol = untyped[i];

                RestoreSymbol(tok, RetrieveString(symbol));
                if (strlen(line) + strlen(tok) > Control::PRINT_LINE_SIZE)
                {
                    option -> report.Put(line);
                    option -> report.Put("\n");
                    ReportLargeToken(line, tok, "    ", MAX_LENGTH);
                }
                else
                {
                    strcat(line, tok);
                    strcat(line, " ");
                }
                strcat(line, " ");
            }

            option -> report.Put(line);
            option -> report.Put("\n");
        }
    }

    //
    // Generate the output.
    //
    GenerateDefaultTitle(notice_actions);
    ProcessCodeActions(header_actions, typestring, processed_rule_map);
    if (option -> automatic_ast != Option::NONE)
    {
        assert(classname_set.Size() == classname.Length());
        ProcessAstActions(ast_actions, notice_actions, initial_actions, typestring, processed_rule_map, classname_set, classname);
    }
    else 
    {
        ProcessCodeActions(initial_actions, typestring, processed_rule_map);
        ProcessCodeActions(code_actions, typestring, processed_rule_map);
    }
    ProcessCodeActions(trailer_actions, typestring, processed_rule_map);

    action_blocks -> Flush(); // Print out all action buffers and close action files.

    return;
}


void Grammar::SetName(VariableSymbol *symbol, VariableSymbol *symbol_name, bool negate)
{
    int index = name.NextIndex(),
        image = symbol -> SymbolIndex();
    name[index] = symbol_name -> Name();
    symbol -> SetNameIndex(index);
    symbol_index[image].external_name_index = (negate ? -index : index);

    return;
}


void Grammar::ProcessNames(void)
{
    if (option -> error_maps)
    {
        //
        // We take care of accept first to that he gets assigned #0
        // which we can ignore later.
        //
        SetName(accept_symbol, null_symbol);

        for (int i = 0; i < parser.names.Length(); i++)
        {
            VariableSymbol *symbol = GetSymbol(parser.names[i].lhs_index);
            int image = (symbol ? symbol -> SymbolIndex() : 0);
            if (image == 0)
                lex_stream -> EmitError("Undefined symbol", parser.names[i].lhs_index);
            if (symbol_index[image].external_name_index != OMEGA)
                lex_stream -> EmitError("This symbol has been named more than once", parser.names[i].lhs_index);
            VariableSymbol *rhs_symbol = GetSymbol(parser.names[i].rhs_index);
            assert(rhs_symbol);
            SetName(symbol, rhs_symbol);
        }

        if (error_image == DEFAULT_SYMBOL) // some default, anything!
            symbol_index[DEFAULT_SYMBOL].external_name_index = symbol_index[accept_image].external_name_index;

        for (int j = FirstTerminal(); j <= LastTerminal(); j++)
        {
            if (symbol_index[j].external_name_index == OMEGA)
                SetName(symbol_index[j].symbol, symbol_index[j].symbol);
        }

        for (int k = FirstNonTerminal(); k <= LastNonTerminal(); k++)
        {
            if (symbol_index[k].external_name_index == OMEGA)
            {
                if (option -> names == Option::MAXIMUM)
                     SetName(symbol_index[k].symbol, symbol_index[k].symbol);
                else if (option -> names == Option::OPTIMIZED)
                     SetName(symbol_index[k].symbol, symbol_index[k].symbol, true);
                else
                {
                    assert(option -> names == Option::MINIMUM);
                    symbol_index[k].external_name_index = symbol_index[error_image].external_name_index;
                }
            }
        }
    }

    num_names = name.Length() - 1;

    return;
}


//
//
//
const char *Grammar::SkipMargin(TextBuffer *buffer, const char *cursor, const char *tail)
{
    int length = option -> margin;

    for (int i = 0; length > 0 && cursor < tail; i++, cursor++)
    {
        if (*cursor == Code::HORIZONTAL_TAB)
             length -= (Tab::TabSize() - (i % Tab::TabSize()));
        else if (Code::IsSpaceButNotNewline(*cursor))
             length--;
        else break;
    }

    //
    // If length is positive (length > 0) then we found some valid data
    // in the margin area, in which case, we issue a warning message.
    // If length has a negative value, the absolute value of length
    // indicates that the number of blank characters that should be
    // written to the output file. Note that this situation occurs
    // iff a tab character is used as the last character in a margin
    // space and, when fully expanded, it overlaps the data space.
    //
    if (length < 0)
    {
        for (int j = length; j > 0; j++)
             buffer -> Put(' ');
    }

    return cursor;
}


//
//
//
void Grammar::ProcessActionBlock(ActionBlockElement &action)
{
    BlockSymbol *block = lex_stream -> GetBlockSymbol(action.block_token);
    TextBuffer *buffer = action.buffer;
    int rule_number = action.rule_number,
        line_no = lex_stream -> Line(action.block_token),
        start = lex_stream -> StartLocation(action.block_token) + block -> BlockBeginLength(),
        end   = lex_stream -> EndLocation(action.block_token) - block -> BlockEndLength() + 1;
    const char *head = &(lex_stream -> InputBuffer(action.block_token)[start]),
               *tail = &(lex_stream -> InputBuffer(action.block_token)[end]);

    //
    // If the block opener marker is immediately followed by a newline
    // character, skip the new line.
    //
    if (Code::IsNewline(*head))
    {
        head++;
        line_no++;
    }

    //
    // If the block closer marker is immediately preceded by a newline
    // character, skip the new line.
    //
    if (Code::IsNewline(*(tail - 1)))
        tail--;

    //
    //
    //
    while (head < tail)
    {
        const char *cursor;
        for (cursor = head; Code::IsNewline(*cursor); cursor++)
        {
            buffer -> PutChar(*cursor);
            line_no++;
        }
        cursor = SkipMargin(buffer, cursor, tail);
        for (head = cursor; cursor < tail && (! Code::IsNewline(*cursor)); cursor++)
            ;
        if (cursor > head)
            ProcessActionLine(action.location,
                              buffer,
                              lex_stream -> FileName(action.block_token),
                              head,
                              cursor,
                              line_no,
                              rule_number);
        head = cursor;
    }

    //
    // If the macro spanned only a single line, add a
    // line break to the output.
    //
    if (lex_stream -> Line(action.block_token) == lex_stream -> EndLine(action.block_token))
        buffer -> Put("\n");

    return;
}


//
// Process an action line consisting only of the macro whose name is
// specified by the parameter name. Note that it is assumed that:
//
//  1. The name does not contain the escape prefix which is added here.
//  2. The location of the output is ActionBlockElement::BODY
//  3. The file name associated with this output is the filename of the
//     default block.
//  4. The line number associated with this output is 0.
//
void Grammar::ProcessMacro(TextBuffer *buffer, char *name, int rule_no)
{
    int length = strlen(name) + 1;
    char *macroname = new char[length + 1];
    macroname[0] = option -> escape;
    strcpy(&macroname[1], name);

    ProcessActionLine(ActionBlockElement::BODY,
                      buffer,
                      option -> DefaultBlock() -> ActionfileSymbol() -> Name(),
                      macroname,
                      &macroname[length],
                      0,
                      rule_no);

    delete [] macroname;

    return;
}


//
// PROCESS_ACTION_LINE takes as arguments a line of text from an action
// block and the rule number with which the block is associated.
// It first scans the text for local macro names and then for
// user defined macro names. If one is found, the macro definition is sub-
// stituted for the name. The modified action text is then printed out in
// the action file.
//
void Grammar::ProcessActionLine(int location, TextBuffer *buffer, const char *filename, const char *cursor, const char *tail, int line_no, int rule_no)
{
    assert(buffer);

    const char *start = cursor;
    while(cursor < tail)
    {
        //
        // If not the escape character, just print it !!!
        //
        if (*cursor != option -> escape)
        {
            buffer -> PutChar(*cursor);

            //
            // If the character just output is a new line marker then
            // skip the margin if one is specified.
            //
            if (Code::IsNewline(*cursor++))
            {
                cursor = SkipMargin(buffer, cursor, tail);
                start = cursor;
            }
        }
        else // all macro names begin with ESCAPE
        {
            //
            // Find macro name.
            //
            const char *end_cursor;
            for (end_cursor = cursor + 1;
                 end_cursor < tail && (Code::IsAlnum(*end_cursor) && *end_cursor != option -> escape);
                 end_cursor++)
                ;

            //
            // First, see if the macro is a local macro.
            // Next, check to see if it is an export_symbol macro
            // Finally, check to see if it is a user-defined macro
            //
            SimpleMacroSymbol *simple_macro;
            MacroSymbol *macro;
            if (simple_macro = FindLocalMacro(cursor, end_cursor - cursor))
            {
                char *value = simple_macro -> Value();
                if (value)
                {
                    if (simple_macro == symbol_declarations)
                    {
                        char *p = value;
                        while(*p)
                        {
                            char *q;
                            for (q = p + 1; *q && *q != '\n'; q++)
                                ;

                            ProcessActionLine(location,
                                              buffer,
                                              filename,
                                              p,
                                              q,
                                              line_no,
                                              rule_no);

                            p = q;
                            if (*p != '\0') // if more info to process, add offset
                            {
                                assert(*p == '\n');
                                buffer -> PutChar(*p++);
                                for (int i = 0; i < cursor - start; i++)
                                    buffer -> PutChar(' ');
                            }
                        }
                    }
                    else buffer -> Put(value);
                }
                else if (simple_macro == rule_number_macro)
                    buffer -> Put(rule_no);
                else if (simple_macro == rule_text_macro)
                {
                    if (rule_no == 0)
                        buffer -> Put("*** No Rule ***");
                    else
                    {
                        int index = rule_no - 1; // original rule index starts at 0
                        buffer -> Put(lex_stream -> NameString(parser.rules[index].lhs_index));
                        buffer -> PutChar(' ');
                        buffer -> Put(rules[rule_no].IsArrowProduction() ? "->" : "::=");

                        for (int j = lex_stream -> Next(parser.rules[index].separator_index);
                             j < parser.rules[index].end_rhs_index;
                             j = lex_stream -> Next(j))
                        {
                            if (lex_stream -> Kind(j) == TK_SYMBOL)
                            {
                                buffer -> PutChar(' ');
                                buffer -> Put(lex_stream -> NameString(j));
                            }
                            else if (lex_stream -> Kind(j) == TK_MACRO_NAME)
                                buffer -> Put(lex_stream -> NameString(j));
                            else if (lex_stream -> Kind(j) == TK_EMPTY_KEY)
                            {
                                buffer -> PutChar(' ');
                                buffer -> PutChar(option -> escape);
                                buffer -> Put("Empty");
                            }
                        }
                    }
                }
                else if (simple_macro == rule_size_macro)
                     buffer -> Put(RhsSize(rule_no));
                else if (simple_macro == input_file_macro)
                     buffer -> Put(filename);
                else if (simple_macro == current_line_macro)
                     buffer -> Put(line_no);
                else if (simple_macro == next_line_macro)
                     buffer -> Put(line_no + 1);
                else assert(false);

                cursor = end_cursor + (*end_cursor == option -> escape ? 1 : 0);
            }
            else if (simple_macro = rule_macro_table.FindName(cursor, end_cursor - cursor))
            {
                char *value = simple_macro -> Value();
                assert(value);
                buffer -> Put(value);

                cursor = end_cursor + (*end_cursor == option -> escape ? 1 : 0);
            }
            else if (simple_macro = filter_macro_table.FindName(cursor, end_cursor - cursor))
            {
                char *value = simple_macro -> Value();
                assert(value);
                buffer -> Put(value);

                cursor = end_cursor + (*end_cursor == option -> escape ? 1 : 0);
            }
            else if (simple_macro = export_macro_table.FindName(cursor, end_cursor - cursor))
            {
                buffer -> Put(option -> exp_prefix);
                buffer -> Put(simple_macro -> Name() + 2); // skip initial escape and '_' characters
                buffer -> Put(option -> exp_suffix);
                cursor = end_cursor + (*end_cursor == option -> escape ? 1 : 0);
            }
            else if (macro = macro_table -> FindName(cursor, end_cursor - cursor))
            {
                if (macro -> IsInUse())
                {
                    option -> report.Put("Loop detected during the expansion of the macro \"");
                    option -> report.Put(macro -> Name());
                    option -> report.Put("\"\n");

                    option -> report.Print(Control::syslis);
                    option -> report.Flush(stdout);

                    exit(12);
                }
                int block_token = macro -> Block();
                BlockSymbol *block = lex_stream -> GetBlockSymbol(block_token);
                if (block)
                {
                    int start = lex_stream -> StartLocation(block_token) + block -> BlockBeginLength(),
                        end   = lex_stream -> EndLocation(block_token) - block -> BlockEndLength() + 1;
                    macro -> MarkInUse();

                    //
                    // If the block containing the macro is a "default" block, then its body should
                    // be output in the current file. Otherwise, output it in the file with which
                    // it is associated.
                    //
                    ProcessActionLine(location,
                                      (block == option -> DefaultBlock()
                                             ? buffer
                                             : block -> Buffer()
                                                      ? block -> Buffer()
                                                      : location == ActionBlockElement::INITIALIZE
                                                                 ? block -> ActionfileSymbol() -> InitialHeadersBuffer()
                                                                 : location == ActionBlockElement::BODY
                                                                            ? block -> ActionfileSymbol() -> BodyBuffer()
                                                                            : block -> ActionfileSymbol() -> FinalTrailersBuffer()),
                                      filename,
                                      &(lex_stream -> InputBuffer(block_token)[start]),
                                      &(lex_stream -> InputBuffer(block_token)[end]),
                                      line_no,
                                      rule_no);
                    macro -> MarkNotInUse();

                    cursor = end_cursor + (*end_cursor == option -> escape ? 1 : 0);
                }
                else
                {
                    Tuple <const char *> msg;
                    IntToString line(line_no);
                    msg.Next() = "The macro \"";
                    msg.Next() = macro -> Name();
                    msg.Next() = "\" used in file ";
                    msg.Next() = filename;
                    msg.Next() = " at line ";
                    msg.Next() = line.String();
                    msg.Next() = " is undefined. No susbtitution made";

                    lex_stream -> EmitWarning(msg, block_token);

                    buffer -> PutChar(*cursor);
                    cursor++;
                }
            }
            else
            {
                option -> report.Put("The macro \"");
                for (const char *p = cursor; p < end_cursor; p++)
                    option -> report.PutChar(*p);
                option -> report.Put("\" used in file ");
                option -> report.Put(filename);
                option -> report.Put(" at line ");
                option -> report.Put(line_no);
                option -> report.Put(" is undefined. No susbtitution made\n");

                option -> report.Print(Control::syslis);
                option -> report.Flush(stdout);

                buffer -> PutChar(*cursor);
                cursor++;
            }
        }
    }

    return;
}


//
// This procedure takes two character strings as arguments: IN and OUT.
// IN identifies a grammar symbol or name that is checked as to whether
// or not it needs to be quoted. If so, the necessary quotes are added
// as IN is copied into the space identified by OUT.
// NOTE that it is assumed that IN and OUT do not overlap each other.
//
void Grammar::RestoreSymbol(char *out, char *in)
{
    int  len = strlen(in);
    if (len > 0)
    {
        if ((len == 1 && in[0] == option -> or_marker) ||
            (in[0] == option -> escape)             ||
            (in[0] == '\'')               ||
            (in[len - 1] == '\'')         ||
            (strchr(in, ' ') != NULL &&
            (in[0] != '<' || in[len - 1] != '>')))
        {
            *(out++) = '\'';
            while(*in != '\0')
            {
                if (*in == '\'')
                    *(out++) = *in;
                *(out++) = *(in++);
            }
            *(out++) = '\'';
            *out = '\0';

            return;
        }
    }

    strcpy(out, in);

    return;
}


//
// PRINT_LARGE_TOKEN generates code to print a token that may exceed the
// limit of its field.  The argument are LINE which is the symbol a varying
// length character string, TOKEN which is the symbol to be printed, INDENT
// which is a character string to be used as an initial prefix to indent the
// output line, and LEN which indicates the maximum number of characters that
// can be printed on a given line.  At the end of this process, LINE will
// have the value of the remaining substring that can fit on the output line.
// If a TOKEN is too large to be indented in a line, but not too large for
// the whole line, we forget the indentation, and printed it. Otherwise, it
// is "chapped up" and printed in pieces that are each indented.
//
void Grammar::PrintLargeToken(char *line, char *token, char *indent, int len)
{
    int toklen;

    char temp[Control::SYMBOL_SIZE + 1];

    toklen = strlen(token);

    if (toklen > len && toklen <= Control::PRINT_LINE_SIZE - 1)
    {
        fprintf(Control::syslis, "\n%s", token);
        strcpy(line, indent);
    }
    else
    {
        for (; toklen > len; toklen = strlen(temp))
        {
            memcpy(temp, token, len);
            temp[len] = '\0';
            fprintf(Control::syslis, "\n%s",temp);
            strcpy(temp, token+len + 1);
            token = temp;
        }
        strcpy(line, indent);
        strcat(line, token);
    }

    return;
}


//
// Same as PrintLargeToken above except that output is sent to report buffer.
//
void Grammar::ReportLargeToken(char *line, char *token, char *indent, int len)
{
    int toklen;

    char temp[Control::SYMBOL_SIZE + 1];

    toklen = strlen(token);

    if (toklen > len && toklen <= Control::PRINT_LINE_SIZE - 1)
    {
        option -> report.Put("\n");
        option -> report.Put(token);
        strcpy(line, indent);
    }
    else
    {
        for (; toklen > len; toklen = strlen(temp))
        {
            memcpy(temp, token, len);
            temp[len] = '\0';
            option -> report.Put("\n");
            option -> report.Put(temp);
            strcpy(temp, token + len + 1);
            token = temp;
        }
        strcpy(line, indent);
        strcat(line, token);
    }

    return;
}


void Grammar::ProcessAstActions(Tuple<ActionBlockElement> &actions,
                                Tuple<ActionBlockElement> &notice_actions,
                                Tuple<ActionBlockElement> &initial_actions,
                                Array<const char *> &typestring,
                                Tuple< Tuple<ProcessedRuleElement> > &processed_rule_map,
                                SymbolLookupTable &classname_set,
                                Tuple<ClassnameElement> &classname)
{
    ActionFileLookupTable ast_filename_table(4096);
    TextBuffer &ast_buffer = *(option -> DefaultBlock() -> ActionfileSymbol() -> BodyBuffer());
    Array<RuleAllocationElement> rule_allocation_map(num_rules + 1);

    //
    // Map each rule into the set of action blocks that is associated with it.
    //
    Array< Tuple<ActionBlockElement> > rule_action_map(num_rules + 1);
    for (int i = 0; i < actions.Length(); i++)
    {
        int rule_no = actions[i].rule_number;
        rule_action_map[rule_no].Next() = actions[i];
    }

    //
    // For each nonterminal A, compute the set of rules that A produces.
    //
    BoundedArray< Tuple<int> > global_map(num_terminals + 1, num_symbols);
    for (int rule_no = FirstRule(); rule_no <= LastRule(); rule_no++)
    {
        int lhs = rules[rule_no].lhs;
        global_map[lhs].Next() = rule_no;
    }

    TTC ttc(global_map, processed_rule_map);

    //
    // Compute the interface depences.
    //
    assert(ast_token_interface == num_symbols + 1);
    BoundedArray< Tuple<int> > extension_of(num_terminals + 1, ast_token_interface);
    BoundedArray<BitSetWithOffset> extension_set(num_terminals + 1, ast_token_interface);
    for (int nt = extension_set.Lbound(); nt <= extension_set.Ubound(); nt++)
        extension_set[nt].Initialize(extension_set.Size() + 1, extension_set.Lbound() - 1);

    for (int rule_no = FirstRule(); rule_no <= LastRule(); rule_no++)
    {
        int lhs = rules[rule_no].lhs;

        if (RhsSize(rule_no) == 1)
        {
            if (lhs != accept_image)
            {
                int symbol = rhs_sym[FirstRhsIndex(rule_no)];
                if (IsNonTerminal(symbol))
                {
                    if (! extension_set[symbol][lhs])
                    {
                        int source_index = rules[rule_no].source_index,
                            array_index = parser.rules[source_index].array_element_type_index;

                        //
                        // If the left-hand side is not an array(list) then it
                        // may extend the right-hand side
                        //
                        if (array_index == 0)
                        {
                            extension_set[symbol].AddElement(lhs);
                            extension_of[symbol].Next() = lhs;
                        }
                    }
                }
                else
                {
                    if (! extension_set[lhs][ast_token_interface])
                    {
                        int source_index = rules[rule_no].source_index,
                            array_index = parser.rules[source_index].array_element_type_index;

                        //
                        // If the left-hand side is not an array(list) then it
                        // may extend the right-hand side
                        //
                        if (array_index == 0)
                        {
                            extension_set[lhs].AddElement(ast_token_interface);
                            extension_of[lhs].Next() = ast_token_interface;
                        }
                    }
                }
            }
        }
    }

    LCA lca(extension_of);

    CompleteClassnameInfo(lca, ttc, global_map, typestring, processed_rule_map, classname_set, classname, rule_allocation_map);

    //
    // Compute a map from each interface into the (transitive closure)
    // of the set of classes that can implement in.
    // (CTC: class transitive closure)
    //
    CTC ctc(classname, typestring, extension_of);
    BoundedArray< Tuple<int> > &interface_map = ctc.GetInterfaceMap();

    //
    // (NTC: Nonterminals that can generate null ASTs.
    //
    NTC ntc(global_map, this);

    //
    // First process the root class, the list class, and the Token class.
    //
    {
        const char *list = "List";
        char *ast_list_classname = new char[strlen(option -> ast_type) + strlen(list) + 1];
        strcpy(ast_list_classname, option -> ast_type);
        strcat(ast_list_classname, list);

        if (option -> automatic_ast == Option::NESTED)
        {
            GenerateAstType(ast_buffer, "    ", option -> ast_type);
            GenerateAstListType(ast_buffer, "    ", ast_list_classname);
            GenerateAstTokenType(ntc, ast_buffer, "    ", ast_token_classname);
        }
        else
        {
            assert(option -> automatic_ast == Option::TOPLEVEL);

            ActionFileSymbol *file_symbol = GenerateTitleAndGlobals(ast_filename_table, notice_actions, option -> ast_type, (parser.ast_block != 0));
            GenerateAstType(*file_symbol -> BodyBuffer(), "", option -> ast_type);
            file_symbol -> Flush();

            file_symbol = GenerateTitleAndGlobals(ast_filename_table, notice_actions, ast_list_classname, false);
            GenerateAstListType(*file_symbol -> BodyBuffer(), "", ast_list_classname);
            file_symbol -> Flush();

            file_symbol = GenerateTitleAndGlobals(ast_filename_table, notice_actions, ast_token_classname, false);
            GenerateAstTokenType(ntc, *file_symbol -> BodyBuffer(), "", ast_token_classname);
            file_symbol -> Flush();
        }

        delete [] ast_list_classname;
    }

    //
    // Generate the token interface
    //
    {
        char *ast_token_interfacename = new char[strlen(ast_token_classname) + 2];
        strcpy(ast_token_interfacename, "I");
        strcat(ast_token_interfacename, ast_token_classname);
    
        if (option -> automatic_ast == Option::NESTED)
            GenerateInterface(true /* is token */,
                              ast_buffer,
                              (char *) "    ",
                              ast_token_interfacename,
                              extension_of[ast_token_interface],
                              interface_map[ast_token_interface],
                              classname);
        else 
        {
            ActionFileSymbol *file_symbol = GenerateTitleAndGlobals(ast_filename_table, notice_actions, ast_token_interfacename, false);
            GenerateInterface(true /* is token */,
                              *(file_symbol -> BodyBuffer()),
                              (char *) "",
                              ast_token_interfacename,
                              extension_of[ast_token_interface],
                              interface_map[ast_token_interface],
                              classname);
            file_symbol -> Flush();
        }

        delete [] ast_token_interfacename;
    }

    //
    // Generate the nonterminal interfaces.
    //
    for (int symbol = num_terminals + 1; symbol <= num_symbols; symbol++)
    {
        if (symbol != accept_image)
        {
            char *interface_name = new char[strlen(RetrieveString(symbol)) + 2];
            strcpy(interface_name, "I");
            strcat(interface_name, RetrieveString(symbol));
    
            if (option -> automatic_ast == Option::NESTED)
                GenerateInterface(ctc.IsTerminalClass(symbol),
                                  ast_buffer,
                                  (char *) "    ",
                                  interface_name,
                                  extension_of[symbol],
                                  interface_map[symbol],
                                  classname);
            else 
            {
                ActionFileSymbol *file_symbol = extension_of[symbol].Length() > 0
                                                    ? GenerateTitle(ast_filename_table, notice_actions, interface_name, false)
                                                    : GenerateTitleAndGlobals(ast_filename_table, notice_actions, interface_name, false);
                GenerateInterface(ctc.IsTerminalClass(symbol),
                                  *(file_symbol -> BodyBuffer()),
                                  (char *) "",
                                  interface_name,
                                  extension_of[symbol],
                                  interface_map[symbol],
                                  classname);
                file_symbol -> Flush();
            }

            delete [] interface_name;
        }
    }

    //
    // Generate the rule classes.
    //
    for (int i = 0; i < classname.Length(); i++)
    {
        //
        // No class is generated for rules that are asoociated with the null classname.
        //
        if (IsNullClassname(classname[i]))
            continue;

        //
        // Figure out whether or not, classname[i] needs the environment
        // and process it if it is a List class.
        //
        Tuple<int> &rule = classname[i].rule;
        if (classname[i].array_element_type_symbol != NULL)
        {
            for (int k = 0; k < rule.Length(); k++)
            {
                int rule_no = rule[k];
                classname[i].rhs_type_index.Reset(); // expected by ProcessAstRule
                classname[i].symbol_set.Reset();     // expected by ProcessAstRule

                ProcessAstRule(classname[i], rule_no, processed_rule_map[rule_no]);

                Tuple<ActionBlockElement> &actions = rule_action_map[rule_no];
                classname[i].needs_environment = classname[i].needs_environment || (actions.Length() > 0);
            }
        }
        else
        {
            if (rule.Length() == 1)
            {
                int rule_no = rule[0];
                Tuple<ActionBlockElement> &actions = rule_action_map[rule_no];
                classname[i].needs_environment = (actions.Length() > 0);
            }
            else
            {
                assert(classname[i].specified_name != classname[i].real_name); // a classname was specified?
                for (int k = 0; k < rule.Length(); k++)
                {
                    int rule_no = rule[k];
                    rule_allocation_map[rule_no].name = classname[i].real_name;
                    Tuple<ActionBlockElement> &actions = rule_action_map[rule_no];
                    classname[i].needs_environment = classname[i].needs_environment || (actions.Length() > 0);
                }
            }
        }

        //
        // If the classes are to be generated as top-level classes, we first obtain
        // a file for this class.
        //
        ActionFileSymbol *file_symbol = (option -> automatic_ast == Option::NESTED
                                                                  ? NULL
                                                                  : GenerateTitleAndGlobals(ast_filename_table,
                                                                                            notice_actions,
                                                                                            classname[i].real_name,
                                                                                            classname[i].needs_environment));
        //
        //
        //
        if (classname[i].array_element_type_symbol != NULL)
        {
            //
            // Generate final info for the allocation of rules associated with this class
            //
            for (int k = 0; k < rule.Length(); k++)
            {
                int rule_no = rule[k];
                Tuple<ActionBlockElement> &actions = rule_action_map[rule_no];
                if (file_symbol != NULL) // option -> automatic_ast == Option::TOPLEVEL
                {
                    for (int j = 0; j < actions.Length(); j++)
                        actions[j].buffer = file_symbol -> BodyBuffer();
                }
            }

            //
            // Generate the class
            //
            GenerateListClass(ctc,
                              ntc, 
                              (option -> automatic_ast == Option::NESTED
                                                        ? ast_buffer
                                                        : *(file_symbol -> BodyBuffer())),
                              (option -> automatic_ast == Option::NESTED
                                                        ? (char *) "    "
                                                        : (char *) ""),
                              classname[i],
                              typestring);
            for (int k = 0; k < rule.Length(); k++)
            {
                int rule_no = rule[k];
                rule_allocation_map[rule_no].needs_environment = classname[i].needs_environment;
                Tuple<ActionBlockElement> &actions = rule_action_map[rule_no];
                ProcessCodeActions(actions, typestring, processed_rule_map);
            }
        }
        else
        {
            if (rule.Length() == 1)
            {
                int rule_no = rule[0];
                Tuple<ActionBlockElement> &actions = rule_action_map[rule_no];
                rule_allocation_map[rule_no].needs_environment = classname[i].needs_environment;
                GenerateRuleClass(ctc,
                                  ntc,
                                  (option -> automatic_ast == Option::NESTED
                                                            ? ast_buffer
                                                            : *(file_symbol -> BodyBuffer())),
                                  (option -> automatic_ast == Option::NESTED
                                                            ? (char *) "    "
                                                            : (char *) ""),
                                  classname[i],
                                  typestring);
                if (file_symbol != NULL) // option -> automatic_ast == Option::TOPLEVEL
                {
                    for (int j = 0; j < actions.Length(); j++)
                        actions[j].buffer = file_symbol -> BodyBuffer();
                }
                ProcessCodeActions(actions, typestring, processed_rule_map);
            }
            else
            {
                assert(classname[i].specified_name != classname[i].real_name); // a classname was specified?
                if (file_symbol != NULL) // option -> automatic_ast == Option::TOPLEVEL
                {
                    for (int k = 0; k < rule.Length(); k++)
                    {
                        int rule_no = rule[k];
                        Tuple<ActionBlockElement> &actions = rule_action_map[rule_no];
                        for (int j = 0; j < actions.Length(); j++)
                            actions[j].buffer = file_symbol -> BodyBuffer();
                    }
                }

                if (classname[i].is_terminal_class)
                     GenerateTerminalMergedClass(ntc,
                                                 (option -> automatic_ast == Option::NESTED
                                                                           ? ast_buffer
                                                                           : *(file_symbol -> BodyBuffer())),
                                                 (option -> automatic_ast == Option::NESTED
                                                                           ? (char *) "    "
                                                                           : (char *) ""),
                                                 classname[i],
                                                 typestring);
                else GenerateMergedClass(ctc,
                                         ntc,
                                         (option -> automatic_ast == Option::NESTED
                                                                   ? ast_buffer
                                                                   : *(file_symbol -> BodyBuffer())),
                                         (option -> automatic_ast == Option::NESTED
                                                                   ? (char *) "    "
                                                                   : (char *) ""),
                                         classname[i],
                                         processed_rule_map,
                                         typestring);

                for (int k = 0; k < rule.Length(); k++)
                {
                    int rule_no = rule[k];
                    rule_allocation_map[rule_no].needs_environment = classname[i].needs_environment;
                    Tuple<ActionBlockElement> &actions = rule_action_map[rule_no];
                    ProcessCodeActions(actions, typestring, processed_rule_map);
                }
            }
        }

        if (option -> automatic_ast == Option::NESTED) // Generate Class Closer
             ast_buffer.Put("    }\n\n");
        else
        {
            file_symbol -> BodyBuffer() -> Put("}\n\n");
            file_symbol -> Flush();
        }
    }

    //
    //
    //
    SymbolLookupTable type_set;
    type_set.FindOrInsertName(ast_token_classname, strlen(ast_token_classname));
    for (int i = 0; i < classname.Length(); i++)
    {
        //
        // No class is generated for rules that are asoociated with the null classname.
        //
        if (! IsNullClassname(classname[i]))
            type_set.FindOrInsertName(classname[i].real_name, strlen(classname[i].real_name));
    }

    //
    // Generate the visitor interfaces and Abstract classes that implements
    // the visitors.
    //
    {
        const char *visitor_type = option -> visitor_type,
                   *argument = "Argument",
                   *result = "Result",
                   *abstract = "Abstract";
        char *argument_visitor_type = new char[strlen(argument) + strlen(visitor_type) + 1],
             *result_visitor_type = new char[strlen(result) + strlen(visitor_type) + 1],
             *result_argument_visitor_type = new char[strlen(result) + strlen(argument) + strlen(visitor_type) + 1],
             *abstract_visitor_type = new char[strlen(abstract) + strlen(visitor_type) + 1],
             *abstract_result_visitor_type = new char[strlen(abstract) + strlen(result) + strlen(visitor_type) + 1];

        strcpy(argument_visitor_type, argument);
        strcat(argument_visitor_type, visitor_type);

        strcpy(result_visitor_type, result);
        strcat(result_visitor_type, visitor_type);

        strcpy(result_argument_visitor_type, result);
        strcat(result_argument_visitor_type, argument);
        strcat(result_argument_visitor_type, visitor_type);

        strcpy(abstract_visitor_type, abstract);
        strcat(abstract_visitor_type, visitor_type);

        strcpy(abstract_result_visitor_type, abstract);
        strcat(abstract_result_visitor_type, result);
        strcat(abstract_result_visitor_type, visitor_type);

        if (option -> visitor == Option::DEFAULT)
        {
            if (option -> automatic_ast == Option::NESTED)
            {
                GenerateSimpleVisitorInterface(ast_buffer, "    ", visitor_type, type_set);
                GenerateArgumentVisitorInterface(ast_buffer, "    ", argument_visitor_type, type_set);
                GenerateResultVisitorInterface(ast_buffer, "    ", result_visitor_type, type_set);
                GenerateResultArgumentVisitorInterface(ast_buffer, "    ", result_argument_visitor_type, type_set);

                GenerateNoResultVisitorAbstractClass(ast_buffer, "    ", abstract_visitor_type, type_set);
                GenerateResultVisitorAbstractClass(ast_buffer, "    ", abstract_result_visitor_type, type_set);
            }
            else
            {
                ActionFileSymbol *file_symbol = GenerateTitle(ast_filename_table, notice_actions, visitor_type, false);
                GenerateSimpleVisitorInterface(*file_symbol -> BodyBuffer(), "", visitor_type, type_set);
                file_symbol -> Flush();

                file_symbol = GenerateTitle(ast_filename_table, notice_actions, argument_visitor_type, false);
                GenerateArgumentVisitorInterface(*file_symbol -> BodyBuffer(), "", argument_visitor_type, type_set);
                file_symbol -> Flush();

                file_symbol = GenerateTitle(ast_filename_table, notice_actions, result_visitor_type, false);
                GenerateResultVisitorInterface(*file_symbol -> BodyBuffer(), "", result_visitor_type, type_set);
                file_symbol -> Flush();

                file_symbol = GenerateTitle(ast_filename_table, notice_actions, result_argument_visitor_type, false);
                GenerateResultArgumentVisitorInterface(*file_symbol -> BodyBuffer(), "", result_argument_visitor_type, type_set);
                file_symbol -> Flush();

                file_symbol = GenerateTitle(ast_filename_table, notice_actions, abstract_visitor_type, false);
                GenerateNoResultVisitorAbstractClass(*file_symbol -> BodyBuffer(), "", abstract_visitor_type, type_set);
                file_symbol -> Flush();

                file_symbol = GenerateTitle(ast_filename_table, notice_actions, abstract_result_visitor_type, false);
                GenerateResultVisitorAbstractClass(*file_symbol -> BodyBuffer(), "", abstract_result_visitor_type, type_set);
                file_symbol -> Flush();
            }
        }
        else if (option -> visitor == Option::PREORDER)
        {
            if (option -> automatic_ast == Option::NESTED)
            {
                GeneratePreorderVisitorInterface(ast_buffer, "    ", visitor_type, type_set);
                GeneratePreorderVisitorAbstractClass(ast_buffer, "    ", abstract_visitor_type, type_set);
            }
            else
            {
                ActionFileSymbol *file_symbol = GenerateTitle(ast_filename_table, notice_actions, visitor_type, false);
                GeneratePreorderVisitorInterface(*file_symbol -> BodyBuffer(), "", visitor_type, type_set);
                file_symbol -> Flush();

                file_symbol = GenerateTitle(ast_filename_table, notice_actions, abstract_visitor_type, false);
                GeneratePreorderVisitorAbstractClass(*file_symbol -> BodyBuffer(), "", abstract_visitor_type, type_set);
                file_symbol -> Flush();
            }
        }

        delete [] argument_visitor_type;
        delete [] result_visitor_type;
        delete [] result_argument_visitor_type;
        delete [] abstract_visitor_type;
        delete [] abstract_result_visitor_type;
    }

    ProcessCodeActions(initial_actions, typestring, processed_rule_map);

    int count = 0;
    for (int rule_no = 1; rule_no < rule_allocation_map.Size(); rule_no++)
    {
        if ((rule_allocation_map[rule_no].name != NULL || RhsSize(rule_no) == 0) &&
            (rule_allocation_map[rule_no].list_kind != RuleAllocationElement::COPY_LIST || rule_allocation_map[rule_no].list_position != 1))
        {
            count++;

            //
            // Check whether or not the rule is a single production
            // and if so, issue an error and stop.
            //
            if (rules[rule_no].IsArrowProduction())
            {
                int source_index = rules[rule_no].source_index;
                lex_stream -> EmitError("Unable to generate Ast allocation for single production", parser.rules[source_index].separator_index);
            }

            if (count % option -> max_cases == 0)
            {
                ProcessMacro(&ast_buffer, "SplitActions", rule_no);
                count++;
            }

            ProcessMacro(&ast_buffer, "BeginAction", rule_no);

            if (rule_allocation_map[rule_no].list_kind != RuleAllocationElement::NOT_A_LIST)
            {
                GenerateListAllocation(ctc,
                                       this,
                                       ast_buffer,
                                       rule_no,
                                       rule_allocation_map[rule_no]);
            }
            else
            {
                 if (RhsSize(rule_no) == 0)
                      GenerateNullAstAllocation(this, ast_buffer, rule_no);
                 else GenerateAstAllocation(ctc,
                                            this,
                                            ast_buffer,
                                            rule_allocation_map[rule_no],
                                            processed_rule_map[rule_no],
                                            typestring,
                                            rule_no);
            }

            GenerateCode(&ast_buffer, "\n          ", rule_no);
            ProcessMacro(&ast_buffer, "EndAction", rule_no);
        }
        else
        {
            //
            // Make sure that no action block is associated with a rule for
            // which no class is allocated when it is reduced.
            //
            for (int k = 0; k < rule_action_map[rule_no].Length(); k++)
            {
                ActionBlockElement &action = rule_action_map[rule_no][k];
                lex_stream -> EmitError("Since no class is associated with this production, this code is unreachable",
                                        action.block_token);
            }

            ProcessMacro(&ast_buffer, "NoAction", rule_no);
        }
    }

    return;
}


void Grammar::CheckRecursivenessAndUpdate(Tuple<int> &nonterminal_list,
                                          BoundedArray< Tuple<int> > &nonterminal_map,
                                          Array<RuleAllocationElement> &rule_allocation_map)
{
    enum { NONE, LEFT_RECURSIVE, RIGHT_RECURSIVE, AMBIGUOUS };
    for (int k = 0; k < nonterminal_list.Length(); k++)
    {
        int lhs = nonterminal_list[k];
        Tuple<int> &rule = nonterminal_map[lhs];
        int recursive = NONE;

        for (int i = 0; i < rule.Length(); i++)
        {
            int rule_no = rule[i];
            if (rule_allocation_map[rule_no].list_kind == RuleAllocationElement::ADD_ELEMENT)
            {
                if (rule_allocation_map[rule_no].list_position < rule_allocation_map[rule_no].element_position)
                {
                    if (recursive == NONE)
                         recursive = LEFT_RECURSIVE;
                    else if (recursive != LEFT_RECURSIVE)
                         recursive = AMBIGUOUS;
                }
                else
                {
                    if (recursive == NONE)
                         recursive = RIGHT_RECURSIVE;
                    else if (recursive != RIGHT_RECURSIVE)
                         recursive = AMBIGUOUS;
                }
            }
        }
    
        if (recursive == AMBIGUOUS)
        {
            int source_index = rules[rule[0]].source_index;
            lex_stream -> EmitError("This nonterminal is associated with both left-recursive and right-recursive rule(s)",
                                    parser.rules[source_index].lhs_index);
        }
        else if (recursive != NONE)
        {
            for (int i = 0; i < rule.Length(); i++)
            {
                int rule_no = rule[i];
                if (rule_allocation_map[rule_no].list_kind == RuleAllocationElement::EMPTY)
                    rule_allocation_map[rule_no].list_kind = (recursive == LEFT_RECURSIVE
                                                                  ? RuleAllocationElement::LEFT_RECURSIVE_EMPTY
                                                                  : RuleAllocationElement::RIGHT_RECURSIVE_EMPTY);
                else if (rule_allocation_map[rule_no].list_kind == RuleAllocationElement::SINGLETON)
                    rule_allocation_map[rule_no].list_kind = (recursive == LEFT_RECURSIVE
                                                                  ? RuleAllocationElement::LEFT_RECURSIVE_SINGLETON
                                                                  : RuleAllocationElement::RIGHT_RECURSIVE_SINGLETON);
            }
        }
        else
        {
            assert(recursive == NONE);

            //
            // Nonrecursive rules are tagged as left-recursive.
            //
            for (int i = 0; i < rule.Length(); i++)
            {
                int rule_no = rule[i];
                if (rule_allocation_map[rule_no].list_kind == RuleAllocationElement::EMPTY)
                     rule_allocation_map[rule_no].list_kind = RuleAllocationElement::LEFT_RECURSIVE_EMPTY;
                else if (rule_allocation_map[rule_no].list_kind == RuleAllocationElement::SINGLETON)
                     rule_allocation_map[rule_no].list_kind = RuleAllocationElement::LEFT_RECURSIVE_SINGLETON;
            }
        }
    }

    return;
}


void Grammar::CompleteClassnameInfo(LCA &lca,
                                    TTC &ttc,
                                    BoundedArray< Tuple<int> > &global_map,
                                    Array<const char *> &typestring,
                                    Tuple< Tuple<ProcessedRuleElement> > &processed_rule_map,
                                    SymbolLookupTable &classname_set,
                                    Tuple<ClassnameElement> &classname,
                                    Array<RuleAllocationElement> &rule_allocation_map)
{
    //
    // Process Array classes and Split ambiguous classnames.
    //
    for (int i = 0; i < classname.Length(); i++)
    {
        //
        // No class is generated for rules that are associated with the null classname.
        //
        if (IsNullClassname(classname[i]))
            continue;

        if (classname[i].array_element_type_symbol != NULL)
        {
            Tuple<int> &rule = classname[i].rule;

            //
            // Compute the set of nonterminals that appear on the left-hand
            // side of the set of rules associated with this list.
            //
            Tuple<int> nonterminals;
            BoundedArray< Tuple<int> > nonterminal_map(num_terminals + 1, num_symbols);
            for (int k = 0; k < rule.Length(); k++)
            {
                int rule_no = rule[k],
                    lhs = rules[rule_no].lhs;
                if (nonterminal_map[lhs].Length() == 0)
                    nonterminals.Next() = lhs;
                nonterminal_map[lhs].Next() = rule_no;
            }

            //
            // Compute the set of nonterminals that can only produce this list.
            // In other words, the set of nonterminals such that all its rules
            // are associated with this list.
            //
            BoundedArray<bool> is_list_nonterminal(num_terminals + 1, num_symbols);
            is_list_nonterminal.MemReset();
            for (int k = 0; k < nonterminals.Length(); k++)
            {
                int lhs = nonterminals[k];
                if (global_map[lhs].Length() == nonterminal_map[lhs].Length())
                    is_list_nonterminal[lhs] = true;
                else if (option -> warnings)
                {
                    int rule_no = global_map[lhs][0],
                        source_index = rules[rule_no].source_index;
                    lex_stream -> EmitWarning("Not all rules generated by this nonterminal are associated with this list",
                                              parser.rules[source_index].lhs_index);
                }
            }

            //
            //
            //
            if (classname[i].array_element_type_symbol -> NameLength() == 0)
            {
                //
                //
                //
                int element_type_symbol_index = -1;
                for (int k = 0; k < rule.Length(); k++)
                {
                    int rule_no = rule[k],
                        offset = FirstRhsIndex(rule_no) - 1;
                    Tuple<ProcessedRuleElement> &processed_rule_elements = processed_rule_map[rule_no];
                    rule_allocation_map[rule_no].name = classname[i].real_name;
                    for (int j = 0;  j < processed_rule_elements.Length(); j++)
                    {
                        ProcessedRuleElement &processed_rule_element = processed_rule_elements[j];
                        int symbol = rhs_sym[offset + processed_rule_element.position];
                        if (IsNonTerminal(symbol) && is_list_nonterminal[symbol])
                        {
                            if (rule_allocation_map[rule_no].list_symbol != 0)
                            {
                                int source_index = rules[rule_no].source_index;
                                lex_stream -> EmitError("Multiple list specified on the right-hand side of this rule",
                                                        parser.rules[source_index].classname_index);
                            }
                            rule_allocation_map[rule_no].list_symbol = symbol;
                            rule_allocation_map[rule_no].list_position = processed_rule_element.position;
                        }
                        else
                        {
                            if (rule_allocation_map[rule_no].element_symbol != 0)
                            {
                                int source_index = rules[rule_no].source_index;
                                lex_stream -> EmitError("Multiple element symbols specified on the right-hand side of this rule",
                                                        parser.rules[source_index].array_element_type_index);
                            }
                            rule_allocation_map[rule_no].element_symbol = symbol;
                            rule_allocation_map[rule_no].element_position = processed_rule_element.position;
                        }
                    }

                    //
                    // A rule of the form A ::= ... B ... where B is a list.
                    // Note that the ... does not have to be the empty string.
                    // For example, we might have a rule of the form:
                    // A ::= ( List )
                    //
                    //   or
                    //
                    // A ::= List ;
                    //
                    if (rule_allocation_map[rule_no].list_position == 0 && rule_allocation_map[rule_no].element_position == 0)
                         rule_allocation_map[rule_no].list_kind = RuleAllocationElement::EMPTY;
                    else if (rule_allocation_map[rule_no].list_symbol == 0)
                    {
                        rule_allocation_map[rule_no].list_kind = RuleAllocationElement::SINGLETON;
                        if (element_type_symbol_index == -1)
                            element_type_symbol_index = (IsTerminal(rule_allocation_map[rule_no].element_symbol)
                                                             ? ast_token_interface
                                                             : rule_allocation_map[rule_no].element_symbol);
                        else if (element_type_symbol_index != rule_allocation_map[rule_no].element_symbol)
                        {
                            element_type_symbol_index =
                                     lca.Find(element_type_symbol_index,
                                              (IsTerminal(rule_allocation_map[rule_no].element_symbol)
                                                   ? ast_token_interface
                                                   : rule_allocation_map[rule_no].element_symbol));
                        }
                    }
                    else if (rule_allocation_map[rule_no].element_symbol == 0)
                         rule_allocation_map[rule_no].list_kind = RuleAllocationElement::COPY_LIST;
                    else rule_allocation_map[rule_no].list_kind = RuleAllocationElement::ADD_ELEMENT;
                }

                //
                // Set the array element type for the classname and each RuleAllocationElement.
                //
                if (element_type_symbol_index == -1)
                    element_type_symbol_index = 0; // if not initialized, initialize it to "Ast" index
                CheckRecursivenessAndUpdate(nonterminals, nonterminal_map, rule_allocation_map);
                for (int j = 0; j < nonterminals.Length(); j++)
                {
                    int lhs = nonterminals[j];
                    for (int k = 0; k < nonterminal_map[lhs].Length(); k++)
                    {
                        int rule_no = nonterminal_map[lhs][k];
                        rule_allocation_map[rule_no].element_type_symbol_index = element_type_symbol_index;
                    }
                }
            }
            else
            {
                //
                //
                //
                for (int k = 0; k < rule.Length(); k++)
                {
                    int rule_no = rule[k];
                    rule_allocation_map[rule_no].name = classname[i].real_name;
                    rule_allocation_map[rule_no].element_type_symbol_index = classname[i].array_element_type_symbol -> SymbolIndex();

                    for (int j = FirstRhsIndex(rule_no); j < EndRhsIndex(rule_no); j++)
                    {
                        int symbol = rhs_sym[j];
                        if (IsNonTerminal(symbol) && is_list_nonterminal[symbol])
                        {
                            if (rule_allocation_map[rule_no].list_symbol != 0)
                            {
                                int source_index = rules[rule_no].source_index;
                                lex_stream -> EmitError("Multiple list specified on the right-hand side of this rule",
                                                        parser.rules[source_index].classname_index);
                            }
                            rule_allocation_map[rule_no].list_symbol = symbol;
                            rule_allocation_map[rule_no].list_position = j - FirstRhsIndex(rule_no) + 1;
                        }
                        else if (symbol == rule_allocation_map[rule_no].element_type_symbol_index)
                        {
                            if (rule_allocation_map[rule_no].element_symbol != 0)
                            {
                                int source_index = rules[rule_no].source_index;
                                lex_stream -> EmitError("Multiple element symbols specified on the right-hand side of this rule",
                                                        parser.rules[source_index].array_element_type_index);
                            }
                            rule_allocation_map[rule_no].element_symbol = symbol;
                            rule_allocation_map[rule_no].element_position = j - FirstRhsIndex(rule_no) + 1;
                        }
                    }

                    //
                    // A rule of the form A ::= ... B ... where B is a list.
                    // Note that the ... does not have to be the empty string.
                    // For example, we might have a rule of the form:
                    // A ::= ( List )
                    //
                    //   or
                    //
                    // A ::= List ;
                    //
                    if (rule_allocation_map[rule_no].list_position == 0 && rule_allocation_map[rule_no].element_position == 0)
                         rule_allocation_map[rule_no].list_kind = RuleAllocationElement::EMPTY;
                    else if (rule_allocation_map[rule_no].list_symbol == 0)
                         rule_allocation_map[rule_no].list_kind = RuleAllocationElement::SINGLETON;
                    else if (rule_allocation_map[rule_no].element_symbol == 0)
                         rule_allocation_map[rule_no].list_kind = RuleAllocationElement::COPY_LIST;
                    else rule_allocation_map[rule_no].list_kind = RuleAllocationElement::ADD_ELEMENT;
                }
            }

            //
            // Add the set of left-hand side nonterminals to the list of
            // interfaces for this ArrayList class. Note that the recursive
            // nonterminal must be the first element of the list.
            //
            CheckRecursivenessAndUpdate(nonterminals, nonterminal_map, rule_allocation_map);
            for (int j = 0; j < nonterminals.Length(); j++)
            {
                int lhs = nonterminals[j];
                classname[i].interface.Next() = lhs;
            }
            assert(classname[i].ungenerated_rule.Length() == 0);
        }
        else
        {
            Tuple<int> &rule = classname[i].rule;
            if (rule.Length() == 1)
            {
                int rule_no = rule[0];
                classname[i].interface.Next() = rules[rule_no].lhs;
                ProcessAstRule(classname[i], rule_no, processed_rule_map[rule_no]);
                classname[i].is_terminal_class = (RhsSize(rule_no) == 1 && ttc.IsTerminalSymbol(rhs_sym[FirstRhsIndex(rule_no)]));
                rule_allocation_map[rule_no].name = classname[i].real_name;
                rule_allocation_map[rule_no].is_terminal_class = classname[i].is_terminal_class;
            }
            else if (classname[i].specified_name != classname[i].real_name) // a classname was specified?
            {
                ComputeInterfaces(classname[i], typestring, rule);

                ProcessAstMergedRules(lca, classname[i], rule, processed_rule_map);

                //
                // A merged class is considered a terminal class if it contains
                // only a single attribute that is a terminal attribute. This is
                // possible when each rule associated with the merged class in
                // question is a single production; all the right-hand side 
                // symbols of these single productions are mapped into the same
                // name; and each right-hand side symbol is either a terminal or
                // it's a nonterminal associated only with terminal classes.
                //
//                classname[i].is_terminal_class = (classname[i].rhs_type_index.Length() == 1);
                classname[i].is_terminal_class = true;
                for (int k = 0; classname[i].is_terminal_class && k < rule.Length(); k++)
                {
                    int rule_no = rule[k];
                    classname[i].is_terminal_class = classname[i].is_terminal_class &&
                                                     (RhsSize(rule_no) == 1 && ttc.IsTerminalSymbol(rhs_sym[FirstRhsIndex(rule_no)]));
                }

                for (int k = 0; k < rule.Length(); k++)
                    rule_allocation_map[rule[k]].is_terminal_class = classname[i].is_terminal_class;
            }
            else // generate independent classes.
            {
                for (int k = 0; k < rule.Length(); k++)
                {
                    int rule_no = rule[k];

                    IntToString suffix(k); // Using suffix(k) is more invariant than suffix(rule_no);
                    int length = strlen(classname[i].real_name) + strlen(suffix.String());
                    char *new_name = new char[length + 1];
                    strcpy(new_name, classname[i].real_name);
                    strcat(new_name, suffix.String());

                    ClassnameElement &additional_classname = classname.Next();
                    additional_classname.rule.Next() = rule_no;
                    additional_classname.specified_name = classname_set.FindOrInsertName(new_name, strlen(new_name)) -> Name();
                    additional_classname.real_name = additional_classname.specified_name;

                    delete [] new_name;
                }
                classname[i].rule.Reset();
            }
        }
    }

    return;
}


//
//
//
void Grammar::ProcessAstRule(ClassnameElement &classname, int rule_no, Tuple<ProcessedRuleElement> &processed_rule_elements)
{
    SymbolLookupTable &symbol_set = classname.symbol_set;
    Tuple<int> &rhs_type_index = classname.rhs_type_index;

    assert(symbol_set.Size() == 0);
    assert(rhs_type_index.Length() == 0);

    int offset = FirstRhsIndex(rule_no) - 1;

    for (int i = 0; i < processed_rule_elements.Length(); i++)
    {
        ProcessedRuleElement &processed_rule_element = processed_rule_elements[i];

        VariableSymbol *variable_symbol = GetSymbol(processed_rule_element.token_index);
        int image = (variable_symbol ? variable_symbol -> SymbolIndex() : 0);
        const char *variable_name;
        int length;
        if (image != 0)
        {
            variable_name = RetrieveString(image);
            length = strlen(variable_name);
        }
        else
        {
            variable_name = lex_stream -> NameString(processed_rule_element.token_index);
            length = lex_stream -> NameStringLength(processed_rule_element.token_index);
        }

        if (*variable_name == option -> escape)
        {
            variable_name++;
            length--;
        }
        Symbol *symbol;
        if (! symbol_set.FindName(variable_name, length))
            symbol = symbol_set.InsertName(variable_name, length);
        else
        {
            IntToString suffix(processed_rule_element.position);
            length += strlen(suffix.String());
            char *new_name = new char[length + 1];
            strcpy(new_name, variable_name);
            strcat(new_name, suffix.String());
            symbol = symbol_set.InsertName(new_name, length);

            delete [] new_name;
        }

        assert(rhs_type_index.Length() == symbol -> Index());
        processed_rule_element.symbol_index = symbol -> Index();
        processed_rule_element.type_index = rhs_sym[offset + processed_rule_element.position];
        rhs_type_index.Next() = processed_rule_element.type_index;
    }

    return;
}


void Grammar::ProcessAstMergedRules(LCA &lca, ClassnameElement &element, Tuple<int> &rule, Tuple< Tuple<ProcessedRuleElement> > &processed_rule_map)
{
    SymbolLookupTable &symbol_set = element.symbol_set;
    Tuple<int> &rhs_type_index = element.rhs_type_index;
    assert(symbol_set.Size() == 0);
    assert(rhs_type_index.Length() == 0);

    Tuple<bool> is_nonterminal;
    Tuple<int> token_of_symbol;
    Tuple<int> rule_of_symbol;
    Array< Tuple<int> > rule_map(rule.Length());

    for (int k = 0; k < rule.Length(); k++)
    {
        int rule_no = rule[k],
            offset = FirstRhsIndex(rule_no) - 1;
        Tuple<ProcessedRuleElement> &processed_rule_elements = processed_rule_map[rule_no];

        SymbolLookupTable local_set;
        for (int i = 0; i < processed_rule_elements.Length(); i++) // Insert the macros for the rules associated with this block
        {
            ProcessedRuleElement &processed_rule_element = processed_rule_elements[i];

            VariableSymbol *variable_symbol = GetSymbol(processed_rule_element.token_index);
            int image = (variable_symbol ? variable_symbol -> SymbolIndex() : 0);
            const char *variable_name;
            int length;
            if (image != 0)
            {
                variable_name = RetrieveString(image);
                length = strlen(variable_name);
            }
            else
            {
                variable_name = lex_stream -> NameString(processed_rule_element.token_index);
                length = lex_stream -> NameStringLength(processed_rule_element.token_index);
            }

            if (*variable_name == option -> escape)
            {
                variable_name++;
                length--;
            }

            Symbol *symbol = symbol_set.FindName(variable_name, length);
            if (! symbol)
            {
                (void) local_set.InsertName(variable_name, length);
                symbol = symbol_set.InsertName(variable_name, length);
                assert(rhs_type_index.Length() == symbol -> Index());
                rhs_type_index.Next() = rhs_sym[offset + processed_rule_element.position];
                is_nonterminal.Next() = IsNonTerminal(rhs_sym[offset + processed_rule_element.position]);
                token_of_symbol.Next() = processed_rule_element.token_index - 1;
                rule_of_symbol.Next() = rule_no;
            }
            else if (local_set.FindName(variable_name, length))
            {
                IntToString suffix(processed_rule_element.position);
                length += strlen(suffix.String());
                char *new_name = new char[length + 1];
                strcpy(new_name, variable_name);
                strcat(new_name, suffix.String());
                symbol = symbol_set.FindName(new_name, length);
                if (! symbol)
                {
                    symbol = symbol_set.InsertName(new_name, length);

                    assert(rhs_type_index.Length() == symbol -> Index());
                    rhs_type_index.Next() = rhs_sym[offset + processed_rule_element.position];
                    is_nonterminal.Next() = IsNonTerminal(rhs_sym[offset + processed_rule_element.position]);
                    token_of_symbol.Next() = processed_rule_element.token_index - 1;
                    rule_of_symbol.Next() = rule_no;
                }
                else
                {
                    if (IsNonTerminal(rhs_sym[offset + processed_rule_element.position]) && is_nonterminal[symbol -> Index()])
                        rhs_type_index[symbol -> Index()] = lca.Find(rhs_type_index[symbol -> Index()],
                                                                     rhs_sym[offset + processed_rule_element.position]);
                }
                (void) local_set.InsertName(new_name, length);
            }
            else
            {
                (void) local_set.FindOrInsertName(variable_name, length);
                if (IsNonTerminal(rhs_sym[offset + processed_rule_element.position]) && is_nonterminal[symbol -> Index()])
                    rhs_type_index[symbol -> Index()] = lca.Find(rhs_type_index[symbol -> Index()],
                                                                 rhs_sym[offset + processed_rule_element.position]);
            }

            rule_map[k].Next() = symbol -> Index();
        }
    }

    //
    //
    //
    Array<ProcessedRuleElement> symbol_map(symbol_set.Size());
    for (int i = 0; i < rule.Length(); i++)
    {
        int rule_no = rule[i];
        Tuple<ProcessedRuleElement> &processed_rule_elements = processed_rule_map[rule_no];
        assert(processed_rule_elements.Length() == rule_map[i].Length());
        symbol_map.MemReset();
        for (int j = 0; j < rule_map[i].Length(); j++)
        {
            processed_rule_elements[j].symbol_index = rule_map[i][j];
            processed_rule_elements[j].type_index = rhs_type_index[rule_map[i][j]];

            symbol_map[processed_rule_elements[j].symbol_index] = processed_rule_elements[j];
        }

        //
        // Now, we will reconstruct the processed_rule_elements for rule_no. We
        // also properly set the symbol and type indexes for null arguments.
        //
        processed_rule_elements.Reset();
        for (int j = 0; j < symbol_map.Size(); j++)
        {
            ProcessedRuleElement &processed_rule_element = processed_rule_elements.Next();
            processed_rule_element = symbol_map[j];
            if (processed_rule_element.position == 0)
            {
                processed_rule_element.symbol_index = j;
                processed_rule_element.type_index = rhs_type_index[j];
            }
        }
    }

    return;
}


void Grammar::ProcessCodeActions(Tuple<ActionBlockElement> &actions, Array<const char *> &typestring, Tuple< Tuple<ProcessedRuleElement> > &processed_rule_map)
{
    //
    // We now process all action blocks that are associated with the rules.
    // During this processing the output information is buffered.
    //
    for (int k = 0; k < actions.Length(); k++)
    {
        rule_macro_table.Push(); // prepare to insert macros for this rule
        local_macro_table.Push(); // prepare to insert macros for this rule

        int rule_number = actions[k].rule_number;

        //
        // Map the position of each symbol in the right-hand side of the rule to
        // the type of the symbol in question.
        //
        Array<const char *> rule_type(RhsSize(rule_number) + 1, NULL);
        Array<bool> is_terminal(RhsSize(rule_number) + 1, false);
        if (parser.types.Length() > 0)
        {
            for (int i = FirstRhsIndex(rule_number), k = 1; i < EndRhsIndex(rule_number); i++, k++)
            {
                int image = rhs_sym[i];
                is_terminal[k] = IsTerminal(image);
                rule_type[k] = typestring[image];
            }
        }

        Tuple<ProcessedRuleElement> &processed_rule_elements = processed_rule_map[rule_number];
        Array<SimpleMacroSymbol *> macros(processed_rule_elements.Length());
        int length = 0;
        for (int i = 0; i < processed_rule_elements.Length(); i++) // Insert the macros for the rules associated with this block
        {
            ProcessedRuleElement &processed_rule_element = processed_rule_elements[i];
            if (processed_rule_element.position == 0) // a null argument that is not associated with any symbol
                continue;

            const char *name = lex_stream -> NameString(processed_rule_element.token_index);
            int macro_name_length = strlen(name);
            if (*name != option -> escape)
            {
                if (! Code::IsValidVariableName(name))
                {
                    VariableSymbol *symbol = GetSymbol(processed_rule_element.token_index);
                    int image = (symbol ? symbol -> SymbolIndex() : 0);
                    if (image != 0)
                    {
                        name = RetrieveString(image);
                        macro_name_length = strlen(name);
                    }
                }
                macro_name_length++;
            }

            char *macro_name = new char[macro_name_length + 1];
            if (*name == option -> escape)
                strcpy(macro_name, name);
            else
            {
                macro_name[0] = option -> escape;
                strcpy(&(macro_name[1]), name);
            }

            if (FindLocalMacro(macro_name, macro_name_length) ||
                rule_macro_table.FindName(macro_name, macro_name_length))
                 lex_stream -> EmitError("Illegal respecification of a predefined macro", processed_rule_element.token_index);

            if (! Code::IsValidVariableName(macro_name + 1))
                lex_stream -> EmitError("Invalid use of a symbol name as a macro", processed_rule_element.token_index);

            macros[i] = InsertRuleMacro(macro_name, processed_rule_element.position);

            if (parser.types.Length() > 0 && option -> automatic_ast == Option::NONE)
            {
                if (is_terminal[processed_rule_element.position])
                     length += (2 * strlen(rule_type[processed_rule_element.position]) +
                                macro_name_length +
                                strlen("  = () getRhsIToken();\n") +
                                strlen(macros[i] -> Value())
                                + 1);
                else length += (2 * strlen(rule_type[processed_rule_element.position]) +
                                macro_name_length +
                                strlen("  = () getRhsSym();\n") +
                                strlen(macros[i] -> Value())
                                + 1);
            }
            delete [] macro_name;
        }

        SimpleMacroSymbol *save_symbol_declarations = symbol_declarations;
        char *str = new char[length + 1];
        if (length > 0)
        {
            *str = '\0';
            for (int i = 0; i < processed_rule_elements.Length(); i++) // Insert the macros for the rules associated with this block
            {
                ProcessedRuleElement &processed_rule_element = processed_rule_elements[i];
                SimpleMacroSymbol *macro = macros[i];
                if (macro) // A valid macro?
                {
                    int position = processed_rule_element.position;
                    char *macro_name = macro -> Name() + 1; // +1 to skip escape

                    strcat(str, rule_type[position]);
                    strcat(str, " ");
                    strcat(str, macro_name);
                    strcat(str, " = (");
                    strcat(str, rule_type[position]);
                    strcat(str, ") ");
                    if (is_terminal[position])
                         strcat(str, "getRhsIToken(");
                    else strcat(str, "getRhsSym(");
                    strcat(str, macro -> Value());
                    strcat(str, ");");

                    if (i < processed_rule_elements.Length() - 1)
                        strcat(str, "\n");
                }
            }

            assert(strlen(str) <= (unsigned) length);
            symbol_declarations = InsertLocalMacro("symbol_declarations", str);
        }

        ProcessActionBlock(actions[k]);

        delete [] str;
        symbol_declarations = save_symbol_declarations;

        local_macro_table.Pop(); // remove the macros for the rules associated with this block
        rule_macro_table.Pop(); // remove the macros for the rules associated with this block
    }

    return;
}

//
//
//
void Grammar::DisplayString(const char *name, const char delimiter)
{
    fprintf(Control::syslis, " %c", delimiter);
    for (const char *p = name; *p != '\0'; p++)
    {
        switch(*p)
        {
            case '\b':
                fprintf(Control::syslis, "\\b");
                break;
            case '\t':
                fprintf(Control::syslis, "\\t");
                break;
            case '\n':
                fprintf(Control::syslis, "\\n");
                break;
            case '\f':
                fprintf(Control::syslis, "\\f");
                break;
            case '\r':
                fprintf(Control::syslis, "\\r");
                break;
            case '\\':
                fprintf(Control::syslis, "\\\\");
                break;
            case '\"':
                fprintf(Control::syslis, "\\\"");
                break;
            case '\'':
                fprintf(Control::syslis, "\\\'");
                break;
            default:
                putc(*p, Control::syslis);
                break;
        }
    }
    fprintf(Control::syslis, "%c", delimiter);
}

//
//
//
void Grammar::DisplaySymbol(const char *name)
{
    int length = strlen(name);
    if (strcmp(name, "\"")   == 0 ||
        strcmp(name, " ")    == 0 ||
        strcmp(name, "--")   == 0 ||
        strcmp(name, "|")    == 0 ||
        strcmp(name, "::=")  == 0 ||
        strcmp(name, "::=?") == 0 ||
        strcmp(name, "->")   == 0 ||
        strcmp(name, "->?")  == 0)
         fprintf(Control::syslis, " \'%s\'", name);
    else if (name[0] == '<' && name[length - 1] == '>')
         fprintf(Control::syslis, " %s", name);
    else if (strcmp(name, "\'") == 0) // a string consisting of 1 single quote
         fprintf(Control::syslis, " \"\'\"");
    else if (strpbrk(name, "\"") != NULL)
         DisplayString(name, '\'');
    else if (strpbrk(name, "\b\t\n\f\r\' ") != NULL)
         DisplayString(name, '\"');
    else if (name[0] == option -> escape)   // does name start with escape?
         fprintf(Control::syslis, " \'%s\'", name);
    else fprintf(Control::syslis, " %s", name);

    return;
}

//
//  If a listing is requested, this prints all the macros(if any), followed
// by the aliases(if any), followed by the terminal symbols, followed by the
// rules.
//   This grammar information is printed on lines no longer than
// PRINT_LINE_SIZE characters long.  If all the symbols in a rule cannot fit
// on one line, it is continued on a subsequent line beginning at the
// position after the equivalence symbol (::= or ->) or the middle of the
// print_line, whichever is smaller.  If a symbol cannot fit on a line
// beginning at the proper offset, it is laid out on successive lines,
// beginning at the proper offset.
//
void Grammar::DisplayInput(void)
{
    //
    // First, flush any data left in the report buffer.
    //
    option -> report.Print(Control::syslis);
    option -> report.Flush(stdout);

    fprintf(Control::syslis, "\nPredefined Macros:\n\n");
    for (int i = 0; i < local_macro_table.Size(); i++)
        fprintf(Control::syslis, "    %s\n", local_macro_table[i] -> Name());
    putc('\n', Control::syslis);

    //
    // Print the filter macros, if any.
    //
    if (filter_macro_table.Size() > 0)
    {
        fprintf(Control::syslis, "\nFilter Macros:\n\n");
        for (int i = 0; i < filter_macro_table.Size(); i++)
            fprintf(Control::syslis, "    %c%s\n", option -> escape, filter_macro_table[i] -> Name());
        putc('\n', Control::syslis);
    }

    //
    // Print the Exported symbols, if any.
    //
    if (parser.exports.Length() > 0)
    {
        fprintf(Control::syslis, "\nExported symbols:\n\n");
        for (int i = 0; i < parser.exports.Length(); i++)
        {
            fprintf(Control::syslis, "   ");
            DisplaySymbol(lex_stream -> NameString(parser.exports[i]));
            putc('\n', Control::syslis);
        }
    }

    //
    // Print the Macro definitions, if any.
    //
    if (macro_table -> Size() > 0)
    {
        fprintf(Control::syslis, "\n\nDefined Symbols:\n\n");
        for (int k = 0; k < macro_table -> Size(); k++)
        {
            MacroSymbol *macro = macro_table -> Element(k);
            int block_token = macro -> Block();
            if (block_token) // a defined macro
            {
                fprintf(Control::syslis, "\n\n    %s\n    ", macro -> Name());

                int start = lex_stream -> StartLocation(block_token),
                    end   = lex_stream -> EndLocation(block_token) + 1;
                for (char *head = &(lex_stream -> InputBuffer(block_token)[start]),
                          *tail = &(lex_stream -> InputBuffer(block_token)[end]);
                     head < tail;
                     head++)
                {
                    putc(*head, Control::syslis);
                    if (Code::IsNewline(*head))
                        fprintf(Control::syslis, "    ");
                }
                putc('\n', Control::syslis);
            }
        }
    }

    //
    //   Print the Aliases, if any.
    //
    if (parser.aliases.Length() > 0)
    {
        fprintf(Control::syslis, "\n\n%s\n\n", (parser.aliases.Length() == 1 ? "Alias:" : "Aliases:"));

        for (int i = 0; i < parser.aliases.Length(); i++)
        {
            fprintf( Control::syslis, "   ");
            int start = lex_stream -> StartLocation(parser.aliases[i].lhs_index),
                length = lex_stream -> EndLocation(parser.aliases[i].lhs_index) - start + 1;
            char *name = new char[length + 1];
            strncpy(name, &(lex_stream -> InputBuffer(parser.aliases[i].lhs_index)[start]), length);
            name[length] = '\0';
            DisplaySymbol(name);
            delete [] name;

            fprintf(Control::syslis, " ::=");

            start = lex_stream -> StartLocation(parser.aliases[i].rhs_index);
            length = lex_stream -> EndLocation(parser.aliases[i].rhs_index) - start + 1;
            name = new char[length + 1];
            strncpy(name, &(lex_stream -> InputBuffer(parser.aliases[i].rhs_index)[start]), length);
            name[length] = '\0';
            DisplaySymbol(name);

            putc('\n', Control::syslis);
        }
    }

    //
    // Print special symbols.
    //
    if (parser.start_index != 0)
    {
         fprintf(Control::syslis, "\n\nStart:\n\n");
         fprintf(Control::syslis, "   ");
         DisplaySymbol(RetrieveString(start_image));
    }
    if (parser.identifier_index != 0)
    {
         fprintf(Control::syslis, "\n\nIdentifier:\n\n");
         fprintf(Control::syslis, "   ");
         DisplaySymbol(RetrieveString(identifier_image));
    }
    if (parser.eol_index != 0)
    {
         fprintf(Control::syslis, "\n\nEol:\n\n");
         fprintf(Control::syslis, "   ");
         DisplaySymbol(RetrieveString(eol_image));
    }
    if (parser.eof_index != 0)
    {
         fprintf(Control::syslis, "\n\nEof:\n\n");
         fprintf(Control::syslis, "   ");
         DisplaySymbol(RetrieveString(eof_image));
    }
    if (parser.error_index != 0)
    {
         fprintf(Control::syslis, "\n\nError:\n\n");
         fprintf(Control::syslis, "   ");
         DisplaySymbol(RetrieveString(error_image));
    }

    //
    // Print the terminals.
    //
    fprintf(Control::syslis, "\n\nTerminals:\n\n");
    for (int i = FirstTerminal(); i <= LastTerminal(); i++)
    {
        if (i != empty)
        {
            if (! IsKeyword(i))
            {
                fprintf(Control::syslis, "   ");
                DisplaySymbol(RetrieveString(i));
                putc('\n', Control::syslis);
            }
        }
    }

    fprintf(Control::syslis, "\n\nSoft Keywords:\n\n");
    for (int i = FirstTerminal(); i <= LastTerminal(); i++)
    {
        if (i != empty)
        {
            if (IsKeyword(i))
            {
                fprintf(Control::syslis, "   ");
                DisplaySymbol(RetrieveString(i));
                putc('\n', Control::syslis);
            }
        }
    }

    //
    //    Print the Rules
    //
    fprintf(Control::syslis, "\n\nRules:\n\n");
    int alternate_space = 0;
    for (int rule_no = 0; rule_no < num_rules; rule_no++)
    {
        int rule_index = rule_no + 1;
        fprintf(Control::syslis, "%-4d  ", rule_index);
        if (rules[rule_index].IsAlternateProduction())
        {
            for (int i = 0; i < alternate_space; i++)
                putc(' ', Control::syslis);
            putc(option -> or_marker, Control::syslis);
        }
        else
        {
            const char *lhs_name = lex_stream -> NameString(parser.rules[rule_no].lhs_index);
            alternate_space = strlen(lhs_name) + (rules[rule_index].IsArrowProduction() ? 3 : 4);
            DisplaySymbol(lhs_name);
            int classname_index = parser.rules[rule_no].classname_index,
                array_element_type_index = parser.rules[rule_no].array_element_type_index;
            if (classname_index != 0 && array_element_type_index == 0)
                 fprintf(Control::syslis, "%s", lex_stream -> NameString(classname_index));
            else if (classname_index == 0 && array_element_type_index != 0)
                 fprintf(Control::syslis, "%c%c%s", option -> escape,
                                                  option -> escape,
                                                  lex_stream -> NameString(classname_index));
            else if (classname_index != 0 && array_element_type_index != 0)
                 fprintf(Control::syslis, "%s%c%s", lex_stream -> NameString(classname_index),
                                                  option -> escape,
                                                  lex_stream -> NameString(array_element_type_index));
            else assert (classname_index == 0 && array_element_type_index == 0);

            if (rules[rule_index].IsArrowProduction())
                 fprintf(Control::syslis, " ->");
            else fprintf(Control::syslis, " ::=");

        }

        for (int j = lex_stream -> Next(parser.rules[rule_no].separator_index);
                 j < parser.rules[rule_no].end_rhs_index;
                 j = lex_stream -> Next(j))
        {
            if (lex_stream -> Kind(j) == TK_SYMBOL)
                 DisplaySymbol(lex_stream -> NameString(j));
            else if (lex_stream -> Kind(j) == TK_MACRO_NAME)
                 fprintf(Control::syslis, " %s", lex_stream -> NameString(j));
            else if (lex_stream -> Kind(j) == TK_EMPTY_KEY)
                 fprintf(Control::syslis, " %c%s", option -> escape, "Empty");
        }

        putc('\n', Control::syslis);
    }

    //
    //    Print the Types
    //
    if (parser.types.Length() > 0)
    {
        fprintf(Control::syslis, "\n\nTypes:\n\n");
        for (int k = 0; k < parser.types.Length(); k++)
        {
            fprintf(Control::syslis, "    ");
            if (lex_stream -> Kind(parser.types[k].separator_index) == TK_OR_MARKER)
            {
                for (int i = 0; i < alternate_space; i++)
                    putc(' ', Control::syslis);
                putc(option -> or_marker, Control::syslis);
            }
            else
            {
                alternate_space = lex_stream -> NameStringLength(parser.types[k].type_index) + 3;
                fprintf(Control::syslis, "%s ::=", lex_stream -> NameString(parser.types[k].type_index));
            }

            fprintf(Control::syslis, " %s", lex_stream -> NameString(parser.types[k].symbol_index));
            putc('\n', Control::syslis);
        }
        putc('\n', Control::syslis);
    }

    return;
}


void Grammar::GenerateCode(TextBuffer *ast_buffer, const char *code, int rule_no)
{
    ProcessActionLine(ActionBlockElement::BODY,
                      ast_buffer,
                      option -> DefaultBlock() -> ActionfileSymbol() -> Name(),
                      code,
                      &code[strlen(code)],
                      0,
                      rule_no);
}

//*********************************************************************************
//*
//* TODO: The functions below are placed here only temporarily so that I can debug
//* them. They actually belong to the file Java.cpp. Similar functions need to be
//* written for the other languages, C, C++, ...
//*
//*********************************************************************************

//
//
//
void Grammar::GenerateDefaultTitle(Tuple<ActionBlockElement> &notice_actions)
{
    //
    // If one or more notice blocks were specified, process and
    // print the notice at the beginning of each action file.
    //
    if (notice_actions.Length() > 0)
    {
        for (int i = 0; i < notice_actions.Length(); i++)
            ProcessActionBlock(notice_actions[i]);
        TextBuffer *buffer = notice_actions[0].buffer; // get proper buffer from first action
        buffer -> Put("\n");
        action_blocks -> PutNotice(*buffer);
    }

    //
    // Issue the package state
    //
    TextBuffer *buffer = (option -> DefaultBlock() -> Buffer()
                              ? option -> DefaultBlock() -> Buffer()
                              : option -> DefaultBlock() -> ActionfileSymbol() -> InitialHeadersBuffer());
    if (*option -> package != '\0')
    {
        buffer -> Put("package ");
        buffer -> Put(option -> package);
        buffer -> Put(";\n\n");
    }
    if (option -> automatic_ast &&
        strcmp(option -> package, option -> ast_package) != 0 &&
        *option -> ast_package != '\0')
    {
        buffer -> Put("import ");
        buffer -> Put(option -> ast_package);
        buffer -> Put(".*;\n");
    }

    return;
}


//
// First construct a file for this type. Write the title information to its header
// buffer and return the file.
//
ActionFileSymbol *Grammar::GenerateTitle(ActionFileLookupTable &ast_filename_table,
                                         Tuple<ActionBlockElement> &notice_actions,
                                         const char *type_name,
                                         bool needs_environment)
{
    const char *filetype = (option -> programming_language == Option::JAVA
                                ? ".java"
                                : (option -> programming_language == Option::ML
                                       ? ".ml"
                                       : (option -> programming_language == Option::PLX || option -> programming_language == Option::PLXASM
                                              ? ".copy"
                                              : (option -> programming_language == Option::C || option -> programming_language == Option::CPP
                                                     ? ".h"
                                                     : ".xml"))));
    int filename_length = strlen(option -> ast_directory_prefix) + strlen(type_name) + strlen(filetype);
    char *filename = new char[filename_length + 1];
    strcpy(filename, option -> ast_directory_prefix);
    strcat(filename, type_name);
    strcat(filename, filetype);

    ActionFileSymbol *file_symbol = ast_filename_table.FindOrInsertName(filename, filename_length);
    TextBuffer *buffer = file_symbol -> InitialHeadersBuffer();
    if (notice_actions.Length() > 0)
    {
        //
        // Copy each notice action block, in turn, into a new
        // ActionBLockElement; redirect its output to this buffer
        // and process it.
        //
        for (int i = 0; i < notice_actions.Length(); i++)
        {
            ActionBlockElement action = notice_actions[i];
            action.buffer = buffer;
            ProcessActionBlock(action);
        }
        buffer -> Put("\n");
    }
    if (*option -> ast_package != '\0')
    {
        buffer -> Put("package ");
        buffer -> Put(option -> ast_package);
        buffer -> Put(";\n\n");
    }

    if (needs_environment &&
        strcmp(option -> ast_package, option -> package) != 0 &&
        *option -> package != '\0')
    {
        buffer -> Put("import ");
        buffer -> Put(option -> package);
        buffer -> Put(".*;\n");
    }

    delete [] filename;

    return file_symbol;
}


ActionFileSymbol *Grammar::GenerateTitleAndGlobals(ActionFileLookupTable &ast_filename_table,
                                                   Tuple<ActionBlockElement> &notice_actions,
                                                   const char *type_name,
                                                   bool needs_environment)
{
    ActionFileSymbol *file_symbol = GenerateTitle(ast_filename_table, notice_actions, type_name, needs_environment);
    for (int i = 0; i < parser.global_blocks.Length(); i++)
    {
        ActionBlockElement action;
        action.rule_number = 0;
        action.location = ActionBlockElement::INITIALIZE; // does not matter - block must be default block...
        action.block_token = parser.global_blocks[i];
        action.buffer = file_symbol -> InitialHeadersBuffer();

        ProcessActionBlock(action);
        action.buffer -> Put("\n");
    }

    return file_symbol;
}


//
//
//
void Grammar::GenerateEnvironmentDeclaration(TextBuffer &ast_buffer, const char *indentation)
{
    ast_buffer.Put(indentation); ast_buffer.Put("    private ");
                                 ast_buffer.Put(option -> action_type);
                                 ast_buffer.Put(" environment;\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    public ");
                                 ast_buffer.Put(option -> action_type);
                                 ast_buffer.Put(" getEnvironment() { return environment; }\n\n");
}

//
//
//
void Grammar::GenerateVisitorHeaders(TextBuffer &ast_buffer, const char *indentation, const char *modifiers)
{
    if (option -> visitor != Option::NONE)
    {
        char *header = new char[strlen(indentation) + strlen(modifiers) + 9];
        strcpy(header, indentation);
        strcat(header, modifiers);

        ast_buffer.Put(header);
        ast_buffer.Put("void accept(");
        ast_buffer.Put(option -> visitor_type);
        ast_buffer.Put(" v);");

        if (option -> visitor == Option::DEFAULT)
        {
            ast_buffer.Put("\n");

            ast_buffer.Put(header);
            ast_buffer.Put("void accept(Argument");
            ast_buffer.Put(option -> visitor_type);
            ast_buffer.Put(" v, Object o);\n");

            ast_buffer.Put(header);
            ast_buffer.Put("Object accept(Result");
            ast_buffer.Put(option -> visitor_type);
            ast_buffer.Put(" v);\n");

            ast_buffer.Put(header);
            ast_buffer.Put("Object accept(ResultArgument");
            ast_buffer.Put(option -> visitor_type);
            ast_buffer.Put(" v, Object o);");
        }
        ast_buffer.Put("\n");

        delete [] header;
    }

    return;
}


//
//
//
void Grammar::GenerateVisitorMethods(NTC &ntc,
                                     TextBuffer &ast_buffer,
                                     const char *indentation,
                                     ClassnameElement &element,
                                     BitSet &optimizable_symbol_set)
{
    if (option -> visitor == Option::DEFAULT)
    {
        ast_buffer.Put("\n");
        ast_buffer.Put(indentation); ast_buffer.Put("    public void accept(");
                                     ast_buffer.Put(option -> visitor_type);
                                     ast_buffer.Put(" v) { v.visit(this); }\n");

        ast_buffer.Put(indentation); ast_buffer.Put("    public void accept(Argument");
                                     ast_buffer.Put(option -> visitor_type);
                                     ast_buffer.Put(" v, Object o) { v.visit(this, o); }\n");

        ast_buffer.Put(indentation); ast_buffer.Put("    public Object accept(Result");
                                     ast_buffer.Put(option -> visitor_type);
                                     ast_buffer.Put(" v) { return v.visit(this); }\n");

        ast_buffer.Put(indentation); ast_buffer.Put("    public Object accept(ResultArgument");
                                     ast_buffer.Put(option -> visitor_type);
                                     ast_buffer.Put(" v, Object o) { return v.visit(this, o); }\n");
    }
    else if (option -> visitor == Option::PREORDER)
    {
        ast_buffer.Put("\n");
        ast_buffer.Put(indentation); ast_buffer.Put("    public void accept(");
                                     ast_buffer.Put(option -> visitor_type);
                                     ast_buffer.Put(" v)\n");
        ast_buffer.Put(indentation); ast_buffer.Put("    {\n");
        ast_buffer.Put(indentation); ast_buffer.Put("        v.preVisit(this);\n");
        ast_buffer.Put(indentation); ast_buffer.Put("        enter(v);\n");
        ast_buffer.Put(indentation); ast_buffer.Put("        v.postVisit(this);\n");
        ast_buffer.Put(indentation); ast_buffer.Put("    }\n\n");

        ast_buffer.Put(indentation); ast_buffer.Put("    public void enter(");
                                     ast_buffer.Put(option -> visitor_type);
                                     ast_buffer.Put(" v)\n");
        ast_buffer.Put(indentation); ast_buffer.Put("    {\n");
        SymbolLookupTable &symbol_set = element.symbol_set;
        Tuple<int> &rhs_type_index = element.rhs_type_index;
        if (element.is_terminal_class || symbol_set.Size() == 0)
        {
            ast_buffer.Put(indentation); ast_buffer.Put("        v.visit(this);\n");
        }
        else
        {
            ast_buffer.Put(indentation); ast_buffer.Put("        boolean checkChildren = v.visit(this);\n");
            ast_buffer.Put(indentation); ast_buffer.Put("        if (checkChildren)\n");
            if (symbol_set.Size() > 1)
            {
                ast_buffer.Put(indentation); ast_buffer.Put("        {\n");
            }

            for (int i = 0; i < symbol_set.Size(); i++)
            {
                ast_buffer.Put(indentation); ast_buffer.Put("            ");
                if ((! optimizable_symbol_set[i]) || ntc.CanProduceNullAst(rhs_type_index[i]))
                {
                    ast_buffer.Put("if (_");
                    ast_buffer.Put(symbol_set[i] -> Name());
                    ast_buffer.Put(" != null) ");
                }
                ast_buffer.Put("_");
                ast_buffer.Put(symbol_set[i] -> Name());
                ast_buffer.Put(".accept(v);\n");
            }

            if (symbol_set.Size() > 1)
            {
                ast_buffer.Put(indentation); ast_buffer.Put("        }\n");
            }
        }
        ast_buffer.Put(indentation); ast_buffer.Put("        v.endVisit(this);\n");
        ast_buffer.Put(indentation); ast_buffer.Put("    }\n");
    }

    return;
}


//
//
//
void Grammar::GenerateEqualsMethod(NTC &ntc,
                                   TextBuffer &ast_buffer,
                                   const char *indentation,
                                   ClassnameElement &element,
                                   BitSet &optimizable_symbol_set)
{
    if (! element.is_terminal_class)
    {
        SymbolLookupTable &symbol_set = element.symbol_set;
        Tuple<int> &rhs_type_index = element.rhs_type_index;

        ast_buffer.Put("\n");
        ast_buffer.Put(indentation); ast_buffer.Put("    public boolean equals(Object o)\n");
        ast_buffer.Put(indentation); ast_buffer.Put("    {\n");
        ast_buffer.Put(indentation); ast_buffer.Put("        if (o == this) return true;\n");
        ast_buffer.Put(indentation); ast_buffer.Put("        //\n");
                                     ast_buffer.Put("        // The supers call is not required for now because Ast nodes\n");
                                     ast_buffer.Put("        // can only extend the root Ast, AstToken and AstList and none\n");
                                     ast_buffer.Put("        // of these nodes contain children.\n");
                                     ast_buffer.Put("        //\n");
                                     ast_buffer.Put("        // if (! super.equals(o)) return false;\n");
                                     ast_buffer.Put("        //\n");
        ast_buffer.Put(indentation); ast_buffer.Put("        if (! (o instanceof ");
                                     ast_buffer.Put(element.real_name);
                                     ast_buffer.Put(")) return false;\n");
        if (symbol_set.Size() > 0)
        {
            ast_buffer.Put(indentation); ast_buffer.Put("        ");
                                         ast_buffer.Put(element.real_name);
                                         ast_buffer.Put(" other = (");
                                         ast_buffer.Put(element.real_name);
            ast_buffer.Put(indentation); ast_buffer.Put(") o;\n");

            for (int i = 0; i < symbol_set.Size(); i++)
            {
                ast_buffer.Put(indentation); ast_buffer.Put("        ");
                if ((! optimizable_symbol_set[i]) || ntc.CanProduceNullAst(rhs_type_index[i]))
                {
                                                 ast_buffer.Put("if (_");
                                                 ast_buffer.Put(symbol_set[i] -> Name());
                                                 ast_buffer.Put(" == null && other.get");
                                                 ast_buffer.Put(symbol_set[i] -> Name());
                                                 ast_buffer.Put("() != null) return false;\n");
                    ast_buffer.Put(indentation); ast_buffer.Put("        else ");
                }
                ast_buffer.Put("if (! _");
                ast_buffer.Put(symbol_set[i] -> Name());
                ast_buffer.Put(".equals(other.get");
                ast_buffer.Put(symbol_set[i] -> Name());
                ast_buffer.Put("())) return false;\n");
            }
        }
        ast_buffer.Put(indentation); ast_buffer.Put("        return true;\n");
        ast_buffer.Put(indentation); ast_buffer.Put("    }\n");
    }

    return;
}


//
//
//
void Grammar::GenerateHashcodeMethod(NTC &ntc,
                                     TextBuffer &ast_buffer,
                                     const char *indentation,
                                     ClassnameElement &element,
                                     BitSet &optimizable_symbol_set)
{
    SymbolLookupTable &symbol_set = element.symbol_set;
    Tuple<int> &rhs_type_index = element.rhs_type_index;

    ast_buffer.Put("\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    public int hashCode()\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    {\n");
    if (element.is_terminal_class)
    {
        ast_buffer.Put(indentation); ast_buffer.Put("        return toString().hashCode();\n");
    }
    else
    {
        ast_buffer.Put(indentation); ast_buffer.Put("        int hash = 7;\n");
        for (int i = 0; i < symbol_set.Size(); i++)
        {
            ast_buffer.Put(indentation); ast_buffer.Put("        hash = hash * 31 + (get");
            ast_buffer.Put(symbol_set[i] -> Name());
            if ((! optimizable_symbol_set[i]) || ntc.CanProduceNullAst(rhs_type_index[i]))
            {
                ast_buffer.Put("() == null ? 0 : get");
                ast_buffer.Put(symbol_set[i] -> Name());
            }
            ast_buffer.Put("().hashCode());\n");
        }
        ast_buffer.Put(indentation); ast_buffer.Put("        return hash;\n");
    }
    ast_buffer.Put(indentation); ast_buffer.Put("    }\n");

    return;
}


//
//
//
void Grammar::GenerateSimpleVisitorInterface(TextBuffer &ast_buffer, const char *indentation, const char *interface_name, SymbolLookupTable &type_set)
{
    ast_buffer.Put(indentation); ast_buffer.Put("public interface ");
                                 ast_buffer.Put(interface_name);
                                 ast_buffer.Put("\n");
    ast_buffer.Put(indentation); ast_buffer.Put("{\n");
    for (int i = 0; i < type_set.Size(); i++)
    {
        Symbol *symbol = type_set[i];
        ast_buffer.Put(indentation); ast_buffer.Put("    void visit");
                                     ast_buffer.Put("(");
                                     ast_buffer.Put(symbol -> Name());
                                     ast_buffer.Put(" n);\n");
    }

    ast_buffer.Put(indentation); ast_buffer.Put("}\n");
}

//
//
//
void Grammar::GenerateArgumentVisitorInterface(TextBuffer &ast_buffer, const char *indentation, const char *interface_name, SymbolLookupTable &type_set)
{
    ast_buffer.Put(indentation); ast_buffer.Put("public interface ");
                                 ast_buffer.Put(interface_name);
                                 ast_buffer.Put("\n");
    ast_buffer.Put(indentation); ast_buffer.Put("{\n");
    for (int i = 0; i < type_set.Size(); i++)
    {
        Symbol *symbol = type_set[i];
        ast_buffer.Put(indentation); ast_buffer.Put("    void visit");
                                     ast_buffer.Put("(");
                                     ast_buffer.Put(symbol -> Name());
                                     ast_buffer.Put(" n, Object o);\n");
    }

    ast_buffer.Put(indentation); ast_buffer.Put("}\n");
}

//
//
//
void Grammar::GenerateResultVisitorInterface(TextBuffer &ast_buffer, const char *indentation, const char *interface_name, SymbolLookupTable &type_set)
{
    ast_buffer.Put(indentation); ast_buffer.Put("public interface ");
                                 ast_buffer.Put(interface_name);
                                 ast_buffer.Put("\n");
    ast_buffer.Put(indentation); ast_buffer.Put("{\n");
    for (int i = 0; i < type_set.Size(); i++)
    {
        Symbol *symbol = type_set[i];
        ast_buffer.Put(indentation); ast_buffer.Put("    Object visit");
                                     ast_buffer.Put("(");
                                     ast_buffer.Put(symbol -> Name());
                                     ast_buffer.Put(" n);\n");
    }

    ast_buffer.Put(indentation); ast_buffer.Put("}\n");
}

//
//
//
void Grammar::GenerateResultArgumentVisitorInterface(TextBuffer &ast_buffer,
                                                     const char *indentation,
                                                     const char *interface_name,
                                                     SymbolLookupTable &type_set)
{
    ast_buffer.Put(indentation); ast_buffer.Put("public interface ");
    ast_buffer.Put(interface_name);
    ast_buffer.Put("\n");
    ast_buffer.Put(indentation); ast_buffer.Put("{\n");
    for (int i = 0; i < type_set.Size(); i++)
    {
        Symbol *symbol = type_set[i];
        ast_buffer.Put(indentation); ast_buffer.Put("    Object visit");
                                     ast_buffer.Put("(");
                                     ast_buffer.Put(symbol -> Name());
                                     ast_buffer.Put(" n, Object o);\n");
    }

    ast_buffer.Put(indentation); ast_buffer.Put("}\n");
}


//
//
//
void Grammar::GeneratePreorderVisitorInterface(TextBuffer &ast_buffer,
                                               const char *indentation,
                                               const char *interface_name,
                                               SymbolLookupTable &type_set)
{
    assert(option -> visitor == Option::PREORDER);
    ast_buffer.Put(indentation); ast_buffer.Put("public interface ");
                                 ast_buffer.Put(interface_name);
                                 ast_buffer.Put("\n");
    ast_buffer.Put(indentation); ast_buffer.Put("{\n");

    ast_buffer.Put(indentation); ast_buffer.Put("    void preVisit(");
                                 ast_buffer.Put(option -> ast_type);
                                 ast_buffer.Put(" element);\n\n");

    ast_buffer.Put(indentation); ast_buffer.Put("    void postVisit(");
                                 ast_buffer.Put(option -> ast_type);
                                 ast_buffer.Put(" element);\n\n");
    for (int i = 0; i < type_set.Size(); i++)
    {
        Symbol *symbol = type_set[i];
        ast_buffer.Put(indentation); ast_buffer.Put("    boolean visit");
                                     ast_buffer.Put("(");
                                     ast_buffer.Put(symbol -> Name());
                                     ast_buffer.Put(" n);\n");
        ast_buffer.Put(indentation); ast_buffer.Put("    void endVisit");
                                     ast_buffer.Put("(");
                                     ast_buffer.Put(symbol -> Name());
                                     ast_buffer.Put(" n);\n");
        if (i + 1 < type_set.Size())
            ast_buffer.Put("\n");
    }

    ast_buffer.Put(indentation); ast_buffer.Put("}\n\n");

    return;
}


//
//
//
void Grammar::GenerateNoResultVisitorAbstractClass(TextBuffer &ast_buffer,
                                                   const char *indentation,
                                                   const char *classname,
                                                   SymbolLookupTable &type_set)
{
    ast_buffer.Put(indentation); ast_buffer.Put(option -> automatic_ast == Option::NESTED ? "static " : "");
                                 ast_buffer.Put("public abstract class ");
                                 ast_buffer.Put(classname);
                                 ast_buffer.Put(" implements ");
                                 ast_buffer.Put(option -> visitor_type);
                                 ast_buffer.Put(", Argument");
                                 ast_buffer.Put(option -> visitor_type);
                                 ast_buffer.Put("\n");
    ast_buffer.Put(indentation); ast_buffer.Put("{\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    public abstract void unimplementedVisitor(String s);\n\n");
    for (int i = 0; i < type_set.Size(); i++)
    {
        Symbol *symbol = type_set[i];
        ast_buffer.Put(indentation); ast_buffer.Put("    public void visit");
                                     ast_buffer.Put("(");
                                     ast_buffer.Put(symbol -> Name());
                                     ast_buffer.Put(" n) { unimplementedVisitor(\"visit(");
                                     ast_buffer.Put(symbol -> Name());
                                     ast_buffer.Put(")\"); }\n");
        ast_buffer.Put(indentation); ast_buffer.Put("    public void visit");
                                     ast_buffer.Put("(");
                                     ast_buffer.Put(symbol -> Name());
                                     ast_buffer.Put(" n, Object o) { unimplementedVisitor(\"visit(");
                                     ast_buffer.Put(symbol -> Name());
                                     ast_buffer.Put(", Object)\"); }\n");
        if (i + 1 < type_set.Size())
            ast_buffer.Put("\n");
    }

    ast_buffer.Put(indentation); ast_buffer.Put("}\n");
}

//
//
//
void Grammar::GenerateResultVisitorAbstractClass(TextBuffer &ast_buffer,
                                                 const char *indentation,
                                                 const char *classname,
                                                 SymbolLookupTable &type_set)
{
    ast_buffer.Put(indentation); ast_buffer.Put(option -> automatic_ast == Option::NESTED ? "static " : "");
                                 ast_buffer.Put("public abstract class ");
                                 ast_buffer.Put(classname);
                                 ast_buffer.Put(" implements Result");
                                 ast_buffer.Put(option -> visitor_type);
                                 ast_buffer.Put(", ResultArgument");
                                 ast_buffer.Put(option -> visitor_type);
                                 ast_buffer.Put("\n");
    ast_buffer.Put(indentation); ast_buffer.Put("{\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    public abstract Object unimplementedVisitor(String s);\n\n");
    for (int i = 0; i < type_set.Size(); i++)
    {
        Symbol *symbol = type_set[i];
        ast_buffer.Put(indentation); ast_buffer.Put("    public Object visit(");
                                     ast_buffer.Put(symbol -> Name());
                                     ast_buffer.Put(" n) { return unimplementedVisitor(\"visit(");
                                     ast_buffer.Put(symbol -> Name());
                                     ast_buffer.Put(")\"); }\n");
        ast_buffer.Put(indentation); ast_buffer.Put("    public Object visit(");
                                     ast_buffer.Put(symbol -> Name());
                                     ast_buffer.Put(" n, Object o) { return  unimplementedVisitor(\"visit(");
                                     ast_buffer.Put(symbol -> Name());
                                     ast_buffer.Put(", Object)\"); }\n");
        if (i + 1 < type_set.Size())
            ast_buffer.Put("\n");
    }

    ast_buffer.Put(indentation); ast_buffer.Put("}\n");
}


//
//
//
void Grammar::GeneratePreorderVisitorAbstractClass(TextBuffer &ast_buffer,
                                                   const char *indentation,
                                                   const char *classname,
                                                   SymbolLookupTable &type_set)
{
    assert(option -> visitor == Option::PREORDER);
    ast_buffer.Put(indentation); ast_buffer.Put(option -> automatic_ast == Option::NESTED ? "static " : "");
                                 ast_buffer.Put("public abstract class ");
                                 ast_buffer.Put(classname);
                                 ast_buffer.Put(" implements ");
                                 ast_buffer.Put(option -> visitor_type);
                                 ast_buffer.Put("\n");
    ast_buffer.Put(indentation); ast_buffer.Put("{\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    public abstract void unimplementedVisitor(String s);\n\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    public void preVisit(");
                                 ast_buffer.Put(option -> ast_type);
                                 ast_buffer.Put(" element) {}\n\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    public void postVisit(");
                                 ast_buffer.Put(option -> ast_type);
                                 ast_buffer.Put(" element) {}\n\n");
    for (int i = 0; i < type_set.Size(); i++)
    {
        Symbol *symbol = type_set[i];
        ast_buffer.Put(indentation); ast_buffer.Put("    public boolean visit");
                                     ast_buffer.Put("(");
                                     ast_buffer.Put(symbol -> Name());
                                     ast_buffer.Put(" n) { unimplementedVisitor(\"visit(");
                                     ast_buffer.Put(symbol -> Name());
                                     ast_buffer.Put(")\"); return true; }\n");
        ast_buffer.Put(indentation); ast_buffer.Put("    public void endVisit");
                                     ast_buffer.Put("(");
                                     ast_buffer.Put(symbol -> Name());
                                     ast_buffer.Put(" n) { unimplementedVisitor(\"endVisit(");
                                     ast_buffer.Put(symbol -> Name());
                                     ast_buffer.Put(")\"); }\n");
        if (i + 1 < type_set.Size())
            ast_buffer.Put("\n");
    }

    ast_buffer.Put(indentation); ast_buffer.Put("}\n");

    return;
}


//
// Generate the the Ast root classes
//
void Grammar::GenerateAstType(TextBuffer &ast_buffer,
                              const char *indentation,
                              const char *classname)
{
    /*
     * First, generate the main root class
     */
    ast_buffer.Put(indentation); ast_buffer.Put(option -> automatic_ast == Option::NESTED ? "static " : "");
                                 ast_buffer.Put("public abstract class ");
                                 ast_buffer.Put(classname);
                                 ast_buffer.Put("\n");
    ast_buffer.Put(indentation); ast_buffer.Put("{\n");
    if (option -> parent_saved)
    {
        ast_buffer.Put(indentation); ast_buffer.Put("    protected ");
                                     ast_buffer.Put(classname);
                                     ast_buffer.Put(" parent = null;\n");
    }
    ast_buffer.Put(indentation); ast_buffer.Put("    protected IToken leftIToken,\n");
    ast_buffer.Put(indentation); ast_buffer.Put("                     rightIToken;\n");
    if (option -> parent_saved)
    {
        ast_buffer.Put(indentation); ast_buffer.Put("    protected void");
                                     ast_buffer.Put(" setParent(");
                                     ast_buffer.Put(classname);
                                     ast_buffer.Put(" parent) { this.parent = parent; }\n");
        ast_buffer.Put(indentation); ast_buffer.Put("    public ");
                                     ast_buffer.Put(classname);
                                     ast_buffer.Put(" getParent() { return parent; }\n");\
    }
    ast_buffer.Put(indentation); ast_buffer.Put("    public IToken getLeftIToken() { return leftIToken; }\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    public IToken getRightIToken() { return rightIToken; }\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    public IToken[] getPrecedingAdjuncts() { return leftIToken.getPrecedingAdjuncts(); }\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    public IToken[] getFollowingAdjuncts() { return rightIToken.getPrecedingAdjuncts(); }\n");

    //
    // TODO: This toString function works only if the content of the
    // parse stream came from a single source file. If multiple source
    // files were used then we need to itereate over the sequence of
    // tokens and concatenate them., Perhaps this function should be
    // specified in the PrsStream toString(i, j) where i..j identify
    // a sequence of tokens.
    //
    ast_buffer.Put(indentation); ast_buffer.Put("    public String toString()\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    {\n");
    ast_buffer.Put(indentation); ast_buffer.Put("        PrsStream prsStream = leftIToken.getPrsStream();\n");
    ast_buffer.Put(indentation); ast_buffer.Put("        return new String(prsStream.getInputChars(),\n");
    ast_buffer.Put(indentation); ast_buffer.Put("                          leftIToken.getStartOffset(),\n");
    ast_buffer.Put(indentation); ast_buffer.Put("                          rightIToken.getEndOffset() - leftIToken.getStartOffset() + 1);\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    }\n");

    ast_buffer.Put(indentation); ast_buffer.Put("    public ");
                                 ast_buffer.Put(classname);
                                 ast_buffer.Put("(IToken token) { this.leftIToken = this.rightIToken = token; }\n");

    ast_buffer.Put(indentation); ast_buffer.Put("    public ");
                                 ast_buffer.Put(classname);
                                 ast_buffer.Put("(IToken leftIToken, IToken rightIToken)\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    {\n");
    ast_buffer.Put(indentation); ast_buffer.Put("        this.leftIToken = leftIToken;\n");
    ast_buffer.Put(indentation); ast_buffer.Put("        this.rightIToken = rightIToken;\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    }\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    void initialize() {}\n");
    if (parser.ast_block != 0)
    {
        ActionBlockElement action;
        action.rule_number = 0;
        action.location = ActionBlockElement::INITIALIZE; // does not matter - block must be default block...
        action.block_token = parser.ast_block;
        action.buffer = &ast_buffer;
        ProcessActionBlock(action);
    }

    ast_buffer.Put(indentation); ast_buffer.Put("    /**\n");
                                 ast_buffer.Put("     * Since the Ast type has no children, any two instances of it are equal.\n");
                                 ast_buffer.Put("     */\n");
                                 ast_buffer.Put("    public boolean equals(Object o) { return o instanceof ");
                                 ast_buffer.Put(classname);
                                 ast_buffer.Put("; }\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    public abstract int hashCode();\n");

    GenerateVisitorHeaders(ast_buffer, indentation, "    public abstract ");
    ast_buffer.Put(indentation); ast_buffer.Put("}\n\n");

    return;
}


//
// Generate the the Ast list class
//
void Grammar::GenerateAstListType(TextBuffer &ast_buffer,
                                  const char *indentation,
                                  const char *classname)
{
    /*
     * Generate the List root class
     */
    ast_buffer.Put(indentation); ast_buffer.Put(option -> automatic_ast == Option::NESTED ? "static " : "");
                                 ast_buffer.Put("public abstract class ");
                                 ast_buffer.Put(classname);
                                 ast_buffer.Put(" extends ");
                                 ast_buffer.Put(option -> ast_type);
                                 ast_buffer.Put("\n");
    ast_buffer.Put(indentation); ast_buffer.Put("{\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    private boolean leftRecursive;\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    private java.util.ArrayList list;\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    public int size() { return list.size(); }\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    public ");
                                 ast_buffer.Put(option -> ast_type);
                                 ast_buffer.Put(" getElementAt(int i) { return (");
                                 ast_buffer.Put(option -> ast_type);
                                 ast_buffer.Put(") list.get(leftRecursive ? i : list.size() - 1 - i); }\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    public java.util.ArrayList getArrayList()\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    {\n");
    ast_buffer.Put(indentation); ast_buffer.Put("        if (! leftRecursive) // reverse the list \n");
    ast_buffer.Put(indentation); ast_buffer.Put("        {\n");
    ast_buffer.Put(indentation); ast_buffer.Put("            for (int i = 0, n = list.size() - 1; i < n; i++, n--)\n");
    ast_buffer.Put(indentation); ast_buffer.Put("            {\n");
    ast_buffer.Put(indentation); ast_buffer.Put("                Object ith = list.get(i),\n");
    ast_buffer.Put(indentation); ast_buffer.Put("                       nth = list.get(n);\n");
    ast_buffer.Put(indentation); ast_buffer.Put("                list.set(i, nth);\n");
    ast_buffer.Put(indentation); ast_buffer.Put("                list.set(n, ith);\n");
    ast_buffer.Put(indentation); ast_buffer.Put("            }\n");
    ast_buffer.Put(indentation); ast_buffer.Put("            leftRecursive = true;\n");
    ast_buffer.Put(indentation); ast_buffer.Put("        }\n");
    ast_buffer.Put(indentation); ast_buffer.Put("        return list;\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    }\n");

    ast_buffer.Put(indentation); ast_buffer.Put("    public void add(");
                                 ast_buffer.Put(option -> ast_type);
                                 ast_buffer.Put(" element)\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    {\n");
    ast_buffer.Put(indentation); ast_buffer.Put("        list.add(element);\n");
    ast_buffer.Put(indentation); ast_buffer.Put("        if (leftRecursive)\n");
    ast_buffer.Put(indentation); ast_buffer.Put("             rightIToken = element.getRightIToken();\n");
    ast_buffer.Put(indentation); ast_buffer.Put("        else leftIToken = element.getLeftIToken();\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    }\n\n");

    // generate constructors for list class
    ast_buffer.Put(indentation); ast_buffer.Put("    public ");
                                 ast_buffer.Put(classname);
                                 ast_buffer.Put("(IToken leftIToken, IToken rightIToken, boolean leftRecursive)\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    {\n");
    ast_buffer.Put(indentation); ast_buffer.Put("        super(leftIToken, rightIToken);\n");
    ast_buffer.Put(indentation); ast_buffer.Put("        this.leftRecursive = leftRecursive;\n");
    ast_buffer.Put(indentation); ast_buffer.Put("        list = new java.util.ArrayList();\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    }\n\n");

    ast_buffer.Put(indentation); ast_buffer.Put("    public ");
                                 ast_buffer.Put(classname);
                                 ast_buffer.Put("(");
                                 ast_buffer.Put(option -> ast_type);
                                 ast_buffer.Put(" element, boolean leftRecursive)\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    {\n");
    ast_buffer.Put(indentation); ast_buffer.Put("        this(element.getLeftIToken(), element.getRightIToken(), leftRecursive);\n");
    ast_buffer.Put(indentation); ast_buffer.Put("        list.add(element);\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    }\n");

    ast_buffer.Put(indentation); ast_buffer.Put("\n");
                                 ast_buffer.Put("    public abstract boolean equals(Object o);\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    public abstract int hashCode();\n");
    ast_buffer.Put(indentation); ast_buffer.Put("}\n\n");

    return;
}


//
// Generate the the Ast token class
//
void Grammar::GenerateAstTokenType(NTC &ntc, TextBuffer &ast_buffer,
                                   const char *indentation,
                                   const char *classname)
{
    /*
     * Generate the Token root class
     */
    ast_buffer.Put(indentation); ast_buffer.Put(option -> automatic_ast == Option::NESTED ? "static " : "");
                                 ast_buffer.Put("public class ");
                                 ast_buffer.Put(classname);
                                 ast_buffer.Put(" extends ");
                                 ast_buffer.Put(option -> ast_type);
                                 ast_buffer.Put(" implements I");
                                 ast_buffer.Put(classname);
                                 ast_buffer.Put("\n");
    ast_buffer.Put(indentation); ast_buffer.Put("{\n");

    ast_buffer.Put(indentation); ast_buffer.Put("    public ");
                                 ast_buffer.Put(classname);
                                 ast_buffer.Put("(IToken token) { super(token); }\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    public IToken getIToken() { return leftIToken; }\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    public String toString() { return leftIToken.toString(); }\n");

    ClassnameElement element; // generate a temporary element with no symbols in its symbol set.
    element.real_name = (char *) classname;
    BitSet optimizable_symbol_set(element.symbol_set.Size(), BitSet::UNIVERSE);

    ast_buffer.Put("\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    public boolean equals(Object o)\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    {\n");
    ast_buffer.Put(indentation); ast_buffer.Put("        if (o == this) return true;\n");
    ast_buffer.Put(indentation); ast_buffer.Put("        if (! (o instanceof ");
                                 ast_buffer.Put(classname);
                                 ast_buffer.Put(")) return false;\n");
    ast_buffer.Put(indentation); ast_buffer.Put("        ");
                                 ast_buffer.Put(classname);
                                 ast_buffer.Put(" other = (");
                                 ast_buffer.Put(classname);
    ast_buffer.Put(indentation); ast_buffer.Put(") o;\n");
    ast_buffer.Put(indentation); ast_buffer.Put("        return toString().equals(other.toString());\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    }\n");

    GenerateHashcodeMethod(ntc, ast_buffer, indentation, element, optimizable_symbol_set);
    GenerateVisitorMethods(ntc, ast_buffer, indentation, element, optimizable_symbol_set);

    ast_buffer.Put(indentation); ast_buffer.Put("}\n\n");

    return;
}


//
//
//
void Grammar::GenerateCommentHeader(TextBuffer &ast_buffer,
                                    const char *indentation,
                                    ClassnameElement &element)
{
    const char *rule_info = " *<li>Rule $rule_number:  $rule_text",
               *ungenerated_rule_info = " *<li>Rule $rule_number:  $rule_text";
    ast_buffer.Put(indentation); ast_buffer.Put("/**");
    if (element.ungenerated_rule.Length() > 0)
    {
        ast_buffer.Put("\n");
        ast_buffer.Put(indentation);
        ast_buffer.Put(" *<em>");
        for (int i = 0; i < element.ungenerated_rule.Length(); i++)
        {
            int rule_no = element.ungenerated_rule[i];
            ast_buffer.Put("\n");
            ast_buffer.Put(indentation);
            ProcessActionLine(ActionBlockElement::BODY,
                              &ast_buffer,
                              option -> DefaultBlock() -> ActionfileSymbol() -> Name(),
                              ungenerated_rule_info,
                              &ungenerated_rule_info[strlen(ungenerated_rule_info)],
                              0,
                              rule_no);
        }
        ast_buffer.Put("\n");
        ast_buffer.Put(indentation);
        ast_buffer.Put(" *</em>\n");
        ast_buffer.Put(indentation);
        ast_buffer.Put(" *<p>");
    }
    ast_buffer.Put("\n");
    ast_buffer.Put(indentation);
    ast_buffer.Put(" *<b>");
    for (int i = 0; i < element.rule.Length(); i++)
    {
        int rule_no = element.rule[i];
        ast_buffer.Put("\n");
        ast_buffer.Put(indentation);
        ProcessActionLine(ActionBlockElement::BODY,
                          &ast_buffer,
                          option -> DefaultBlock() -> ActionfileSymbol() -> Name(),
                          rule_info,
                          &rule_info[strlen(rule_info)],
                          0,
                          rule_no);
    }

    ast_buffer.Put("\n");
    ast_buffer.Put(indentation);
    ast_buffer.Put(" *</b>\n");
    ast_buffer.Put(indentation);
    ast_buffer.Put(" */\n");
}


void Grammar::ComputeInterfaces(ClassnameElement &element, Array<const char *> &typestring, Tuple<int> &rule)
{
    SymbolLookupTable interface_set;
    for (int i = 0; i < rule.Length(); i++)
    {
        int rule_no = rule[i];
        const char *name = typestring[rules[rule_no].lhs];
        if (interface_set.FindName(name, strlen(name)) == NULL)
        {
            interface_set.InsertName(name, strlen(name));
            element.interface.Next() = rules[rule_no].lhs;
        }
    }
}


//
//
//
void Grammar::GenerateListClass(CTC &ctc,
                                NTC &ntc,
                                TextBuffer &ast_buffer,
                                const char *indentation,
                                ClassnameElement &element,
                                Array<const char *> &typestring)
{
    Tuple<int> &interface = element.interface;
    assert(element.array_element_type_symbol != NULL);
    const char *classname = element.real_name,
               *element_name = element.array_element_type_symbol -> Name(),
               *element_type = ctc.FindBestTypeFor(element.array_element_type_symbol -> SymbolIndex());

    GenerateCommentHeader(ast_buffer, indentation, element);

    ast_buffer.Put(indentation); ast_buffer.Put(option -> automatic_ast == Option::NESTED ? "static " : "");
                                 ast_buffer.Put("public class ");
                                 ast_buffer.Put(classname);
                                 ast_buffer.Put(" extends ");
                                 ast_buffer.Put(option -> ast_type);
                                 ast_buffer.Put("List implements ");
    for (int i = 0; i < interface.Length() - 1; i++)
    {
        ast_buffer.Put(typestring[element.interface[i]]);
        ast_buffer.Put(", ");
    }
    ast_buffer.Put(typestring[element.interface[interface.Length() - 1]]);
    ast_buffer.Put("\n");
    ast_buffer.Put(indentation); ast_buffer.Put("{\n");

    if (element.needs_environment)
        GenerateEnvironmentDeclaration(ast_buffer, indentation);

    if (ntc.CanProduceNullAst(element.array_element_type_symbol -> SymbolIndex()))
    {
        ast_buffer.Put(indentation); ast_buffer.Put("    /**\n");
        ast_buffer.Put(indentation); ast_buffer.Put("     * The value returned by <b>get");
                                     ast_buffer.Put(element_name);
                                     ast_buffer.Put("At</b> may be <b>null</b>\n");
        ast_buffer.Put(indentation); ast_buffer.Put("     */\n");
    }
    ast_buffer.Put(indentation); ast_buffer.Put("    public ");
                                 ast_buffer.Put(element_type);
                                 ast_buffer.Put(" get");
                                 ast_buffer.Put(element_name);
                                 ast_buffer.Put("At(int i) { return (");
                                 ast_buffer.Put(element_type);
                                 ast_buffer.Put(") getElementAt(i); }\n\n");

    //
    // generate constructors
    //
    ast_buffer.Put(indentation); ast_buffer.Put("    public ");
                                 ast_buffer.Put(classname);
                                 ast_buffer.Put("(");
                                 if (element.needs_environment)
                                 {
                                     ast_buffer.Put(option -> action_type);
                                     ast_buffer.Put(" environment, ");
                                 }
                                 ast_buffer.Put("IToken leftIToken, IToken rightIToken, boolean leftRecursive)\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    {\n");
    ast_buffer.Put(indentation); ast_buffer.Put("        super(leftIToken, rightIToken, leftRecursive);\n");
    if (element.needs_environment)
    {
        ast_buffer.Put(indentation);
        ast_buffer.Put("        this.environment = environment;\n");
    }
    ast_buffer.Put(indentation); ast_buffer.Put("        initialize();\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    }\n\n");

    ast_buffer.Put(indentation); ast_buffer.Put("    public ");
                                 ast_buffer.Put(classname);
                                 ast_buffer.Put("(");
                                 if (element.needs_environment)
                                 {
                                     ast_buffer.Put(option -> action_type);
                                     ast_buffer.Put(" environment, ");
                                 }
                                 ast_buffer.Put(element_type);
                                 ast_buffer.Put(" _");
                                 ast_buffer.Put(element_name);
                                 ast_buffer.Put(", boolean leftRecursive)\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    {\n");
    ast_buffer.Put(indentation); ast_buffer.Put("        super((");
                                 ast_buffer.Put(option -> ast_type);
                                 ast_buffer.Put(") _");
                                 ast_buffer.Put(element_name);
                                 ast_buffer.Put(", leftRecursive);\n");
    if (option -> parent_saved)
    {
        ast_buffer.Put(indentation); ast_buffer.Put("        ");
        if (ntc.CanProduceNullAst(element.array_element_type_symbol -> SymbolIndex()))
        {
            ast_buffer.Put("if (_");
            ast_buffer.Put(element_name);
            ast_buffer.Put(" != null) ");
        }
        ast_buffer.Put("((");
        ast_buffer.Put(option -> ast_type);
        ast_buffer.Put(") _");
        ast_buffer.Put(element_name);
        ast_buffer.Put(").setParent(this);\n");
    }
    if (element.needs_environment)
    {
        ast_buffer.Put(indentation);
        ast_buffer.Put("        this.environment = environment;\n");
    }
    ast_buffer.Put(indentation); ast_buffer.Put("        initialize();\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    }\n");
    ast_buffer.Put("\n");

    ast_buffer.Put(indentation); ast_buffer.Put("    public ");
                                 ast_buffer.Put("void add(");
                                 ast_buffer.Put(element_type);
                                 ast_buffer.Put(" _");
                                 ast_buffer.Put(element_name);
                                 ast_buffer.Put(")\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    {\n");
    ast_buffer.Put(indentation); ast_buffer.Put("        super.add((");
                                 ast_buffer.Put(option -> ast_type);
                                 ast_buffer.Put(") _");
                                 ast_buffer.Put(element_name);
                                 ast_buffer.Put(");\n");
    if (option -> parent_saved)
    {
        ast_buffer.Put(indentation); ast_buffer.Put("        ");
        if (ntc.CanProduceNullAst(element.array_element_type_symbol -> SymbolIndex()))
        {
            ast_buffer.Put("if (_");
            ast_buffer.Put(element_name);
            ast_buffer.Put(" != null) ");
        }
        ast_buffer.Put("((");
        ast_buffer.Put(option -> ast_type);
        ast_buffer.Put(") _");
        ast_buffer.Put(element_name);
        ast_buffer.Put(").setParent(this);\n");
    }
    ast_buffer.Put(indentation); ast_buffer.Put("    }\n");

    //
    // Generate the "equals" method
    //
    ast_buffer.Put("\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    public boolean equals(Object o)\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    {\n");
    ast_buffer.Put(indentation); ast_buffer.Put("        if (o == this) return true;\n");
    ast_buffer.Put(indentation); ast_buffer.Put("        if (! (o instanceof ");
                                 ast_buffer.Put(classname);
                                 ast_buffer.Put(")) return false;\n");
    ast_buffer.Put(indentation); ast_buffer.Put("        ");
                                 ast_buffer.Put(classname);
                                 ast_buffer.Put(" other = (");
                                 ast_buffer.Put(classname);
    ast_buffer.Put(indentation); ast_buffer.Put(") o;\n");
    ast_buffer.Put(indentation); ast_buffer.Put("        if (size() != other.size()) return false;\n");
    ast_buffer.Put(indentation); ast_buffer.Put("        for (int i = 0; i < size(); i++)\n");
    ast_buffer.Put(indentation); ast_buffer.Put("        {\n");
    ast_buffer.Put(indentation); ast_buffer.Put("            ");
    char *element_typename = ctc.FindUniqueTypeFor(element.array_element_type_symbol -> SymbolIndex());
    if (element_typename != NULL)
         ast_buffer.Put(element_typename);
    else ast_buffer.Put(typestring[element.array_element_type_symbol -> SymbolIndex()]);
    ast_buffer.Put(" element = get");
    ast_buffer.Put(element_name);
    ast_buffer.Put("At(i);\n");
    ast_buffer.Put(indentation); ast_buffer.Put("            ");
    if (ntc.CanProduceNullAst(element.array_element_type_symbol -> SymbolIndex()))
    {
        ast_buffer.Put("if (element == null && other.get");
                                     ast_buffer.Put(element_name);
                                     ast_buffer.Put("At(i) != null) return false;\n");
        ast_buffer.Put(indentation); ast_buffer.Put("            else ");
    }
    ast_buffer.Put(indentation); ast_buffer.Put("if (! element.equals(other.get");
                                 ast_buffer.Put(element_name);
                                 ast_buffer.Put("At(i))) return false;\n");
    ast_buffer.Put(indentation); ast_buffer.Put("        }\n");
    ast_buffer.Put(indentation); ast_buffer.Put("        return true;\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    }\n");

    //
    // Generate the "hashCode" method
    //
    ast_buffer.Put("\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    public int hashCode()\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    {\n");
    ast_buffer.Put(indentation); ast_buffer.Put("        int hash = 7;\n");
    ast_buffer.Put(indentation); ast_buffer.Put("        for (int i = 0; i < size(); i++)\n");
    ast_buffer.Put(indentation); ast_buffer.Put("            hash = hash * 31 + (get");
                                 ast_buffer.Put(element_name);
    if (ntc.CanProduceNullAst(element.array_element_type_symbol -> SymbolIndex()))
    {
        ast_buffer.Put("At(i) == null ? 0 : get");
        ast_buffer.Put(element_name);
    }
    ast_buffer.Put("At(i).hashCode());\n");
    ast_buffer.Put(indentation); ast_buffer.Put("        return hash;\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    }\n");

    //
    // Generate visitor methods.
    //
    if (option -> visitor == Option::DEFAULT)
    {
        ast_buffer.Put("\n");
        ast_buffer.Put(indentation); ast_buffer.Put("    public void accept(");
                                     ast_buffer.Put(option -> visitor_type);
        if (ctc.FindUniqueTypeFor(element.array_element_type_symbol -> SymbolIndex()) != NULL)
        {
            ast_buffer.Put(" v) { for (int i = 0; i < size(); i++) v.visit"
                           "("
                           "get");
            ast_buffer.Put(element_name);
            ast_buffer.Put("At(i)"
                           "); }\n");
        }
        else
        {
            ast_buffer.Put(" v) { for (int i = 0; i < size(); i++) get");
            ast_buffer.Put(element_name);
            ast_buffer.Put("At(i).accept(v); }\n");
        }

        ast_buffer.Put(indentation); ast_buffer.Put("    public void accept(Argument");
                                     ast_buffer.Put(option -> visitor_type);
        if (ctc.FindUniqueTypeFor(element.array_element_type_symbol -> SymbolIndex()) != NULL)
        {
            ast_buffer.Put(" v, Object o) { for (int i = 0; i < size(); i++) v.visit"
                           "("
                           "get");
            ast_buffer.Put(element_name);
            ast_buffer.Put("At(i), o");
            ast_buffer.Put("); }\n");
        }
        else
        {
            ast_buffer.Put(" v, Object o) { for (int i = 0; i < size(); i++) get");
            ast_buffer.Put(element_name);
            ast_buffer.Put("At(i).accept(v, o); }\n");
        }

        //
        // Code cannot be generated to automatically visit a node that
        // can return a value. These cases are left up to the user.
        //
        ast_buffer.Put(indentation); ast_buffer.Put("    public Object accept(Result");
                                     ast_buffer.Put(option -> visitor_type);
        if (ctc.FindUniqueTypeFor(element.array_element_type_symbol -> SymbolIndex()) != NULL)
        {
            ast_buffer.Put(" v) { for (int i = 0; i < size(); i++) v.visit"
                           "("
                           "get");
            ast_buffer.Put(element_name);
            ast_buffer.Put("At(i)"
                           "); return null; }\n");
        }
        else
        {
            ast_buffer.Put(" v) { for (int i = 0; i < size(); i++) get");
            ast_buffer.Put(element_name);
            ast_buffer.Put("At(i).accept(v); return null; }\n");
        }

        ast_buffer.Put(indentation); ast_buffer.Put("    public Object accept(ResultArgument");
                                     ast_buffer.Put(option -> visitor_type);
        if (ctc.FindUniqueTypeFor(element.array_element_type_symbol -> SymbolIndex()) != NULL)
        {
            ast_buffer.Put(" v, Object o) { for (int i = 0; i < size(); i++) v.visit"
                           "("
                           "get");
            ast_buffer.Put(element_name);
            ast_buffer.Put("At(i), o"
                           "); return null; }\n");
        }
        else
        {
            ast_buffer.Put(" v, Object o) { for (int i = 0; i < size(); i++) get");
            ast_buffer.Put(element_name);
            ast_buffer.Put("At(i).accept(v, o); return null; }\n");
        }
    }
    else if (option -> visitor == Option::PREORDER)
    {
        ast_buffer.Put("\n");
        ast_buffer.Put(indentation); ast_buffer.Put("    public void accept(");
                                     ast_buffer.Put(option -> visitor_type);
                                     ast_buffer.Put(" v)\n");
        ast_buffer.Put(indentation); ast_buffer.Put("    {\n");
        ast_buffer.Put(indentation); ast_buffer.Put("        v.preVisit(this);\n");
        ast_buffer.Put(indentation); ast_buffer.Put("        enter(v);\n");
        ast_buffer.Put(indentation); ast_buffer.Put("        v.postVisit(this);\n");
        ast_buffer.Put(indentation); ast_buffer.Put("    }\n");
        ast_buffer.Put(indentation); ast_buffer.Put("    public void enter(");
                                     ast_buffer.Put(option -> visitor_type);
                                     ast_buffer.Put(" v)\n");
        ast_buffer.Put(indentation); ast_buffer.Put("    {\n");
        ast_buffer.Put(indentation); ast_buffer.Put("        boolean checkChildren = v.visit(this);\n");
        ast_buffer.Put(indentation); ast_buffer.Put("        if (checkChildren)\n");
        ast_buffer.Put(indentation); ast_buffer.Put("        {\n");
        ast_buffer.Put(indentation); ast_buffer.Put("            for (int i = 0; i < size(); i++)\n");
        ast_buffer.Put(indentation); ast_buffer.Put("            {\n");
        ast_buffer.Put(indentation); ast_buffer.Put("                ");

        char *element_typename = ctc.FindUniqueTypeFor(element.array_element_type_symbol -> SymbolIndex());
        if (element_typename != NULL)
        {
            ast_buffer.Put(element_typename);
            ast_buffer.Put(" element = get");
            ast_buffer.Put(element_name);
            ast_buffer.Put("At(i);\n");
            if (ntc.CanProduceNullAst(element.array_element_type_symbol -> SymbolIndex()))
            {
                ast_buffer.Put(indentation); ast_buffer.Put("                if (element != null)");
                ast_buffer.Put(indentation); ast_buffer.Put("                {\n");
                ast_buffer.Put(indentation); ast_buffer.Put("                    v.preVisit(element);\n");
                ast_buffer.Put(indentation); ast_buffer.Put("                    element.enter(v);\n");
                ast_buffer.Put(indentation); ast_buffer.Put("                    v.postVisit(element);\n");
                ast_buffer.Put(indentation); ast_buffer.Put("                }\n");
            }
            else
            {
                ast_buffer.Put(indentation); ast_buffer.Put("                v.preVisit(element);\n");
                ast_buffer.Put(indentation); ast_buffer.Put("                element.enter(v);\n");
                ast_buffer.Put(indentation); ast_buffer.Put("                v.postVisit(element);\n");
            }
        }
        else
        {
            ast_buffer.Put(typestring[element.array_element_type_symbol -> SymbolIndex()]);
            ast_buffer.Put(" element = get");
            ast_buffer.Put(element_name);
            ast_buffer.Put("At(i);\n");
            ast_buffer.Put("                ");
            if (ntc.CanProduceNullAst(element.array_element_type_symbol -> SymbolIndex()))
                ast_buffer.Put("if (element != null) ");
            ast_buffer.Put("element.accept(v);\n");
        }
        ast_buffer.Put(indentation); ast_buffer.Put("            }\n");
        ast_buffer.Put(indentation); ast_buffer.Put("        }\n");
        ast_buffer.Put(indentation); ast_buffer.Put("        v.endVisit(this);\n");
        ast_buffer.Put(indentation); ast_buffer.Put("    }\n");
    }

    return;
}


//
// Generate a generic rule class
//
void Grammar::GenerateRuleClass(CTC &ctc,
                                NTC &ntc,
                                TextBuffer &ast_buffer,
                                const char *indentation,
                                ClassnameElement &element,
                                Array<const char *> &typestring)
{
    char *classname = element.real_name;
    SymbolLookupTable &symbol_set = element.symbol_set;
    Tuple<int> &rhs_type_index = element.rhs_type_index;

    BitSet optimizable_symbol_set(element.symbol_set.Size(), BitSet::UNIVERSE);

    GenerateCommentHeader(ast_buffer, indentation, element);

    assert(element.rule.Length() == 1);
    int rule_no = element.rule[0];

    ast_buffer.Put(indentation); ast_buffer.Put(option -> automatic_ast == Option::NESTED ? "static " : "");
                                 ast_buffer.Put("public class ");
                                 ast_buffer.Put(classname);
                                 ast_buffer.Put(" extends ");
    if (element.is_terminal_class)
    {
        ast_buffer.Put(ast_token_classname);
        ast_buffer.Put(" implements ");
        ast_buffer.Put(typestring[rules[rule_no].lhs]);
        ast_buffer.Put("\n");
        ast_buffer.Put(indentation); ast_buffer.Put("{\n");
        if (element.needs_environment)
            GenerateEnvironmentDeclaration(ast_buffer, indentation);
        if (symbol_set.Size() == 1) // if the right-hand side contains a symbol ...
        {
            ast_buffer.Put(indentation); ast_buffer.Put("    public IToken get");
                                         ast_buffer.Put(symbol_set[0] -> Name());
                                         ast_buffer.Put("() { return leftIToken; }\n\n");
        }
        ast_buffer.Put(indentation); ast_buffer.Put("    public ");
                                     ast_buffer.Put(classname);
                                     ast_buffer.Put("(");
        if (element.needs_environment)
        {
            ast_buffer.Put(option -> action_type);
            ast_buffer.Put(" environment, IToken token)");
            ast_buffer.Put(indentation); ast_buffer.Put("    {\n");
            ast_buffer.Put(indentation); ast_buffer.Put("        super(token);\n");
            ast_buffer.Put(indentation); ast_buffer.Put("        this.environment = environment;\n");
            ast_buffer.Put(indentation); ast_buffer.Put("        initialize();\n");
            ast_buffer.Put(indentation); ast_buffer.Put("    }\n");
        }
        else ast_buffer.Put("IToken token) { super(token); initialize(); }\n");
    }
    else 
    {
        ast_buffer.Put(option -> ast_type);
        ast_buffer.Put(" implements ");
        ast_buffer.Put(typestring[rules[rule_no].lhs]);
        ast_buffer.Put("\n");
        ast_buffer.Put(indentation); ast_buffer.Put("{\n");
        if (element.needs_environment)
            GenerateEnvironmentDeclaration(ast_buffer, indentation);

        if (symbol_set.Size() > 0)
        {
            for (int i = 0; i < symbol_set.Size(); i++)
            {
                ast_buffer.Put(indentation); ast_buffer.Put("    private ");
                                             ast_buffer.Put(ctc.FindBestTypeFor(rhs_type_index[i]));
                                             ast_buffer.Put(" _");
                                             ast_buffer.Put(symbol_set[i] -> Name());
                                             ast_buffer.Put(";\n");
            }
            ast_buffer.Put("\n");

            for (int i = 0; i < symbol_set.Size(); i++)
            {
                if (ntc.CanProduceNullAst(rhs_type_index[i]))
                {
                    ast_buffer.Put(indentation); ast_buffer.Put("    /**\n");
                    ast_buffer.Put(indentation); ast_buffer.Put("     * The value returned by <b>get");
                                                 ast_buffer.Put(symbol_set[i] -> Name());
                                                 ast_buffer.Put("</b> may be <b>null</b>\n");
                    ast_buffer.Put(indentation); ast_buffer.Put("     */\n");
                }

                ast_buffer.Put(indentation); ast_buffer.Put("    public ");
                                             ast_buffer.Put(ctc.FindBestTypeFor(rhs_type_index[i]));
                                             ast_buffer.Put(" get");
                                             ast_buffer.Put(symbol_set[i] -> Name());
                                             ast_buffer.Put("() { return _");
                                             ast_buffer.Put(symbol_set[i] -> Name());
                                             ast_buffer.Put("; }\n");
            }
            ast_buffer.Put("\n");
        }

        //
        // generate constructor
        //
        const char *header = "    public ";
        ast_buffer.Put(indentation);
        ast_buffer.Put(header);
        ast_buffer.Put(classname);
        int length = strlen(indentation) + strlen(header) + strlen(classname);

        ast_buffer.Put("(");
        if (element.needs_environment)
        {
            ast_buffer.Put(option -> action_type);
            ast_buffer.Put(" environment, ");
        }
        ast_buffer.Put("IToken leftIToken, IToken rightIToken");
        ast_buffer.Put(symbol_set.Size() == 0 ? ")\n" : ",\n");
        for (int i = 0; i < symbol_set.Size(); i++)
        {
            for (int k = 0; k <= length; k++)
                ast_buffer.PutChar(' ');
            ast_buffer.Put(ctc.FindBestTypeFor(rhs_type_index[i]));
            ast_buffer.Put(" _");
            ast_buffer.Put(symbol_set[i] -> Name());
            ast_buffer.Put(i == symbol_set.Size() - 1 ? ")\n" : ",\n");
        }
        ast_buffer.Put(indentation); ast_buffer.Put("    {\n");
        ast_buffer.Put(indentation); ast_buffer.Put("        super(leftIToken, rightIToken);\n\n");
        if (element.needs_environment)
        {
            ast_buffer.Put(indentation);
            ast_buffer.Put("        this.environment = environment;\n");
        }
        for (int i = 0; i < symbol_set.Size(); i++)
        {
            ast_buffer.Put(indentation); ast_buffer.Put("        this._");
                                         ast_buffer.Put(symbol_set[i] -> Name());
                                         ast_buffer.Put(" = _");
                                         ast_buffer.Put(symbol_set[i] -> Name());
                                         ast_buffer.Put(";\n");

            if (option -> parent_saved)
            {
                ast_buffer.Put(indentation); ast_buffer.Put("        ");
                if ((! optimizable_symbol_set[i]) || ntc.CanProduceNullAst(rhs_type_index[i]))
                {
                    ast_buffer.Put("if (_");
                    ast_buffer.Put(symbol_set[i] -> Name());
                    ast_buffer.Put(" != null) ");
                }

                ast_buffer.Put("((");
                ast_buffer.Put(option -> ast_type);
                ast_buffer.Put(") _");
                ast_buffer.Put(symbol_set[i] -> Name());
                ast_buffer.Put(").setParent(this);\n");
            }
        }

        ast_buffer.Put(indentation); ast_buffer.Put("        initialize();\n");
        ast_buffer.Put(indentation); ast_buffer.Put("    }\n");
    }

    GenerateEqualsMethod(ntc, ast_buffer, indentation, element, optimizable_symbol_set);
    GenerateHashcodeMethod(ntc, ast_buffer, indentation, element, optimizable_symbol_set);
    GenerateVisitorMethods(ntc, ast_buffer, indentation, element, optimizable_symbol_set);

    return;
}


//
// Generate Ast class
//
void Grammar::GenerateTerminalMergedClass(NTC &ntc, 
                                          TextBuffer &ast_buffer,
                                          const char *indentation,
                                          ClassnameElement &element,
                                          Array<const char *> &typestring)
{
    char *classname = element.real_name;
    GenerateCommentHeader(ast_buffer, indentation, element);

    ast_buffer.Put(indentation); ast_buffer.Put(option -> automatic_ast == Option::NESTED ? "static " : "");
                                 ast_buffer.Put("public class ");
                                 ast_buffer.Put(classname);
                                 ast_buffer.Put(" extends ");
                                 ast_buffer.Put(ast_token_classname);
                                 ast_buffer.Put(" implements ");
    for (int i = 0; i < element.interface.Length() - 1; i++)
    {
        ast_buffer.Put(typestring[element.interface[i]]);
        ast_buffer.Put(", ");
    }
    ast_buffer.Put(typestring[element.interface[element.interface.Length() - 1]]);
    ast_buffer.Put("\n");
    ast_buffer.Put(indentation); ast_buffer.Put("{\n");
    if (element.needs_environment)
        GenerateEnvironmentDeclaration(ast_buffer, indentation);
    SymbolLookupTable &symbol_set = element.symbol_set;
    if (symbol_set.Size() == 1) // if the right-hand side contains a symbol ...
    {
        ast_buffer.Put(indentation); ast_buffer.Put("    public IToken get");
                                     ast_buffer.Put(symbol_set[0] -> Name());
                                     ast_buffer.Put("() { return leftIToken; }\n\n");
    }
    ast_buffer.Put(indentation); ast_buffer.Put("    public ");
                                 ast_buffer.Put(classname);
                                 ast_buffer.Put("(");
                                 if (element.needs_environment)
                                 {
                                     ast_buffer.Put(option -> action_type);
                                     ast_buffer.Put(" environment, IToken token)");
                                     ast_buffer.Put(indentation); ast_buffer.Put("    {\n");
                                     ast_buffer.Put(indentation); ast_buffer.Put("        super(token);\n");
                                     ast_buffer.Put(indentation); ast_buffer.Put("        this.environment = environment;\n");
                                     ast_buffer.Put(indentation); ast_buffer.Put("        initialize();\n");
                                     ast_buffer.Put(indentation); ast_buffer.Put("    }\n");
                                 }
                                 else ast_buffer.Put("IToken token) { super(token); initialize(); }\n");

    BitSet optimizable_symbol_set(element.symbol_set.Size(), BitSet::UNIVERSE);
    GenerateHashcodeMethod(ntc, ast_buffer, indentation, element, optimizable_symbol_set);
    GenerateVisitorMethods(ntc, ast_buffer, indentation, element, optimizable_symbol_set);

    return;
}


//
// Generate Ast class
//
void Grammar::GenerateMergedClass(CTC &ctc,
                                  NTC &ntc,
                                  TextBuffer &ast_buffer,
                                  const char *indentation,
                                  ClassnameElement &element,
                                  Tuple< Tuple<ProcessedRuleElement> > &processed_rule_map,
                                  Array<const char *> &typestring)
{
    char *classname = element.real_name;
    SymbolLookupTable &symbol_set = element.symbol_set;
    Tuple<int> &rhs_type_index = element.rhs_type_index;

    GenerateCommentHeader(ast_buffer, indentation, element);

    ast_buffer.Put(indentation); ast_buffer.Put(option -> automatic_ast == Option::NESTED ? "static " : "");
                                 ast_buffer.Put("public class ");
                                 ast_buffer.Put(classname);
                                 ast_buffer.Put(" extends ");
                                 ast_buffer.Put(option -> ast_type);
                                 ast_buffer.Put(" implements ");
    for (int i = 0; i < element.interface.Length() - 1; i++)
    {
        ast_buffer.Put(typestring[element.interface[i]]);
        ast_buffer.Put(", ");
    }
    ast_buffer.Put(typestring[element.interface[element.interface.Length() - 1]]);
    ast_buffer.Put("\n");
    ast_buffer.Put(indentation); ast_buffer.Put("{\n");
    if (element.needs_environment)
        GenerateEnvironmentDeclaration(ast_buffer, indentation);
    for (int i = 0; i < symbol_set.Size(); i++)
    {
        ast_buffer.Put(indentation); ast_buffer.Put("    private ");
                                     ast_buffer.Put(ctc.FindBestTypeFor(rhs_type_index[i]));
                                     ast_buffer.Put(" _");
                                     ast_buffer.Put(symbol_set[i] -> Name());
                                     ast_buffer.Put(";\n");
    }
    ast_buffer.Put("\n");

    //
    // Compute the set of symbols that always appear in an instance creation
    // of this merged class for which a null instance allocation will never occur.
    //
    BitSet optimizable_symbol_set(element.symbol_set.Size(), BitSet::UNIVERSE);
    Tuple<int> &rule = element.rule;
    for (int i = 0; i < rule.Length(); i++)
    {
        int rule_no = rule[i];
        Tuple<ProcessedRuleElement> &processed_rule_elements = processed_rule_map[rule_no];
        for (int k = 0; k < processed_rule_elements.Length(); k++)
        {
            if (processed_rule_elements[k].position == 0 ||
                ntc.CanProduceNullAst(rhs_sym[processed_rule_elements[k].position]))
                 optimizable_symbol_set.RemoveElement(k);
        }
    }

    for (int i = 0; i < symbol_set.Size(); i++)
    {
        if ((! optimizable_symbol_set[i]) || ntc.CanProduceNullAst(rhs_type_index[i]))
        {
            ast_buffer.Put(indentation); ast_buffer.Put("    /**\n");
            ast_buffer.Put(indentation); ast_buffer.Put("     * The value returned by <b>get");
                                         ast_buffer.Put(symbol_set[i] -> Name());
                                         ast_buffer.Put("</b> may be <b>null</b>\n");
            ast_buffer.Put(indentation); ast_buffer.Put("     */\n");
        }

        ast_buffer.Put(indentation); ast_buffer.Put("    public ");
                                     ast_buffer.Put(ctc.FindBestTypeFor(rhs_type_index[i]));
                                     ast_buffer.Put(" get");
                                     ast_buffer.Put(symbol_set[i] -> Name());
                                     ast_buffer.Put("() { return _");
                                     ast_buffer.Put(symbol_set[i] -> Name());
                                     ast_buffer.Put("; }\n");
    }
    ast_buffer.Put("\n");


    //
    // generate merged constructor
    //
    const char *header = "    public ";
    ast_buffer.Put(indentation);
    ast_buffer.Put(header);
    ast_buffer.Put(classname);
    int length = strlen(indentation) + strlen(header) + strlen(classname);

    ast_buffer.Put("(");
    if (element.needs_environment)
    {
        ast_buffer.Put(option -> action_type);
        ast_buffer.Put(" environment, ");
    }
    ast_buffer.Put("IToken leftIToken, IToken rightIToken");
    ast_buffer.Put(symbol_set.Size() == 0 ? ")\n" : ",\n");
    for (int i = 0; i < symbol_set.Size(); i++)
    {
        for (int k = 0; k <= length; k++)
            ast_buffer.PutChar(' ');
        ast_buffer.Put(ctc.FindBestTypeFor(rhs_type_index[i]));
        ast_buffer.Put(" _");
        ast_buffer.Put(symbol_set[i] -> Name());
        ast_buffer.Put(i == symbol_set.Size() - 1 ? ")\n" : ",\n");
    }
    ast_buffer.Put(indentation); ast_buffer.Put("    {\n");
    ast_buffer.Put(indentation); ast_buffer.Put("        super(leftIToken, rightIToken);\n\n");
    if (element.needs_environment)
    {
        ast_buffer.Put(indentation);
        ast_buffer.Put("        this.environment = environment;\n");
    }
    for (int i = 0; i < symbol_set.Size(); i++)
    {
        ast_buffer.Put(indentation); ast_buffer.Put("        this._");
                                     ast_buffer.Put(symbol_set[i] -> Name());
                                     ast_buffer.Put(" = _");
                                     ast_buffer.Put(symbol_set[i] -> Name());
                                     ast_buffer.Put(";\n");

        if (option -> parent_saved)
        {
            ast_buffer.Put(indentation); ast_buffer.Put("        ");
            if ((! optimizable_symbol_set[i]) || ntc.CanProduceNullAst(rhs_type_index[i]))
            {
                ast_buffer.Put("if (_");
                ast_buffer.Put(symbol_set[i] -> Name());
                ast_buffer.Put(" != null) ");
            }

            ast_buffer.Put("((");
            ast_buffer.Put(option -> ast_type);
            ast_buffer.Put(") _");
            ast_buffer.Put(symbol_set[i] -> Name());
            ast_buffer.Put(").setParent(this);\n");
        }
    }

    ast_buffer.Put(indentation); ast_buffer.Put("        initialize();\n");
    ast_buffer.Put(indentation); ast_buffer.Put("    }\n");

    GenerateEqualsMethod(ntc, ast_buffer, indentation, element, optimizable_symbol_set);
    GenerateHashcodeMethod(ntc, ast_buffer, indentation, element, optimizable_symbol_set);
    GenerateVisitorMethods(ntc, ast_buffer, indentation, element, optimizable_symbol_set);

    return;
}


void Grammar::GenerateInterface(bool is_terminal,
                                TextBuffer &ast_buffer,
                                const char *indentation,
                                const char *interface_name,
                                Tuple<int> &extension,
                                Tuple<int> &classes,
                                Tuple<ClassnameElement> &classname)
{
    ast_buffer.Put(indentation); ast_buffer.Put("/**");
    if (is_terminal)
    {
        ast_buffer.Put("\n");
        ast_buffer.Put(indentation);  ast_buffer.Put(" * is always implemented by <b>");
                                      ast_buffer.Put(ast_token_classname);
                                      ast_buffer.Put("</b>. It is also implemented by");
    }
    else 
    {
        ast_buffer.Put("\n");
        ast_buffer.Put(indentation);
        ast_buffer.Put(" * is implemented by");
    }

    if (classes.Length() == 1)
    {
        ast_buffer.Put(" <b>");
        ast_buffer.Put(classname[classes[0]].real_name);
        ast_buffer.Put("</b>");
    }
    else
    {
        ast_buffer.Put(":\n");
        ast_buffer.Put(indentation);
        ast_buffer.Put(" *<b>\n");
        ast_buffer.Put(indentation); ast_buffer.Put(" *<ul>");
        for (int i = 0; i < classes.Length(); i++)
        {
            ast_buffer.Put("\n");
            ast_buffer.Put(indentation);
            ast_buffer.Put(" *<li>");
            ast_buffer.Put(classname[classes[i]].real_name);
        }
        ast_buffer.Put("\n");
        ast_buffer.Put(indentation);
        ast_buffer.Put(" *</ul>\n");
        ast_buffer.Put(indentation);
        ast_buffer.Put(" *</b>");
    }

    ast_buffer.Put("\n");
    ast_buffer.Put(indentation);
    ast_buffer.Put(" */\n");

    ast_buffer.Put(indentation); ast_buffer.Put("public interface ");
                                 ast_buffer.Put(interface_name);
    if (extension.Length() > 0)
    {
        ast_buffer.Put(" extends ");
        for (int k = 0; k < extension.Length() - 1; k++)
        {
            ast_buffer.PutChar('I');
            ast_buffer.Put(extension[k] == ast_token_interface
                               ? ast_token_classname
                               : RetrieveString(extension[k]));
            ast_buffer.Put(", ");
        }
        ast_buffer.PutChar('I');
        ast_buffer.Put(extension[extension.Length() - 1] == ast_token_interface
                               ? ast_token_classname
                               : RetrieveString(extension[extension.Length() - 1]));
        ast_buffer.Put(" {}\n\n");
    }
    else
    {
        ast_buffer.Put("\n");
        ast_buffer.Put(indentation); ast_buffer.Put("{\n");
        ast_buffer.Put(indentation); ast_buffer.Put("    public IToken getLeftIToken();\n");
        ast_buffer.Put(indentation); ast_buffer.Put("    public IToken getRightIToken();\n");
        ast_buffer.Put("\n");
        GenerateVisitorHeaders(ast_buffer, indentation, "    ");
        ast_buffer.Put(indentation); ast_buffer.Put("}\n\n");
    }

    return;
}


//
//
//
void Grammar::GenerateNullAstAllocation(Grammar *grammar, TextBuffer &ast_buffer, int rule_no)
{
    const char *code = "\n                    setResult(null);";
    grammar -> GenerateCode(&ast_buffer, code, rule_no);

    return;
}

//
//
//
void Grammar::GenerateAstAllocation(CTC &ctc,
                                    Grammar *grammar,
                                    TextBuffer &ast_buffer,
                                    RuleAllocationElement &allocation_element,
                                    Tuple<ProcessedRuleElement> &processed_rule_elements,
                                    Array<const char *> &typestring,
                                    int rule_no)
{
    char *classname = allocation_element.name;

    // 
    // Copy these two arrays into a local tuple for CONVENIENCE.
    // 
    Tuple<int> position,
               type_index;
    for (int i = 0; i < processed_rule_elements.Length(); i++)
    {
        position.Next() = processed_rule_elements[i].position;
        type_index.Next() = processed_rule_elements[i].type_index;
    }

    // 
    // Convenient constant declarations.
    // 
    const char *space = "\n                    ",
               *space4 = "    ",
               *newkey = option -> factory,
               *lparen = "(",
               *comma = ",",
               *rparen = ")",
               *trailer = ");";
    int extra_space_length = strlen(space) + strlen(space4) + strlen(newkey) + strlen(classname) + 1;
    char *extra_space = new char[extra_space_length + 1];
    extra_space[0] = '\n';
    for (int i = 1; i < extra_space_length; i++)
        extra_space[i] = ' ';
    extra_space[extra_space_length] = '\0';

    //
    // TODO: We simply generate a comment as a reminder that the previous nonterminal
    // allocated for this token should be deleted when using a language such as C/C++
    // that does not have a garbage collector.
    //
    //    if (allocation_element.is_terminal_class && type_index.Length() == 1 && IsNonTerminal(type_index[0]))
    //    {
    //        grammar -> GenerateCode(&ast_buffer, space, rule_no);
    //        grammar -> GenerateCode(&ast_buffer, "// When garbage collection is not available, delete ", rule_no);
    //        grammar -> GenerateCode(&ast_buffer, "getRhsSym(", rule_no);
    //        IntToString index(position[0]);
    //        grammar -> GenerateCode(&ast_buffer, index.String(), rule_no);
    //        grammar -> GenerateCode(&ast_buffer, rparen, rule_no);
    //    }
    if (allocation_element.is_terminal_class && (RhsSize(rule_no) == 1 && IsNonTerminal(rhs_sym[FirstRhsIndex(rule_no)])))
    {
        grammar -> GenerateCode(&ast_buffer, space, rule_no);
        grammar -> GenerateCode(&ast_buffer, "//", rule_no);
        grammar -> GenerateCode(&ast_buffer, space, rule_no);
        grammar -> GenerateCode(&ast_buffer, "// When garbage collection is not available, delete ", rule_no);
        grammar -> GenerateCode(&ast_buffer, "getRhsSym(1)", rule_no);
        grammar -> GenerateCode(&ast_buffer, space, rule_no);
        grammar -> GenerateCode(&ast_buffer, "//", rule_no);
    }
    grammar -> GenerateCode(&ast_buffer, space, rule_no);
    grammar -> GenerateCode(&ast_buffer, "setResult(", rule_no);
    grammar -> GenerateCode(&ast_buffer, space, rule_no);
    grammar -> GenerateCode(&ast_buffer, space4, rule_no);
    grammar -> GenerateCode(&ast_buffer, newkey, rule_no);
    grammar -> GenerateCode(&ast_buffer, classname, rule_no);
    grammar -> GenerateCode(&ast_buffer, lparen, rule_no);
    if (allocation_element.needs_environment)
    {
        grammar -> GenerateCode(&ast_buffer, option -> action_type, rule_no);
        grammar -> GenerateCode(&ast_buffer, ".this, ", rule_no);
    }
    if (allocation_element.is_terminal_class)
    {
        grammar -> GenerateCode(&ast_buffer, "getRhsIToken(1)", rule_no);
        //
        // TODO: Old bad idea. Remove at some point...
        //
        //
	//        assert(position.Length() <= 1);
	//
	//        grammar -> GenerateCode(&ast_buffer, "getRhsIToken(", rule_no);
	//        IntToString index(position.Length() == 0 ? 1 : position[0]);
	//        grammar -> GenerateCode(&ast_buffer, index.String(), rule_no);
	//        grammar -> GenerateCode(&ast_buffer, rparen, rule_no);
    }
    else
    {
        grammar -> GenerateCode(&ast_buffer, "getLeftIToken()", rule_no);
        grammar -> GenerateCode(&ast_buffer, ", ", rule_no);
        grammar -> GenerateCode(&ast_buffer, "getRightIToken()", rule_no);
        if (position.Length() > 0)
        {
            grammar -> GenerateCode(&ast_buffer, comma, rule_no);
            grammar -> GenerateCode(&ast_buffer, extra_space, rule_no);
    
            int offset = FirstRhsIndex(rule_no) - 1;
            for (int i = 0; i < position.Length(); i++)
            {
                if (position[i] == 0)
                {
                    grammar -> GenerateCode(&ast_buffer, lparen, rule_no);
                    grammar -> GenerateCode(&ast_buffer, ctc.FindBestTypeFor(type_index[i]), rule_no);
                    grammar -> GenerateCode(&ast_buffer, rparen, rule_no);
                    grammar -> GenerateCode(&ast_buffer, "null", rule_no);
                }
                else
                {
                    int symbol = rhs_sym[offset + position[i]];
                    if (IsTerminal(symbol))
                    {
                        grammar -> GenerateCode(&ast_buffer, newkey, rule_no);
                        grammar -> GenerateCode(&ast_buffer, ast_token_classname, rule_no);
                        grammar -> GenerateCode(&ast_buffer, lparen, rule_no);
                        grammar -> GenerateCode(&ast_buffer, "getRhsIToken(", rule_no);
                        IntToString index(position[i]);
                        grammar -> GenerateCode(&ast_buffer, index.String(), rule_no);
                        grammar -> GenerateCode(&ast_buffer, rparen, rule_no);
                    }
                    else
                    {
                        grammar -> GenerateCode(&ast_buffer, lparen, rule_no);
                        grammar -> GenerateCode(&ast_buffer, ctc.FindBestTypeFor(type_index[i]), rule_no);
                        grammar -> GenerateCode(&ast_buffer, rparen, rule_no);
                        grammar -> GenerateCode(&ast_buffer, "getRhsSym(", rule_no);
                        IntToString index(position[i]);
                        grammar -> GenerateCode(&ast_buffer, index.String(), rule_no);
                    }
    
                    grammar -> GenerateCode(&ast_buffer, rparen, rule_no);
                }
        
                if (i != position.Length() - 1)
                {
                    grammar -> GenerateCode(&ast_buffer, comma, rule_no);
                    grammar -> GenerateCode(&ast_buffer, extra_space, rule_no);
                }
            }
        }
    }

    grammar -> GenerateCode(&ast_buffer, rparen, rule_no);
    grammar -> GenerateCode(&ast_buffer, space, rule_no);
    grammar -> GenerateCode(&ast_buffer, trailer, rule_no);

    delete [] extra_space;

    return;
}

//
//
//
void Grammar::GenerateListAllocation(CTC &ctc,
                                     Grammar *grammar,
                                     TextBuffer &ast_buffer,
                                     int rule_no,
                                     RuleAllocationElement &allocation_element)
{
    const char *space = "\n                    ",
               *space4 = "    ",
               *newkey = option -> factory,
               *lparen = "(",
               *comma = ",",
               *rparen = ")",
               *trailer = ");";
    int extra_space_length = strlen(space) + strlen(space4) + strlen(newkey) + strlen(allocation_element.name) + 1;
    char *extra_space = new char[extra_space_length + 1];
    extra_space[0] = '\n';
    for (int i = 1; i < extra_space_length; i++)
        extra_space[i] = ' ';
    extra_space[extra_space_length] = '\0';

    if (allocation_element.list_kind == RuleAllocationElement::LEFT_RECURSIVE_EMPTY ||
        allocation_element.list_kind == RuleAllocationElement::RIGHT_RECURSIVE_EMPTY ||
        allocation_element.list_kind == RuleAllocationElement::LEFT_RECURSIVE_SINGLETON ||
        allocation_element.list_kind == RuleAllocationElement::RIGHT_RECURSIVE_SINGLETON)
    {
        grammar -> GenerateCode(&ast_buffer, space, rule_no);
        grammar -> GenerateCode(&ast_buffer, "setResult(", rule_no);
        grammar -> GenerateCode(&ast_buffer, space, rule_no);
        grammar -> GenerateCode(&ast_buffer, space4, rule_no);

        grammar -> GenerateCode(&ast_buffer, newkey, rule_no);
        grammar -> GenerateCode(&ast_buffer, allocation_element.name, rule_no);
        grammar -> GenerateCode(&ast_buffer, lparen, rule_no);
        if (allocation_element.needs_environment)
        {
            grammar -> GenerateCode(&ast_buffer, option -> action_type, rule_no);
            grammar -> GenerateCode(&ast_buffer, ".this, ", rule_no);
        }
        if (allocation_element.list_kind == RuleAllocationElement::LEFT_RECURSIVE_EMPTY ||
            allocation_element.list_kind == RuleAllocationElement::RIGHT_RECURSIVE_EMPTY)
        {
            grammar -> GenerateCode(&ast_buffer, "getLeftIToken()", rule_no);
            grammar -> GenerateCode(&ast_buffer, ", ", rule_no);
            grammar -> GenerateCode(&ast_buffer, "getRightIToken()", rule_no);
            grammar -> GenerateCode(&ast_buffer, comma, rule_no);
            if (allocation_element.list_kind == RuleAllocationElement::LEFT_RECURSIVE_EMPTY)
                 grammar -> GenerateCode(&ast_buffer, " true /* left recursive */", rule_no);
            else grammar -> GenerateCode(&ast_buffer, " false /* not left recursive */", rule_no);
        }
        else
        {
            assert(allocation_element.list_kind == RuleAllocationElement::LEFT_RECURSIVE_SINGLETON ||
                   allocation_element.list_kind == RuleAllocationElement::RIGHT_RECURSIVE_SINGLETON);

            if (IsTerminal(allocation_element.element_symbol))
            {
                grammar -> GenerateCode(&ast_buffer, newkey, rule_no);
                grammar -> GenerateCode(&ast_buffer, ast_token_classname, rule_no);
                grammar -> GenerateCode(&ast_buffer, lparen, rule_no);
                grammar -> GenerateCode(&ast_buffer, "getRhsIToken(", rule_no);
                IntToString index(allocation_element.element_position);
                grammar -> GenerateCode(&ast_buffer, index.String(), rule_no);
                grammar -> GenerateCode(&ast_buffer, rparen, rule_no);
            }
            else
            {
                grammar -> GenerateCode(&ast_buffer, lparen, rule_no);
                grammar -> GenerateCode(&ast_buffer, ctc.FindBestTypeFor(allocation_element.element_type_symbol_index), rule_no);
                grammar -> GenerateCode(&ast_buffer, rparen, rule_no);
                grammar -> GenerateCode(&ast_buffer, "getRhsSym(", rule_no);
                IntToString index(allocation_element.element_position);
                grammar -> GenerateCode(&ast_buffer, index.String(), rule_no);
            }
    
            grammar -> GenerateCode(&ast_buffer, rparen, rule_no);
            grammar -> GenerateCode(&ast_buffer, comma, rule_no);
            if (allocation_element.list_kind == RuleAllocationElement::LEFT_RECURSIVE_SINGLETON)
                 grammar -> GenerateCode(&ast_buffer, " true /* left recursive */", rule_no);
            else grammar -> GenerateCode(&ast_buffer, " false /* not left recursive */", rule_no);
        }

        grammar -> GenerateCode(&ast_buffer, rparen, rule_no);
        grammar -> GenerateCode(&ast_buffer, space, rule_no);
    }
    else
    {
        //
        // Add new element to list
        //
        if (allocation_element.list_kind == RuleAllocationElement::ADD_ELEMENT)
        {
            grammar -> GenerateCode(&ast_buffer, space, rule_no);
            grammar -> GenerateCode(&ast_buffer, lparen, rule_no);
            grammar -> GenerateCode(&ast_buffer, lparen, rule_no);
            grammar -> GenerateCode(&ast_buffer, allocation_element.name, rule_no);
            grammar -> GenerateCode(&ast_buffer, rparen, rule_no);
            grammar -> GenerateCode(&ast_buffer, "getRhsSym(", rule_no);
            IntToString index(allocation_element.list_position);
            grammar -> GenerateCode(&ast_buffer, index.String(), rule_no);
            grammar -> GenerateCode(&ast_buffer, ")).add(", rule_no);
            if (IsTerminal(allocation_element.element_symbol))
            {
                grammar -> GenerateCode(&ast_buffer, newkey, rule_no);
                grammar -> GenerateCode(&ast_buffer, ast_token_classname, rule_no);
                grammar -> GenerateCode(&ast_buffer, lparen, rule_no);
                grammar -> GenerateCode(&ast_buffer, "getRhsIToken(", rule_no);
                IntToString index(allocation_element.element_position);
                grammar -> GenerateCode(&ast_buffer, index.String(), rule_no);
                grammar -> GenerateCode(&ast_buffer, rparen, rule_no);
            }
            else
            {
                grammar -> GenerateCode(&ast_buffer, lparen, rule_no);
                grammar -> GenerateCode(&ast_buffer, ctc.FindBestTypeFor(allocation_element.element_type_symbol_index), rule_no);
                grammar -> GenerateCode(&ast_buffer, rparen, rule_no);
                grammar -> GenerateCode(&ast_buffer, "getRhsSym(", rule_no);
                IntToString index(allocation_element.element_position);
                grammar -> GenerateCode(&ast_buffer, index.String(), rule_no);
            }

            if (allocation_element.list_position != 1) // a right-recursive rule? set the list as result
            {
                grammar -> GenerateCode(&ast_buffer, rparen, rule_no);
                grammar -> GenerateCode(&ast_buffer, trailer, rule_no);

                grammar -> GenerateCode(&ast_buffer, space, rule_no);
                grammar -> GenerateCode(&ast_buffer, "setResult(", rule_no);
                grammar -> GenerateCode(&ast_buffer, "getRhsSym(", rule_no);
                IntToString index(allocation_element.list_position);
                grammar -> GenerateCode(&ast_buffer, index.String(), rule_no);
            }
        }

        //
        // Copy a list that is not the first element on the right-hand side of the rule
        //
        else
        {
            assert(allocation_element.list_kind == RuleAllocationElement::COPY_LIST);

            grammar -> GenerateCode(&ast_buffer, space, rule_no);
            grammar -> GenerateCode(&ast_buffer, "setResult(", rule_no);
            grammar -> GenerateCode(&ast_buffer, lparen, rule_no);
            grammar -> GenerateCode(&ast_buffer, allocation_element.name, rule_no);
            grammar -> GenerateCode(&ast_buffer, rparen, rule_no);
            grammar -> GenerateCode(&ast_buffer, "getRhsSym(", rule_no);
            IntToString index(allocation_element.list_position);
            grammar -> GenerateCode(&ast_buffer, index.String(), rule_no);
        }

        grammar -> GenerateCode(&ast_buffer, rparen, rule_no);
    }

    grammar -> GenerateCode(&ast_buffer, trailer, rule_no);
 
    delete [] extra_space;

    return;
}
